//-------------------------------------------------------------------------
/**
 * @file 	p25_constants.h
 * @brief	Defines constants used by the P25 Waveform
 *
 * @author	Cal(it)2
 * @author  Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009, The Regents of the University of California
 * All Rights Reserve
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
//-------------------------------------------------------------------------

#ifndef _P25_CONSTANTS_H
#define _P25_CONSTANTS_H

namespace P25 {

//-------------------------------------------------------------------------
// P25 Shared Frame components
//-------------------------------------------------------------------------
/** @brief FrameSync bit length. */
#define FS_BLEN             48
/** @brief Network ID bit length. */
#define NID_BLEN            64
/** @brief Sync bit length. */
#define SS_BLEN             2
/** @brief Voice Code Word bit length. */
#define VC_BLEN             144

//-------------------------------------------------------------------------
// Header Data Unit (HDU)
//-------------------------------------------------------------------------
/** @brief Header Data Unit (HDU) total bit length. */
#define HDU_BLEN			792
/** @brief HDU payload bit length. */
#define HCW_BLEN            648
/** @brief HDU padding bit length. */
#define HDU_PAD_BLEN        10

//-------------------------------------------------------------------------
// Termination Data Unit (TDU)
//-------------------------------------------------------------------------
/** @brief Termination Data Unit (TDU) with Link Control Word (LCW) bit length. */
#define TDU_LCW_BLEN		432
/** @brief TDU with LCW payload bit length. */
#define TDU_LC_BLEN        	288
/** @brief TDU with LCW padding bit length. */
#define TDU_LCW_PAD_BLEN    20
/** @brief TDU bit length. */
#define TDU_BLEN			144
/** @brief TDU padding bit length. */
#define TDU_PAD_BLEN        28

//-------------------------------------------------------------------------
// Logical Data Unit (LDU)
//-------------------------------------------------------------------------
/** @brief Logical Data Unit (LDU) bit length. */
#define LDU_BLEN			1728
/** @brief LDU payload bit length. */
#define LDU_VC_BLEN         1680-FS_BLEN-NID_BLEN

//-------------------------------------------------------------------------
// Frame Control components
//-------------------------------------------------------------------------
/** @brief Low Speed Data bit length. */
#define LSD_BLEN            32
/** @brief Link Control Word bit length. */
#define LCW_BLEN            240
/** @brief Encryption Sync Word bit length. */
#define ESW_BLEN            240

//-------------------------------------------------------------------------
// Data Unit ID's
//-------------------------------------------------------------------------
/** @brief Header data unit ID. */
#define DUID_HDU        0
/** @brief Terminator data unit ID. */
#define DUID_TDU        3
/** @brief Logical data unit 1 ID. */
#define DUID_LDU1       5
/** @brief Trunking signaling block ID. */
#define DUID_TSB        7
/** @brief Logical data unit 2 ID. */
#define DUID_LDU2       10
/** @brief Packet data unit ID. */
#define DUID_PDU        12
/** @brief Terminator data unit with Link Control Word ID. */
#define DUID_TDU_LCW	15

//-------------------------------------------------------------------------
// Derived constants
//-------------------------------------------------------------------------
/** @brief NID offset. @see P25 Training Guide Chapter 4 */
#define NID1_OFFSET			FS_BLEN
/** @brief First part of NID length. @see P25 Training Guide Chapter 4*/
#define NID1_BLEN			22
/** @brief First status symbol offset. @see P25 Training Guide Chapter 4*/
#define SS_OFFSET			(NID1_OFFSET + NID1_BLEN)
/** @brief Offset for rest of NID. @see P25 Training Guide Chapter 4*/
#define NID2_OFFSET			(SS_OFFSET + SS_BLEN)
/** @brief Remainder of NID size. @see P25 Training Guide Chapter 4*/
#define NID2_BLEN			42
/** @brief Offset of remainder of Header data unit. @see P25 Training Guide Chapter 4*/
#define HDR_BLEN			(NID2_OFFSET + NID2_BLEN)

/** 
 * @brief Maximum bit length of P25 frame. 
 * @see LDU_BLEN
 */
#define MAX_FRAME_SIZE		LDU_BLEN
/** @brief Hexbit size. */
#define HEXBIT      6
/** @brief Size of playback buffer (18 VCWs x length of VCW). */
#define PLAY_BACK_BUF_SIZE  18*VC_BLEN /*LDU*/

//-------------------------------------------------------------------------
// Link Control Word Formats used in Conventional System
//  Below are the Link Control Opcodes
//-------------------------------------------------------------------------
/** @brief Opcode for Group Voice Channel User. */
#define LC_GRP_V_CH_USR         0
/** @ brief Opcode for */
#define LC_UU_V_CH_USR          3
#define LC_TELE_INT_V_CH_USR    6
#define LC_SOURCE_ID_EXT        9
#define LC_UU_V_CH_USR_EXT      10
#define LC_CALL_TRM_CAN         15
#define LC_STS_Q                19
#define LC_STA_UPDT             20
#define LC_MSG_UPDT             21
#define LC_CALL_ALRT            22
#define LC_EXT_FNCT_CMD         23
#define LC_CH_ID_UPDT           24

//-------------------------------------------------------------------------
// Radio Activity Constants
//-------------------------------------------------------------------------
/** @brief Transmit Activity. */
#define TX_ACTIVITY     	21
/** @brief Receive Activity. */
#define RX_ACTIVITY     	22
/** @brief Standby Activity. */
#define STANDBY_ACTIVITY	23

//-------------------------------------------------------------------------
// Radio State Constants
//-------------------------------------------------------------------------
/** @brief Running State. */
#define RUNNING_STATE	11
/** @brief Shutdown State. */
#define SHUTDOWN_STATE	12

//-------------------------------------------------------------------------
// Radio Mode Constants
//-------------------------------------------------------------------------
/** @brief Test Mode. */
#define TEST_MODE     	1
/** @brief Live Mode. */
#define LIVE_MODE     	2


//static unsigned short   frameSync[3] = {0x5575,0xf5ff,0x77ff};
/** @brief P25 Frame Synchronization pattern with each byte representing a bit. */
const char	fs[FS_BLEN] = {
	0x00,0x01,0x00,0x01,0x00,0x01,0x00,0x01,
	0x00,0x01,0x01,0x01,0x00,0x01,0x00,0x01,
	0x01,0x01,0x01,0x01,0x00,0x01,0x00,0x01,
	0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
	0x00,0x01,0x01,0x01,0x00,0x01,0x01,0x01,
	0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01
};
/** @brief Error Threshold for Frame Sync. */
#define FS_ERROR_THRESHOLD 10

//-------------------------------------------------------------------------
// ALGID Constants
//-------------------------------------------------------------------------
/** @brief Unencrypted ALGID. */
#define ALGID_UNENCRYPTED	0x80
/** @brief DES ALGID. */
#define ALGID_DES			0x81

}

#endif
