/*
 * File:    bch.c
 * Title:   Encoder/decoder for binary BCH codes in C (Version 3.1)
 * Author:  Robert Morelos-Zaragoza
 * Date:    August 1994
 * Revised: June 13, 1997
 *
 * ===============  Encoder/Decoder for binary BCH codes in C =================
 *
 * Version 1:   Original program. The user provides the generator polynomial
 *              of the code (cumbersome!).
 * Version 2:   Computes the generator polynomial of the code.
 * Version 3:   No need to input the coefficients of a primitive polynomial of
 *              degree M, used to construct the Galois Field GF(2^M). The
 *              program now works for any binary BCH code of length such that:
 *              2^(M-1) - 1 < length <= 2^M - 1
 *
 * Note:        You may have to change the size of the arrays to make it work.
 *
 * The encoding and decoding methods used in this program are based on the
 * book "Error Control Coding: Fundamentals and Applications", by Lin and
 * Costello, Prentice Hall, 1983.
 *
 * Thanks to Patrick Boyle (pboyle@era.com) for his observation that 'bch2.c'
 * did not work for lengths other than 2^M-1 which led to this new version.
 * Portions of this program are from 'rs.c', a Reed-Solomon encoder/decoder
 * in C, written by Simon Rockliff (simon@augean.ua.oz.au) on 21/9/89. The
 * previous version of the BCH encoder/decoder in C, 'bch2.c', was written by
 * Robert Morelos-Zaragoza (robert@spectra.eng.hawaii.edu) on 5/19/92.
 *
 * NOTE:    
 *          The author is not responsible for any malfunctioning of
 *          this program, nor for any damage caused by it. Please include the
 *          original program along with these comments in any redistribution.
 *
 *  For more information, suggestions, or other ideas on implementing error
 *  correcting codes, please contact me at:
 *
 *                           Robert Morelos-Zaragoza
 *                           5120 Woodway, Suite 7036
 *                           Houston, Texas 77056
 *
 *                    email: r.morelos-zaragoza@ieee.org
 *
 * COPYRIGHT INFO: This computer program is free for non-commercial purposes.
 * You may implement this program for any non-commercial application. You may 
 * also implement this program for commercial purposes, provided that you
 * obtain my written permission. Any modification of this program is covered
 * by this copyright.
 *
 * == Copyright (c) 1994-7,  Robert Morelos-Zaragoza. All rights reserved.  ==
 */

#include "p25/ecc/bch.h"

#define M			6							/* order of GF */
#define N			((1 << M) - 1)				/* code length */
#define K			16							/* data length */
#define T			11							/* max errors  */

static char alpha_to[N + 1];
static char index_of[N + 1];

static char pp[M + 1] = {1, 1, 0, 0, 0, 0, 1};	/* primitive polynomial */
static char gp[N - K + 1] = {					/* generator polynomial */
	1, 1, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1,
	1, 1, 0, 0, 1, 0, 1, 1, 1, 0, 1, 1,
	1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0,
	1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1
};

/*
 * Generate field GF(2^M) from primitive polynomial.
 * alpha_to: return vector form of a power index
 * index_of: return power from of a vector value
 */
static void generate_gf()
{
	int i, mask;

	mask = 1;
	alpha_to[M] = 0;
	for (i = 0; i < M; i++) {
		alpha_to[i] = mask;
		index_of[alpha_to[i]] = i;
		if (pp[i] != 0)
			alpha_to[M] ^= mask;
		mask <<= 1;
	}
	index_of[alpha_to[M]] = M;
	mask >>= 1;
	for (i = M + 1; i < N; i++) {
		if (alpha_to[i - 1] >= mask)
			alpha_to[i] = alpha_to[M]^((alpha_to[i - 1]^mask)<<1);
		else
			alpha_to[i] = alpha_to[i - 1] << 1;
		index_of[alpha_to[i]] = i;
	}
	index_of[0] = -1;
}

static void reverse(char *array, int n)
{
    int i;
    char temp;

    for (i = 0; i < n/2; i++) {
        temp = array[i];
        array[i] = array[n-i-1];
        array[n-i-1] = temp;
    }
}

/*
 * BCH encode. The parity polynomial bp(x) is the remainder after
 * dividing x^(N-K)*data(x) by the generator polynomial g(x).
 */
void libdsp::bch_encode(char *data)
{
	int i, j;
	char feedback;
	char *bp;

	/* FIXME: reverse */
	reverse(data, K);

	bp = data + K;
	for (i = 0; i < N - K; i++)
		bp[i] = 0;

	for (i = K - 1; i >= 0; i--) {
		feedback = data[i] ^ bp[N - K - 1];
		if (feedback != 0) {
			for (j = N - K - 1; j > 0; j--)
				if (gp[j] != 0)
					bp[j] = bp[j - 1] ^ feedback;
				else
					bp[j] = bp[j - 1];
			bp[0] = gp[0] && feedback;
		} else {
			for (j = N - K - 1; j > 0; j--)
				bp[j] = bp[j - 1];
			bp[0] = 0;
		}
	}

	/* FIXME: reverse */
	reverse(data, K);
	reverse(bp, N - K);
}

/* 
 * BCH decode. Following the terminology of Lin and Costello's book:
 * Error Control Coding 1983.
 *
 * Compute the 2*T syndromes by substituting alpha^i into code(X) and
 * evaluating, storing the syndromes in s[i], i=1..2*T (leave s[0] zero).
 * Then we use the Berlekamp algorithm to find the error location polynomial
 * elp[i].
 *
 * If the degree of the elp is >T, then we cannot correct all the errors,
 * and we have detected an uncorrectable error pattern. We output the
 * information bits uncorrected.
 *
 * If the degree of elp is <=T, we substitute alpha^i , i=1..N into
 * the elp to get the roots, hence the inverse roots, the error location
 * numbers. This step is usually called "Chien's search".
 *
 * If the number of errors located is not equal the degree of the elp, then
 * the decoder assumes that there are more than t errors and cannot correct
 * them, only detect them. We output the information bits uncorrected.
 */

int libdsp::bch_decode(char *code)
{
	char s[2*T+1], elp[2*T+1][2*T+1], d[2*T+1], l[2*T+1], u_lu[2*T+1];
   	char loc[N], reg[T+1];
	int i, j, u, q, count, syn_error;
	int ret = 0;

	/* FIXME: reverse */
	reverse(code, N);

	/* 1. Compute syndromes */

	syn_error = 0;
	for (i = 1; i <= 2*T; i++) {
		s[i] = 0;
		for (j = 0; j < N; j++) {
			if (code[j] != 0)
				s[i] ^= alpha_to[(i * j) % N];
		}
		if (s[i] != 0)
			syn_error = 1;
		s[i] = index_of[s[i]];
	}

	if (!syn_error) {
		ret = 0;
		goto out;
	}

	/* 2. Determing error-location polynomial. */

	d[0] = 0;			/* index form */
	d[1] = s[1];		/* index form */
	elp[0][0] = 0;		/* index form */
	elp[1][0] = 1;		/* polynomial form */
	for (i = 1; i < 2*T; i++) {
		elp[0][i] = -1;	/* index form */
		elp[1][i] = 0;	/* polynomial form */
	}
	l[0] = 0;
	l[1] = 0;
	u_lu[0] = -1;
	u_lu[1] = 0;
	u = 0;

	do {
		u++;
		if (d[u] == -1) {
			l[u + 1] = l[u];
			for (i = 0; i <= l[u]; i++) {
				elp[u + 1][i] = elp[u][i];
				elp[u][i] = index_of[elp[u][i]];
			}
		} else {
			/* search for words with greatest u_lu[q] for
			 * which d[q]!=0 */
			q = u - 1;
			while ((d[q] == -1) && (q > 0))
				q--;
			/* have found first non-zero d[q]  */
			if (q > 0) {
				j = q;
				do {
					j--;
					if ((d[j] != -1) && (u_lu[q] < u_lu[j]))
						q = j;
				} while (j > 0);
			}

   			/* have now found q such that d[u]!=0 and
	   		 * u_lu[q] is maximum */
   		
			/* store degree of new elp polynomial */
	   		if (l[u] > l[q] + u - q)
   				l[u + 1] = l[u];
	   		else
   				l[u + 1] = l[q] + u - q;

	   		/* form new elp(x) */
   			for (i = 0; i < 2*T; i++)
   				elp[u + 1][i] = 0;
	   		for (i = 0; i <= l[q]; i++)
   				if (elp[q][i] != -1)
   					elp[u + 1][i + u - q] = 
   						alpha_to[(d[u] + N - d[q] + elp[q][i]) % N];
	   		for (i = 0; i <= l[u]; i++) {
   				elp[u + 1][i] ^= elp[u][i];
   				elp[u][i] = index_of[elp[u][i]];
	   		}
   		}
	   	u_lu[u + 1] = u - l[u + 1];

   		/* form (u+1)th discrepancy */
	   	if (u < 2*T) {	
   			/* no discrepancy computed on last iteration */
   			if (s[u + 1] != -1)
   				d[u + 1] = alpha_to[s[u + 1]];
	   		else
   				d[u + 1] = 0;
	   		for (i = 1; i <= l[u + 1]; i++)
   				if ((s[u + 1 - i] != -1) && (elp[u + 1][i] != 0))
   					d[u + 1] ^= alpha_to[(s[u + 1 - i] 
   						      + index_of[elp[u + 1][i]]) % N];
	   		/* put d[u+1] into index form */
   			d[u + 1] = index_of[d[u + 1]];	
		}
	} while ((u < T*2) && (l[u + 1] <= T));

	u++;
	if (l[u] > T) {
		ret = -1;
		goto out;
	}
	
	/* 3. Determine error locations with Chien search */

	for (i = 0; i <= l[u]; i++)
		reg[i] = index_of[elp[u][i]];

	count = 0;
	for (i = 1; i <= N; i++) {
		q = 1;
		for (j = 1; j <= l[u]; j++)
			if (reg[j] != -1) {
				reg[j] = (reg[j] + j) % N;
				q ^= alpha_to[reg[j]];
			}
		if (!q) {
			loc[count] = N - i;
			count++;
		}
	}
	if (count != l[u]) {
		ret = -2;
		goto out;
	}

	for (i = 0; i < l[u]; i++)
		code[loc[i]] ^= 1;
out:
	reverse(code, N);
	return ret;
}

void libdsp::bch_init()
{
	generate_gf();
}
