///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	intfc.cpp
 * @brief	Function definitions of the dsp::intfc namespace. This class
 * 			defines the interface between the DSP module and the GPP and
 * 			Radio modules of the P25 SDR implementation.
 * @author	Cal(it)2
 * @author  Sivaramakrishnan Chandrasekaran <bscrk29@gmail.com>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009, The Regents of the University of California
 * All Rights Reserve
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cstring>
#include "p25/comm/ipcomm.h"

#include "modem.hpp"
#include "intfc.hpp"

// tick-tock timer macros
#include <sys/time.h>
#include <iostream>
#define TICK() struct timeval tp1, tp2;						\
{															\
	gettimeofday(&tp1, NULL);								\
}

#define _TOCK(fmt, thresh)	{								\
	gettimeofday(&tp2, NULL);								\
	if (((tp2.tv_usec - tp1.tv_usec) / 1e3) > thresh) 		\
		std::cout << fmt << " time:  " << 					\
			(tp2.tv_usec - tp1.tv_usec) / 1e3 << " ms."     \
				<< std::endl;								\
}

#define TOCK(fmt)		_TOCK(fmt, 0)


using namespace std;

clock_t last_fpga_update = 0;


/**
 * @brief Channel Objects
 */
P25::Comm::IPComm* modem2gpp_chan;
P25::Comm::IPComm* gpp2modem_chan;

/**
 * @brief Use socket objects as an alternate for FIFO pipes
 */
P25::Comm::IPComm* modem2uhd_chan;
P25::Comm::IPComm* uhd2modem_chan;

FILE *tx_fm_file;

//Initialize objects
int dsp::intfc::intfc_init(const char *tx_spath , const char *rx_spath, const char *uhd_tx_spath , const char *uhd_rx_spath ) {

	// Setup Modem to GPP channels
	modem2gpp_chan = new P25::Comm::IPComm(tx_spath);
	gpp2modem_chan = new P25::Comm::IPComm(rx_spath);
	modem2gpp_chan->start_connection(P25::Comm::IPComm::TXMode);
	gpp2modem_chan->start_connection(P25::Comm::IPComm::RXMode);

	//Setup Modem to UHD channels
	printf("Opening connection to:  %s\n", uhd_tx_spath);
	modem2uhd_chan = new P25::Comm::IPComm(uhd_tx_spath);
	modem2uhd_chan->start_connection(P25::Comm::IPComm::TXMode);

	printf("Opening connection to:  %s\n", uhd_rx_spath);
	uhd2modem_chan = new P25::Comm::IPComm(uhd_rx_spath);
	uhd2modem_chan->start_connection(P25::Comm::IPComm::RXMode);

	return 0;
}

// Not used
int dsp::intfc::uhdi_tx_en(int en)
{
	return 0;
}

//Not used
int dsp::intfc::uhdi_rx_en(int en)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//Functions that read/write from interfaces between modules
///////////////////////////////////////////////////////////////////////////////

/**
 * Read data (bits) from the GPP module through a socket interface.
 * @param data the bitstream
 * @param len the length of the stream to read
 * @return the length of data read
 */
int dsp::intfc::read_gppi(char *data, int len)
{
	int res, i;
	res = gpp2modem_chan->receive(data, len);
	if (res != len) {
		std::cout << res << " of " << len << " bytes read\n\n";
		return -1;
	}

	return res;
}

/**
 * Write data (bits) to the GPP module through a socket interface.
 * @param data the bitstream
 * @param len the length of the stream to write
 * @return the length of data written
 */
int dsp::intfc::write_gppi(char *data, int len)
{
	int res;
	res = modem2gpp_chan->send(data, len);
	if (res != len) {
		std::cout << res << " of " << len << " bytes sent\n\n";
		return -1;
	}

	return res;
}

//Not used. To be removed.
int dsp::intfc::read_gppi_reset(void)
{
	return 0;
}

//Not used. To be removed
int dsp::intfc::write_gppi_flush(void)
{
	return 0;
}


/**
 * Read data (stream of floats) from the radio (UHD) module through a FIFO pipe.
 * @param data the float stream
 * @param len the length of the stream
 * @return the length of data read
 */

int dsp::intfc::read_uhdi(char *data, int len)
{
	int res = 0;

	res = uhd2modem_chan->receive(data, len);

	if (res != len) {
		std::cout << res << " of " << len << " bytes read\n\n";
		return -1;
	}

  	return res;
}

/**
 * Write data (stream of floats) to the radio (UHD) module through a FIFO pipe.
 * @param data the float stream
 * @param len the length of the stream
 * @return the length of data written
 */

int dsp::intfc::write_uhdi(char *data, int len)
{
	int res = 0;

	res = modem2uhd_chan->send(data, len);
	if (res != len) {
		std::cout << res << " of " << len << " bytes sent\n\n";
		return -1;
	}

	return res;
}

//Not used
int dsp::intfc::read_uhdi_reset(void) //never used
{
	return 0;
}

//Not used. To be removed.
int dsp::intfc::write_uhdi_flush(void) //never used
{
	return 0;
}
