///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	p25_frame.cpp
 * @brief	P25 framing helper functions
 * @details	Much of this code has been ported to linux from the original code
 * 			written by Wenhua Zhao for the SDR4000 TI DSP  platform.
 *
 * @author	Cal(it)2
 * @author  Sivaramakrishnan Chandrasekaran <bscrk29@gmail.com>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <cstdio>

#include "p25/constants.h"
#include "p25/ecc/bch.h"
#include "dsp/p25_frame.hpp"

using namespace libdsp;
/*
 * @brief		get the frame length
 * @param nid	the NID of the frame
 *
 * @return 		the frame length
 */
static int get_frame_length(char *nid)
{
	int duid, len;

	duid = nid[12] << 3 | nid[13] << 2 | nid[14] << 1 | nid[15];

	switch (duid) {
		case DUID_HDU:
			printf("HDU\n");
			len = HDU_BLEN;
			break;
		case DUID_TDU:
			printf("TDU\n");
			len = TDU_BLEN;
			break;
		case DUID_TDU_LCW:
			printf("TDU_LCW\n");
			len = TDU_LCW_BLEN;
			break;
		case DUID_LDU1:
		case DUID_LDU2:
			printf("LDU1/2\n");
			len = LDU_BLEN;
			break;
		case DUID_PDU:
		case DUID_TSB:
		default:
			printf("unknown length for duid %d\n", duid);
			len = -1;
	}

	return len;
}

/*
 * @brief	encode the frame header
 * 			BCH encode the NID field
 *
 * @return 	the frame length
 */
int frame_header_encode(char *hdr)
{
	char nid[NID1_BLEN + NID2_BLEN];

	memcpy(nid, hdr + NID1_OFFSET, NID1_BLEN);
	memcpy(nid + NID1_BLEN, hdr + NID2_OFFSET, NID2_BLEN);

	bch_encode(nid);

	memcpy(hdr + NID1_OFFSET, nid, NID1_BLEN);
	memcpy(hdr + NID2_OFFSET, nid + NID1_BLEN, NID2_BLEN);

	return get_frame_length(nid);
}

/*
 * @brief	decode the frame header
 * 			BCH decode the NID field
 *
 * @return	the frame length
 */
int frame_header_decode(char* hdr)
{
	char nid[NID1_BLEN + NID2_BLEN];

	memcpy(nid, hdr + NID1_OFFSET, NID1_BLEN);
	memcpy(nid + NID1_BLEN, hdr + NID2_OFFSET, NID2_BLEN);

	if (bch_decode(nid) < 0) {
		printf("bch decode failure\n");
		return -1;
	}

	memcpy(hdr + NID1_OFFSET, nid, NID1_BLEN);
	memcpy(hdr + NID2_OFFSET, nid + NID1_BLEN, NID2_BLEN);

	return get_frame_length(nid);
}
