///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	Freq_Mod.hpp
 * @brief	Declarations for the Frequency Modulator (FM).
 *
 * @author	Cal(it)2
 * @author  Minsoo Kang
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////

#ifndef _FREQ_MOD
#define _FREQ_MOD


//// FM modulation
class Freq_Mod {
public:
	/**
	 * Constructor that sets the sampling rate and freq deviation for FM modulator.
	 * 
	 * @param tx_rate 	Sampling rate (= 48ksps)
	 * @param tx_delta 	Frequency deviation
	 * 
	 */
	Freq_Mod(double, unsigned int);
	~Freq_Mod(void);
	
	/**
	 * FM modulation function.
	 */
	void fm_mod(float*, float*, float*, size_t);
	
	void reset();
	
	static float sum;

private:
	/**
	 * Length of complex input array
	 */
	unsigned int len_input;
	
	/**
	 * Sampling rate
	 */
	double rate;
	
	/**
	 * Frequency deviation
	 */
	unsigned int delta;
};



//// FM demodulation
class Freq_Demod {
public:
	/**
	 * Constructor that sets the sampling rate and freq deviation for FM demodulator.
	 *
	 * @param rx_rate 	Sampling rate (= 48ksps)
	 * @param rx_delta 	Frequency deviation
	 *
	 */
	Freq_Demod(double, unsigned int);
	~Freq_Demod(void);
	
	/**
	 * Calculate the angle between two inputs.
	 */
	float angle(float, float);

	/**
	 * Compares the input with PI=3.14xxxxxx.
	 *
	 * @return 	1 if the input is greater than PI
	 * 			0 otherwise
	 */
	size_t comp(float);

	/**
	 * FM demodulation function
	 */
	void fm_demod(float*, float*, float*, size_t);

	static float prev_phase;

	void reset();

private:
	
	// Sampling rate which is 48ksps
	double rate;
	
	// freq deviation
	unsigned int delta;
};




//// generate a sine wave
class gen_sine {
public:
	gen_sine();
	~gen_sine(void);

	float sine(float);

	void gen_sine_wave(float*,double, double, float, size_t);

	float theta;
	
	float* sine_table;

private:
	double rate, wave_freq;

	float cps;
	float tx_ampl;
};

#endif
