///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	radio.cpp
 * @brief	Implementation of the Radio module of the P25 SDR Implementation
 *
 * @author	Cal(it)2
 * @author 	Min Soo Kang
 * @author  Sivaramakrishnan Chandrasekaran <bscrk29@gmail.com>
 * @author  Jeffrey Cuenco <jcuenco@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009, The Regents of the University of California
 * All Rights Reserve
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 *
 *		04/11/2011		    Jeffrey Cuenco	    Runtime frequency-changing
 *												via control channel
 *
 *		04/30/2011			Jeffrey Cuenco		Changed baseband-to-modem
 *												component communication from
 *												FIFOs to Unix Domain Sockets
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////

#include <uhd/utils/thread_priority.hpp>
#include <uhd/utils/safe_main.hpp>
#include <uhd/utils/static.hpp>
#include <uhd/usrp/multi_usrp.hpp>
#include <uhd/types/device_addr.hpp>
#include <uhd/types/metadata.hpp>
#include <boost/program_options.hpp>
#include <boost/thread/thread_time.hpp>
#include <boost/math/special_functions/round.hpp>
#include <boost/format.hpp>
#include <boost/thread.hpp>
#include <boost/function.hpp>
#include <iostream>
#include <complex>
#include <cmath>
#include <fstream>
#include <pthread.h>
#include <cstdio>
#include <cstring>
#include <fcntl.h>
#include <unistd.h>
#include <semaphore.h>

#include "radio.hpp"
#include "Freq_Mod.hpp"
#include "resample.hpp"
#include "p25/comm/ipcomm.h"
#include "filter.hpp"
#include "fir_coefs.h"

namespace po = boost::program_options;
using namespace std;

// Unix Domain Sockets for waveform/control data exchange.
P25::Comm::IPComm* modem2uhd_chan;
P25::Comm::IPComm* uhd2modem_chan;
P25::Comm::IPComm* uhd_modem_ctrl_chan;
P25::Comm::IPComm* uhd_gpp_ctrl_chan;
P25::Comm::IPComm* ctrl_chan_modem_baseband;

// Resampling filter objects
Resample_OP *upsamp_I;
Resample_OP *upsamp_Q;
Resample_OP *downsamp_I;
Resample_OP *downsamp_Q;

float *fir; 	//Filter Coeffs for Tx
float *rx_fir; 	//Filter Coeffs for Rx
uhd::usrp::multi_usrp::sptr sdev; 	// Pointer to USRP parent device
uhd::device::sptr dev; 				// Pointer to a USRP device

// FIFOs for internal loopback test mode
std::string uhd_loopback_tx_fifo, uhd_loopback_rx_fifo;

int ptt_swtch = 0;
sem_t ptt_sem;
pthread_mutex_t ptt_mutex;

bool first_tx = false;
bool first_rx = true;

typedef struct {
	unsigned char signal[8];
	double cfreq;
} control_pkt;

// tick-tock timer macros
#include <sys/time.h>
#include <iostream>
#define TICK() struct timeval tp1, tp2;						\
{															\
	gettimeofday(&tp1, NULL);								\
}

#define _TOCK(fmt, thresh)	{								\
	gettimeofday(&tp2, NULL);								\
	if (((tp2.tv_usec - tp1.tv_usec) / 1e3) > thresh) 		\
		std::cout << fmt << " time:  " << 					\
			(tp2.tv_usec - tp1.tv_usec) / 1e3 << " ms."     \
				<< std::endl;								\
}

#define TOCK(fmt)		_TOCK(fmt, 0)

////////////////////// MAIN ////////////////

int UHD_SAFE_MAIN(int argc, char *argv[]) 
{
	std::string uhd_tx_spath, uhd_rx_spath;
	std::string uhd_modem_ctrl_chan_path, uhd_gpp_ctrl_chan_path;
	std::string ctrl_chan_modem_baseband_path;
	
	size_t spb;
	double tx_rate, tx_freq, tx_gain, tx_bw;
	float tx_ampl;
	double rx_rate, rx_freq, rx_gain, rx_bw;
	std::string args = "";
	std::string file = "";
	void *status;

	// Setup the program options	
	po::options_description desc("Allowed options");
	desc.add_options()
		("help", "help message")
		("args", po::value<std::string>(&args)->default_value(""), 
			"Single uhd device address args")
		("tx_isock", po::value<std::string>(&uhd_rx_spath)->default_value("dsp_uhd_sock"), 
			"TX input socket")
		("rx_osock", po::value<std::string>(&uhd_tx_spath)->default_value("uhd_dsp_sock"), 
			"RX output socket")
		("rc_sock", po::value<std::string>(&uhd_modem_ctrl_chan_path)->default_value("dsp_ctrl_sock"), 
			"Radio control socket")
		("fc_sock", po::value<std::string>(&uhd_gpp_ctrl_chan_path)->default_value("uhd_ctrl_sock"), 
			"Frequency control socket")
		("bbc_sock", po::value<std::string>(&ctrl_chan_modem_baseband_path)->default_value("modem_baseband_ctrl_sock"), 
			"Modem -> Baseband ACK control socket")
		("spb", po::value<size_t>(&spb)->default_value(1000), "samples per buffer")
		("tx_rate", po::value<double>(&tx_rate)->default_value(200e3), 
			"Rate of outgoing samples")
		("rx_rate", po::value<double>(&rx_rate)->default_value(200e3), 
			"Rate of incoming samples")
		("tx_freq", po::value<double>(&tx_freq)->default_value(139.75e6), 
			"RF output center frequency in Hz")
		("rx_freq", po::value<double>(&rx_freq)->default_value(139.75e6), 
			"RF input center frequency in Hz")
		("tx_ampl", po::value<float>(&tx_ampl)->default_value(float(1)), 
			"Amplitude of the output waveform")
		("tx_gain", po::value<double>(&tx_gain)->default_value(35), "Gain for the RF TX chain")
		("rx_gain", po::value<double>(&rx_gain)->default_value(0), "Gain for the RF RX chain")
		("tx_bw", po::value<double>(&tx_bw)->default_value(20e6), "IF BW (MHz) for the TX chain")
		("rx_bw", po::value<double>(&rx_bw)->default_value(20e6), "IF BW (MHz) for the RX chain")
		("tx_lfifo", po::value<std::string>(&uhd_loopback_tx_fifo)->default_value(""),
			"[Optional] Modem (No UHD) Test Loopback FIFO for TX thread")
		("rx_lfifo", po::value<std::string>(&uhd_loopback_rx_fifo)->default_value(""),
			"[Optional] Modem (No UHD) Test Loopback FIFO for RX thread")
	;
	po::variables_map vm;
	po::store(po::parse_command_line(argc, argv, desc), vm);
	po::notify(vm);

	// print the help message
	if (vm.count("help")){
		std::cout << boost::format("P25 Modem/RF Frontend %s") % desc << std::endl;
		return ~0;
	}


	// Setup Modem to UHD channels
	cout << "Opening connection to:  " << uhd_tx_spath.c_str() << endl;
	uhd2modem_chan = new P25::Comm::IPComm(uhd_tx_spath.c_str());
	if (uhd2modem_chan->start_connection(P25::Comm::IPComm::TXMode) != 0) {
		cerr << "Error starting connection to uhd2modem_chan!" << endl;
		return ~0;
	}
	cout << "Opening connection to:  " << uhd_rx_spath.c_str() << endl;
	modem2uhd_chan = new P25::Comm::IPComm(uhd_rx_spath.c_str());
	if (modem2uhd_chan->start_connection(P25::Comm::IPComm::RXMode) != 0) {
		cerr << "Error starting connection to modem2uhd_chan!" << endl;
		return ~0;
	}

	// Setup control channel b/w DSP modem and UHD radio
	uhd_modem_ctrl_chan = new P25::Comm::IPComm(uhd_modem_ctrl_chan_path.c_str());
	if (uhd_modem_ctrl_chan->start_connection(P25::Comm::IPComm::RXMode) != 0) {
		cerr << "Error starting connection to uhd_modem_ctrl_chan!" << endl;
		return ~0;
	}

	// Setup control channel b/w GPP and UHD radio
	uhd_gpp_ctrl_chan = new P25::Comm::IPComm(uhd_gpp_ctrl_chan_path.c_str());
	if (uhd_gpp_ctrl_chan->start_connection(P25::Comm::IPComm::RXMode) != 0) {
		cerr << "Error starting connection to uhd_gpp_ctrl_chan!" << endl;
		return ~0;
	}

	// Setup control channel from MODEM to BASEBAND
	ctrl_chan_modem_baseband = new P25::Comm::IPComm(ctrl_chan_modem_baseband_path.c_str());
	if (ctrl_chan_modem_baseband->start_connection(P25::Comm::IPComm::TXMode) != 0) {
		cerr << "Error starting connection to ctrl_chan_modem_baseband!" << endl;
		return ~0;
	}

	ptt_swtch=0;

	if (sem_init(&ptt_sem, 0, 0) < 0)
		perror("Semaphore ptt_sem init failed");

	pthread_mutex_init(&ptt_mutex, NULL);

///////////////////////////////////////////////////////////////////////////////
//
//Create a USRP device and set its radio parameters
//
///////////////////////////////////////////////////////////////////////////////


	if ((uhd_loopback_tx_fifo == "") || (uhd_loopback_rx_fifo == "")) {
		uhd::set_thread_priority_safe();

		cout << boost::format("Creating the usrp device with: %s...") % args << endl;
		sdev = uhd::usrp::multi_usrp::make(uhd::device_addr_t(args));
		dev = sdev->get_device();
		cout << boost::format("Using Device: %s") % sdev->get_pp_string() << endl;

		//////////// tx ////////////////

		// set the tx sample rate
		cout << boost::format("Setting TX Rate: %f Msps...") % (tx_rate/1e6) << endl;
		sdev->set_tx_rate(tx_rate, uhd::usrp::multi_usrp::ALL_CHANS);
		cout << boost::format("Actual TX Rate: %f Msps...") % (sdev->get_tx_rate()/1e6)
			<< endl << endl;

		// set the tx center frequency
		set_txfreq(tx_freq);

		// set the tx rf gain
		cout << boost::format("Setting TX Gain: %f dB...") % tx_gain << endl;
		sdev->set_tx_gain(tx_gain, 0);
		cout << boost::format("Actual TX Gain: %f dB...") % sdev->get_tx_gain()
			<< endl << endl;

		// set the tx if bandwidth
		cout << boost::format("Setting TX Bandwidth: %f MHz...") % (tx_bw/1e6) << endl;
		sdev->set_tx_bandwidth(tx_bw, 0);
		cout << boost::format("Actual TX Bandwidth: %f MHz...") % (sdev->get_tx_bandwidth() / 1e6)
				<< endl << endl;


		/////////////////////////////////

		//////////// rx /////////////////

		// set the rx sample rate
		cout << boost::format("Setting RX Rate: %f Msps...") % (rx_rate/1e6) << endl;
		sdev->set_rx_rate(rx_rate, uhd::usrp::multi_usrp::ALL_CHANS);
		cout << boost::format("Actual RX Rate: %f Msps...") % (sdev->get_rx_rate()/1e6)
			<< endl << endl;

		// set the rx center frequency
		set_rxfreq(rx_freq);

		// set the rx rf gain
		cout << boost::format("Setting RX Gain: %f dB...") % rx_gain << endl;
		sdev->set_rx_gain(rx_gain, 0);
		cout << boost::format("Actual RX Gain: %f dB...") % sdev->get_rx_gain() << endl << endl;

		// set the rx if bandwidth
		cout << boost::format("Setting RX Bandwidth: %f MHz...") % (rx_bw/1e6) << endl;
		sdev->set_rx_bandwidth(rx_bw, 0);
		cout << boost::format("Actual RX Bandwidth: %f MHz...") % (sdev->get_rx_bandwidth() / 1e6)
				<< endl << endl;

		boost::this_thread::sleep(boost::posix_time::seconds(1)); //allow for some setup time

		///////////////////////////////
	}

//////////////////////////////////End USRP setting/////////////////////////////

	// Read TX filter coefficients
	std::string coef_file = "./txfir_coef.bin";
	fir = new float[TX_FIR_LENGTH];
	fstream coeffile(coef_file.c_str(), ios::in | ios::out | ios::binary);
	coeffile.seekg(0);
	coeffile.read((char*)fir, sizeof(float)*TX_FIR_LENGTH);
	coeffile.close();

	// Read RX filter coefficients
	std::string rx_coef_file = "./rxfir_coef.bin";
	rx_fir = new float[RX_FIR_LENGTH];
	fstream rx_coeffile(rx_coef_file.c_str(), ios::in | ios::out | ios::binary);
	rx_coeffile.seekg(0);
	rx_coeffile.read((char*)rx_fir, sizeof(float)*RX_FIR_LENGTH);
	rx_coeffile.close();

	pthread_t rx_thread, tx_thread, ctrl_thread, freq_ctrl_thread;
	pthread_attr_t tx_attr, rx_attr, ctrl_attr, freq_ctrl_attr;

///////////////////// Start RX & TX /////////////////////////////////

	int ret;

	// Spawn Tx Thread
	struct sched_param sched_param_tx;
	sched_param_tx.sched_priority = 99;
	pthread_attr_init(&tx_attr);
	pthread_attr_setschedpolicy(&tx_attr, SCHED_FIFO);
	pthread_attr_setschedparam(&tx_attr, &sched_param_tx);
	ret = pthread_attr_setstacksize(&tx_attr,131072);
	ret = pthread_attr_setdetachstate(&tx_attr,PTHREAD_CREATE_JOINABLE);
	ret = pthread_create(&tx_thread, &tx_attr, uhd_tx , NULL);
	if (ret) cout<<"Failed to create Tx thread, cannot transmit.\n\n";

	// Spawn Rx Thread
	struct sched_param sched_param_rx;
	sched_param_rx.sched_priority = 99;
	pthread_attr_init(&rx_attr);
	pthread_attr_setschedpolicy(&rx_attr, SCHED_FIFO);
	pthread_attr_setschedparam(&rx_attr, &sched_param_rx);
	ret = pthread_attr_setstacksize(&rx_attr,131072);
	ret = pthread_attr_setdetachstate(&rx_attr,PTHREAD_CREATE_JOINABLE);
	ret = pthread_create(&rx_thread, &rx_attr, uhd_rx, NULL);
	if (ret) cout<<"Failed to create Rx thread, cannot receive...\n\n";

	// Spawn Control Thread
	pthread_attr_init(&ctrl_attr);
	ret = pthread_attr_setstacksize(&ctrl_attr,131072);
	ret = pthread_attr_setdetachstate(&ctrl_attr,PTHREAD_CREATE_JOINABLE);
	ret = pthread_create(&ctrl_thread, &ctrl_attr, control_sig, NULL);
	if (ret) cout<<"Failed to create Control Channel thread,..\n\n";

	// Spawn Frequency Control Thread
	pthread_attr_init(&freq_ctrl_attr);
	ret = pthread_attr_setstacksize(&freq_ctrl_attr,131072);
	ret = pthread_attr_setdetachstate(&freq_ctrl_attr,PTHREAD_CREATE_JOINABLE);
	ret = pthread_create(&freq_ctrl_thread, &freq_ctrl_attr, freq_sig, NULL);
	if (ret) cout<<"Failed to create Frequency Control Channel thread,..\n\n";

	pthread_join(tx_thread, &status);

	return 0;
}


void uhd_start_rx_stream()
{
	if ((uhd_loopback_tx_fifo == "") || (uhd_loopback_rx_fifo == "")) {

		//setup streaming
		int total_num_samps = 0;
		uhd::stream_cmd_t stream_cmd((total_num_samps == 0)?
		    uhd::stream_cmd_t::STREAM_MODE_START_CONTINUOUS:
		    uhd::stream_cmd_t::STREAM_MODE_NUM_SAMPS_AND_DONE
		);

		cout << "Setting up streaming:  STREAM_MODE_START_CONTINUOUS\n";

		stream_cmd.num_samps = total_num_samps;
		stream_cmd.stream_now = true;
		stream_cmd.time_spec = uhd::time_spec_t(0.0);

		sdev->issue_stream_cmd(stream_cmd);
	}
}

void uhd_stop_rx_stream()
{
	if ((uhd_loopback_tx_fifo == "") || (uhd_loopback_rx_fifo == "")) {
		sdev->issue_stream_cmd(uhd::stream_cmd_t::STREAM_MODE_STOP_CONTINUOUS);
	}
}

template<typename samp_type> 
size_t upsample_IQ(const samp_type* I_in, const samp_type* Q_in, std::vector<std::complex<samp_type> >& IQ_out, size_t num_samps)
{
	size_t ocount, odata_count;
	/*static*/ float us_I[10];
	/*static*/ float us_Q[10];

	memset(us_I, 0, 10);
	memset(us_Q, 0, 10);

	odata_count = 0;
	for (size_t i = 0; i < num_samps; i++) {
		ocount = 	upsamp_I->resample(I_in[i], us_I);
				upsamp_Q->resample(Q_in[i], us_Q);

		odata_count += ocount;

		if (ocount > 0) {
			for (size_t k = 0, j = (odata_count - ocount); j < odata_count; j++) {
				k = j - (odata_count - ocount);
				IQ_out[j].real() = us_I[k];
				IQ_out[j].imag() = us_Q[k];
			}
		}
	}

	return odata_count;
}

template<typename samp_type>
size_t downsample_IQ(std::vector<std::complex<samp_type> > const& IQ_in, samp_type* I_out, samp_type* Q_out, size_t num_samps)
{
	size_t ocount, odata_count;
	/*static*/ float ds_I[10];
	/*static*/ float ds_Q[10];

	memset(ds_I, 0, 10);
	memset(ds_Q, 0, 10);

	odata_count = 0;
	for (size_t i = 0; i < num_samps; i++) {
		ocount = 	downsamp_I->resample(IQ_in[i].real(), ds_I);
				downsamp_Q->resample(IQ_in[i].imag(), ds_Q);

		odata_count += ocount;

		if (ocount > 0) {
			for (size_t k = 0, j = (odata_count - ocount); j < odata_count; j++) {
				k = j - (odata_count - ocount);
				I_out[j] = ds_I[k];
				Q_out[j] = ds_Q[k];
			}
		}
	}

	return odata_count;
}

void upsample_IQ_flush()
{
	static float val = 0.0;
	static float rs[10];

	for (size_t i = 0; i < 1000; i++) {
		upsamp_I->resample(val, rs);
		upsamp_Q->resample(val, rs);
	}
}

void downsample_IQ_flush()
{
	static float val = 0.0;
	static float rs[10];

	for (size_t i = 0; i < 1000; i++) {
		downsamp_I->resample(val, rs);
		downsamp_Q->resample(val, rs);
	}
}

void uhd_get_async_md()
{
	std::cout << std::endl << "Waiting for async burst ACK... " << std::flush;
	uhd::async_metadata_t async_md;
	bool got_async_burst_ack = false;

	//loop through all messages for the ACK packet (may have underflow messages in queue)
	while (not got_async_burst_ack and dev->recv_async_msg(async_md, 0)) {
		got_async_burst_ack = (async_md.event_code == uhd::async_metadata_t::EVENT_CODE_BURST_ACK);
 	}
	std::cout << (got_async_burst_ack? "success" : "fail") << std::endl;
	switch (async_md.event_code) {
		case uhd::async_metadata_t::EVENT_CODE_BURST_ACK:
			cout << "Async metadata event:  EVENT_CODE_BURST_ACK" << endl;
			break;
		case uhd::async_metadata_t::EVENT_CODE_UNDERFLOW:
			cout << "Async metadata event:  EVENT_CODE_UNDERFLOW" << endl;
			break;
		case uhd::async_metadata_t::EVENT_CODE_SEQ_ERROR:
			cout << "Async metadata event:  EVENT_CODE_SEQ_ERROR" << endl;
			break;
		case uhd::async_metadata_t::EVENT_CODE_TIME_ERROR:
			cout << "Async metadata event:  EVENT_CODE_TIME_ERROR" << endl;
			break;
		case uhd::async_metadata_t::EVENT_CODE_UNDERFLOW_IN_PACKET:
			cout << "Async metadata event:  EVENT_CODE_UNDERFLOW_IN_PACKET" << endl;
			break;
		case uhd::async_metadata_t::EVENT_CODE_SEQ_ERROR_IN_BURST:
			cout << "Async metadata event:  EVENT_CODE_SEQ_ERROR_IN_BURST" << endl;
			break;
		default:
			cout << "Unknown Async metadata event: " << async_md.event_code << endl;
			break;
	}
}

template<typename samp_type> 
size_t uhd_send(const std::vector<samp_type>& buf, size_t len)
{
	size_t res = 0;
	//double seconds_in_future = 0.005;			// 5 ms
	//double timeout = seconds_in_future + 0.1;
	double timeout = 0.160;
	uhd::tx_metadata_t tx_md;

	tx_md.has_time_spec  = false;
	//tx_md.time_spec = uhd::time_spec_t(seconds_in_future);
	tx_md.start_of_burst = false;
	tx_md.end_of_burst   = false;

	if (len > buf.size()) {
		cout << "uhd_send():  len > buf.size, setting len = buf.size()." << endl;
		len = buf.size();
	}

	if (len != 0) {
	  resend_uhd:
		res = dev->send(&buf.front(), len, tx_md,
			uhd::io_type_t::COMPLEX_FLOAT32,
			uhd::device::SEND_MODE_FULL_BUFF,
			timeout
		);

		if (res == 0) {
			cout << "uhd_send() timeout:  res = " << res << ", len = " << len << endl;
			usleep(10);
			goto resend_uhd;
		}
		else {
			if (res != len) {
				cerr << "uhd_send() fail:  res = " << res << ", len = " << len << endl;
				uhd_get_async_md();
				sleep(1);
			}
		}
	}

	return res;
}

template<typename samp_type>
size_t uhd_recv(std::vector<samp_type>& buf, size_t len)
{
	uhd::rx_metadata_t rx_md;

	// Read data from USRP
	size_t num_samps = dev->recv(
		&buf.front(), len, rx_md,
		uhd::io_type_t::COMPLEX_FLOAT32,
		uhd::device::RECV_MODE_FULL_BUFF
	);

	// Check received metadata for error codes.
	switch (rx_md.error_code) {
		case uhd::rx_metadata_t::ERROR_CODE_NONE:
			break;
		case uhd::rx_metadata_t::ERROR_CODE_TIMEOUT:
			cerr << "RX_MD Error:  No packet received, implementation timed-out." << endl;
			sleep(1);
			break;
		case uhd::rx_metadata_t::ERROR_CODE_LATE_COMMAND:
			cerr << "RX_MD Error:  A stream command was issued in the past." << endl;
			sleep(1);
			break;
		case uhd::rx_metadata_t::ERROR_CODE_BROKEN_CHAIN:
			cerr << "RX_MD Error:  Expected another stream command." << endl;
			sleep(1);
			break;
		case uhd::rx_metadata_t::ERROR_CODE_OVERFLOW:
			cerr << "RX_MD Error:  An internal receive buffer has filled." << endl;
			sleep(1);
			break;
		case uhd::rx_metadata_t::ERROR_CODE_BAD_PACKET:
			cerr << "RX_MD Error:  The packet could not be parsed." << endl;
			sleep(1);
			break;
		default:
			cerr << "Unknown received metadata error code!" << endl;
			sleep(1);
	}

	return num_samps;
}

// Tx thread - Operation is controlled by signals from DSP/Baseband module via control channel
void *uhd_tx(void *)
{
// Variable and class object initializations: //////////////////////////////////
	int res;
	size_t samp_size, odata_count, num_sent, tx_spb;
	float data[CHUNK_LEN];
	float tx_fm_I[CHUNK_LEN];
	float tx_fm_Q[CHUNK_LEN];

	char ctrl_signal[1] = {0xAA};

	// FM Mod object
	Freq_Mod fm(FM_RATE, DELTA);

	// Upsampling output buffer
	tx_spb = (int)(CHUNK_LEN * 25.0 / 6.0);
	std::vector<std::complex<float> > tx_buff(tx_spb);

	// Upsample object
	upsamp_I = new Resample_OP(UP_SAMPL_RATE, DOWN_SAMPL_RATE, TX_NUM_SAMPS_ONE_SIDE,
			TX_FIR_LENGTH, fir);
	upsamp_Q = new Resample_OP(UP_SAMPL_RATE, DOWN_SAMPL_RATE, TX_NUM_SAMPS_ONE_SIDE,
			TX_FIR_LENGTH, fir);

	// FIFO only used when in modem loopback-test mode
	FILE *uhd_loopback_fifo = NULL;
	if ((uhd_loopback_tx_fifo != "") && (uhd_loopback_rx_fifo != "")) {
		if ((uhd_loopback_fifo = fopen (uhd_loopback_tx_fifo.c_str(), "wb")) == NULL) {
			cerr << "Open of uhd_loopback_tx_fifo failed!" << endl;
			return NULL;
		}
	}

#ifdef DEBUG_DUMP
	FILE *tx_c4_48k_file = fopen("tx_c4_48k.dat", "wb");
	if (tx_c4_48k_file == NULL)
		cerr << "Error opening up tx_c4_48k.dat!" << endl;

	FILE *tx_c4fm_48k_I_file = fopen("tx_c4fm_48k_I.dat", "wb");
	if (tx_c4fm_48k_I_file == NULL)
		cerr << "Error opening up tx_c4fm_48k_I.dat!" << endl;

	FILE *tx_c4fm_48k_Q_file = fopen("tx_c4fm_48k_Q.dat", "wb");
	if (tx_c4fm_48k_Q_file == NULL)
		cerr << "Error opening up tx_c4fm_48k_Q.dat!" << endl;

	FILE *tx_c4fm_200k_file = fopen("tx_c4fm_200k.dat", "wb");
	if (tx_c4fm_200k_file == NULL)
		cerr << "Error opening up tx_c4fm_200k.dat!" << endl;
#endif

// Main TX loop: ///////////////////////////////////////////////////////////////
	while(true) {
		// Receive data from modem socket
		samp_size = 0;
		res = modem2uhd_chan->receive((char *)data, CHUNK_LEN*sizeof(float));
		if (res != (int)(sizeof(float) * CHUNK_LEN)) {
			cout << res << " of " << (CHUNK_LEN * sizeof(float)) << " bytes read\n\n";
		}
		else {
			samp_size = (size_t)(res / sizeof(float));
		}

		odata_count = 0;
		if (samp_size != CHUNK_LEN) {
			cerr << "samp_size = " << samp_size << ", CHUNK_LEN = " << CHUNK_LEN << endl;
		}
		else {
			if (first_tx) {
				// Initialize IQ upsampler
				cout << "First TX packet - initializing upsample filter." << endl;
				upsample_IQ_flush();

				first_tx = false;
			}

#ifdef DEBUG_DUMP
			cout << "writing to file" << endl;
			int cnt = fwrite(data, sizeof(float), samp_size, tx_c4_48k_file);
			if (cnt != samp_size)
				cerr << "Error writing to tx_c4_48k.dat!" << endl;
#endif
			// FM modulation: /////////////////////////////////////////////////////////
			fm.fm_mod(data, tx_fm_I, tx_fm_Q, samp_size);

#ifdef DEBUG_DUMP
			cnt = fwrite(tx_fm_I, sizeof(float), samp_size, tx_c4fm_48k_I_file);
			if (cnt != samp_size)
				cerr << "Error writing to tx_c4fm_48k_I.dat!" << endl;

			cnt = fwrite(tx_fm_Q, sizeof(float), samp_size, tx_c4fm_48k_Q_file);
			if (cnt != samp_size)
				cerr << "Error writing to tx_c4fm_48k_Q.dat!" << endl;
#endif

			// Upsample to 200ksps: ///////////////////////////////////////////////////
			odata_count = upsample_IQ<float>(tx_fm_I, tx_fm_Q, tx_buff, samp_size);

#ifdef DEBUG_DUMP
			cnt = fwrite(&tx_buff.front(), sizeof(std::complex<float>), odata_count,
					tx_c4fm_200k_file);
			if (cnt != odata_count)
				cerr << "Error writing to tx_c4fm_200k.dat!" << endl;
#endif

			// Send the entire contents of the buffer: ////////////////////////////////
			num_sent = 0;
			if ((uhd_loopback_tx_fifo != "") || (uhd_loopback_rx_fifo != "")) {
				// Loopback Mode:  Send data to loopback FIFO
				usleep(18000);
				if (uhd_loopback_fifo != NULL) {
					num_sent = fwrite (&tx_buff.front(), sizeof(std::complex<float>), 
						odata_count, uhd_loopback_fifo);
				}
			}
			else {
				// UHD Mode:  Send data to UHD
				num_sent = uhd_send<std::complex<float> > (tx_buff, odata_count);
				if (num_sent != odata_count) {
					cerr << "num_sent = " << num_sent << ", odata_count = "
							<< odata_count << endl;
					uhd_get_async_md();
					sleep(1);
				}
			}
		}
		
		/*
		res = ctrl_chan_modem_baseband->send(ctrl_signal, sizeof(ctrl_signal));
		if (res != sizeof(ctrl_signal)) {
			cout << "Error Sending ACK to Baseband [res = " << res << "]\n";
		}*/

		fm.reset();
	}
}

// Rx thread - Operation is controlled by signals from DSP/Baseband module via control channel
void *uhd_rx(void *) 
{
// Variable and class object initializations: //////////////////////////////////
	int res, tx_swtch;
	size_t odata_count, num_rx_samps, rx_spb;
	float rs_I[CHUNK_LEN];
	float f_rs_I[CHUNK_LEN];
	float rs_Q[CHUNK_LEN];
	float f_rs_Q[CHUNK_LEN];
	float demod_out[CHUNK_LEN];
	struct fir *fir_lpf_I, *fir_lpf_Q, *fir_200_lpf_I, *fir_200_lpf_Q, *fir_lpf_demod;

	// FM Demod object
	Freq_Demod fd(FM_RATE, DELTA);

	// Downsampling input buffer
	rx_spb = (int)(CHUNK_LEN * 25.0 / 6.0);
	std::vector<std::complex<float> > rx_buff(rx_spb);

	// Downsample object
	downsamp_I = new Resample_OP(DOWN_SAMPL_RATE, UP_SAMPL_RATE, RX_NUM_SAMPS_ONE_SIDE,
			RX_FIR_LENGTH, rx_fir);
	downsamp_Q = new Resample_OP(DOWN_SAMPL_RATE, UP_SAMPL_RATE, RX_NUM_SAMPS_ONE_SIDE,
			RX_FIR_LENGTH, rx_fir);

	// Setup FIR pre-FM demod filter
	fir_200_lpf_I = fir_alloc(fir_coefs, FIR_COEFS_SIZE);
	fir_200_lpf_Q = fir_alloc(fir_coefs, FIR_COEFS_SIZE);
	fir_lpf_I = fir_alloc(fir_coefs_48k, FIR_COEFS_48K_SIZE);
	fir_lpf_Q = fir_alloc(fir_coefs_48k, FIR_COEFS_48K_SIZE);
	fir_lpf_demod = fir_alloc(fir_coefs_48k, FIR_COEFS_48K_SIZE);
        if (!fir_lpf_I || !fir_lpf_Q || !fir_lpf_demod) {
                printf("alloc fir_lpf failed\n");
                return NULL;
        }

	// FIFO only used when in modem loopback-test mode
	FILE *rx_uhd_loopback_fifo = NULL;
	if ((uhd_loopback_tx_fifo != "") && (uhd_loopback_rx_fifo != "")) {
		if ((rx_uhd_loopback_fifo = fopen (uhd_loopback_rx_fifo.c_str(), "rb")) == NULL) {
			cerr << "Open of uhd_loopback_rx_fifo failed!" << endl;
			return NULL;
		}
	}

#ifdef DEBUG_DUMP
	FILE *rx_c4fm_200k_file = fopen("rx_c4fm_200k.dat", "wb");
	if (rx_c4fm_200k_file == NULL)
		cerr << "Error opening up rx_c4fm_200k.dat!" << endl;

	FILE *rx_c4fm_48k_I_file = fopen("rx_c4fm_48k_I.dat", "wb");
	if (rx_c4fm_48k_I_file == NULL)
		cerr << "Error opening up rx_c4fm_48k_I.dat!" << endl;

	FILE *rx_c4fm_48k_Q_file = fopen("rx_c4fm_48k_Q.dat", "wb");
	if (rx_c4fm_48k_Q_file == NULL)
		cerr << "Error opening up rx_c4fm_48k_Q.dat!" << endl;

	FILE *rx_c4_48k_file = fopen("rx_c4_48k.dat", "wb");
	if (rx_c4_48k_file == NULL)
		cerr << "Error opening up rx_c4_48k.dat!" << endl;
#endif

	int tempcount = 0;


// Main RX loop: ///////////////////////////////////////////////////////////////

	pthread_mutex_lock(&ptt_mutex);
	tx_swtch = ptt_swtch;
	pthread_mutex_unlock(&ptt_mutex);

	// The outer while loop runs as long as the thread is alive
	while(true) {
		// Start UHD continuous streaming mode
		uhd_start_rx_stream();

		// This loop runs as long as there is nothing to transmit or in 
		// other words until the PTT switch is pressed
		while(!tx_swtch) { 
			odata_count = 0; 
			num_rx_samps = 0;

			// Receive new IQ samples at 200Ksps
			if ((uhd_loopback_tx_fifo != "") && (uhd_loopback_rx_fifo != "")) {
				// LOOPBACK MODE
				if (rx_uhd_loopback_fifo != NULL) {
					num_rx_samps = fread (&rx_buff.front(), sizeof(std::complex<float>),
						rx_buff.size(), rx_uhd_loopback_fifo);
				}
			}
			else {
				// UHD mode
				num_rx_samps = uhd_recv<std::complex<float> > (rx_buff, rx_buff.size());
			}

			// If received IQ samples matches desired size, proceed to rest of chain, else error.
			if (num_rx_samps != (size_t)rx_spb) {
				cout << "num_rx_samps = " << num_rx_samps << ", rx_spb = " << rx_spb
					<< endl;
			}
			else {
				if (first_rx) {
					// Initialize IQ downsampler
					cout << "First RX packet - initializing downsample filter." << endl;
					downsample_IQ_flush();

					first_rx = false;
				}

				// Energy detector
				float avg = 0;
				for (int i = 0; i < num_rx_samps; i++) {
					avg += (rx_buff[i].real()*rx_buff[i].real() 
						+ rx_buff[i].imag()*rx_buff[i].imag());
				}

				if (avg < 1e-5) {
					//cout << "avg = " << avg << endl;	
				}
				else {
				//	cout << "AVG = " << avg << endl;

	#ifdef DEBUG_DUMP
					int cnt = fwrite(&rx_buff.front(), sizeof(std::complex<float>),
							num_rx_samps, rx_c4fm_200k_file);
					if (cnt != num_rx_samps)
						cerr << "Error writing to rx_c4fm_200k.dat!" << endl;
	#endif
				
					// Apply LPF to I and Q data prior to resampling and demod
					//for (int i = 0; i < num_rx_samps; i++) {
					//	rx_buff[i].real() = fir_seq(fir_200_lpf_I, rx_buff[i].real());
					//	rx_buff[i].imag() = fir_seq(fir_200_lpf_Q, rx_buff[i].imag());
					//}

					// Downsampling from 200KHz to 48KHz
					odata_count = downsample_IQ(rx_buff, rs_I, rs_Q, num_rx_samps);

					if (odata_count != CHUNK_LEN) {
						cout << "odata_count = " << odata_count << ", CHUNK_LEN = " << CHUNK_LEN
							<< endl;
					}

	#ifdef DEBUG_DUMP
					cnt = fwrite(rs_I, sizeof(float), odata_count, rx_c4fm_48k_I_file);
					if (cnt != odata_count)
						cerr << "Error writing to rx_c4fm_48k_I.dat!" << endl;

					cnt = fwrite(rs_Q, sizeof(float), odata_count, rx_c4fm_48k_Q_file);
					if (cnt != odata_count)
						cerr << "Error writing to rx_c4fm_48k_Q.dat!" << endl;
	#endif

	

					// Apply FIR LPF to I and Q data prior to demod
		//			for (int i = 0; i < CHUNK_LEN; i++) {
		//				rs_I[i] = fir_seq(fir_lpf_I, rs_I[i]);
		//				rs_Q[i] = fir_seq(fir_lpf_Q, rs_Q[i]);
		//			}

					// FM demodulation
					fd.fm_demod (rs_I, rs_Q, demod_out, odata_count);

					// Apply FIR LPF to post-demodulated data
					//for (int i = 0; i < CHUNK_LEN; i++) {
					//	demod_out[i] = fir_seq(fir_lpf_demod, demod_out[i]);
					//}
	#ifdef DEBUG_DUMP
					cnt = fwrite(demod_out, sizeof(float), odata_count, rx_c4_48k_file);
					if (cnt != odata_count)
						cerr << "Error writing to rx_c4_48k.dat!" << endl;
	#endif

					// Write to modem socket to send to baseband component
					res = uhd2modem_chan->send((char *)demod_out, odata_count*sizeof(float));
					if (res != (int)(sizeof(float) * odata_count)) {
						cout << res << " of " << (odata_count * sizeof(float)) << "bytes written\n\n";
					}
	/*				
					tempcount += odata_count;
				
					if (tempcount > 8640*3) {
						cout << "tempcount = " << tempcount << endl;
						sleep(100);
					}
	*/
				}
			}

			pthread_mutex_lock(&ptt_mutex);
			tx_swtch = ptt_swtch;
			pthread_mutex_unlock(&ptt_mutex);
		}

		// We've been sent a control signal to switch to transmit, so we've broken out of inner while loop
		// stop receiving and wait on semaphore for receive restart command

		// Stop UHD continuous streaming mode
		uhd_stop_rx_stream();

		sem_wait(&ptt_sem);
		pthread_mutex_lock(&ptt_mutex);
		tx_swtch = ptt_swtch;
		pthread_mutex_unlock(&ptt_mutex);

		fir_reset(fir_lpf_I);
		fir_reset(fir_lpf_Q);

		fd.reset();
	}
}

// Set carrier frequency
void change_cfreq(double cfreq)
{
	set_txfreq(cfreq);
	set_rxfreq(cfreq);
}

// Set TX frequency
void set_txfreq(double cfreq)
{
	if ((uhd_loopback_tx_fifo == "") || (uhd_loopback_rx_fifo == "")) {
		cout << boost::format("Setting TX Freq: %f Mhz...") % (cfreq/1e6) << endl;
		sdev->set_tx_freq(cfreq, 0);
		cout << boost::format("Actual TX Freq: %f Mhz...") % (sdev->get_tx_freq()/1e6)
			<< endl << endl;
	}
}

// Set RX frequency
void set_rxfreq(double cfreq)
{
	if ((uhd_loopback_tx_fifo == "") || (uhd_loopback_rx_fifo == "")) {
		cout << boost::format("Setting RX Freq: %f Mhz...") % (cfreq/1e6) << endl;
		sdev->set_rx_freq(cfreq, 0);
		cout << boost::format("Actual RX Freq: %f Mhz...") % (sdev->get_rx_freq()/1e6)
			<< endl << endl;
	}
}

/* Control signal handler thread responds to control signals
 * received via the control channel, sent by the DSP module.
 */
void *control_sig(void*)
{
	int res;
	unsigned char signal[1];

	while(true) {
		res = uhd_modem_ctrl_chan->receive((char *)signal, sizeof(signal));
		if (res != sizeof(signal)) {
			cerr << "Control Channel Error" << endl;
			continue;
		}
		else 
			cout << "Control Signal received." << endl;

		if (signal[0]==0xBB){
			first_tx = true;

			pthread_mutex_lock(&ptt_mutex);
			ptt_swtch = 1;
			pthread_mutex_unlock(&ptt_mutex);

			cout << "Halting Receiver..\n\n" << endl;
		}
		else if	(signal[0]==0xEE){
			first_rx = true;

			pthread_mutex_lock(&ptt_mutex);
			ptt_swtch = 0;
			pthread_mutex_unlock(&ptt_mutex);
			sem_post(&ptt_sem);

			cout << "Restarting Receiver...\n\n" << endl;
		}
	}

	return NULL;
}

// Thread that receives frequency change packets from the packetizer component
// over a control channel.
void *freq_sig(void*)
{
	int res;
	double cfreq;
	control_pkt ctrl_pkt;

	while(true){
		res = uhd_gpp_ctrl_chan->receive((char *)&ctrl_pkt, sizeof(ctrl_pkt));
		if (res != sizeof(ctrl_pkt)) {
			cerr << "Frequency Change Channel Error:  res = " << res
				<< ", sizeof(ctrl_pkt) = " << sizeof(ctrl_pkt) << endl;
			continue;
		}

		cfreq = 0.0;
		if (ctrl_pkt.signal[0] == 0xFC) {
			cfreq = ctrl_pkt.cfreq;

			if ((cfreq >= 15.0e6) && (cfreq <= 2.7e9)) {
				cout << "Changing carrier frequency to:  " << cfreq / 1e6 << "MHz" << endl;
				change_cfreq(cfreq);
			}
		}
	}

	return NULL;
}
