//-------------------------------------------------------------------------
/**
 * @file 	controller.cpp
 * @brief	Implementation of Controller class functions that control the overall
 * 		Mode, State and Activity of the Radio system.
 * @note	none.
 *
 * @author	Cal(it)2
 * @author  	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * File Name:       main.cpp
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 *		1.0 Jul 27 2009 	Anthony Nwokafor    C Model Structure created
 * 		1.1 Feb 22 2011 	Anthony Nwokafor    C++ Model Structure created
 * @endverbatim
 */
//-------------------------------------------------------------------------

#include <iostream>
#include <pthread.h>
#include <signal.h>
#include "p25/format/bit_fifo.h"
#include "p25/format/bit_manipulate.h"
#include "p25/format/bit_stream.h"
#include "p25/ecc/rs.h"
#include "p25/util/logger.h"
#include "controller.h"
#include "receiver.h"
#include "transmitter.h"
#include "vocoder.h"
#include "ui.h"

pthread_mutex_t rc_a;
pthread_mutex_t rc_s;
pthread_mutex_t rc_m;

//-------------------------------------------------------------------------
/**
 * @details	Initializes the VC55 and GUI communication ports
 * @param[in]	_ui_port	Graphical User Interface communication port
 * @param[in]	_vc_sport	VC55 serial port interface
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
Controller::Controller(int _ui_port, char* _vc_sport, char* _vc_rxfile,
		int _r_mode) :
	ui_port(_ui_port), vc_sport(_vc_sport), vc_rxfile(_vc_rxfile) {
	switch (_r_mode) {
	case LIVE_MODE:
		INFO("Setting radio mode to Mode %d [LIVE_MODE]\n",LIVE_MODE)
		;
		set_radio_mode(_r_mode);
		break;
	case TEST_MODE:
		INFO("Setting radio mode to Mode %d [TEST_MODE]\n",TEST_MODE)
		;
		set_radio_mode(_r_mode);
		break;
	default:
		INFO("Unrecognized Radio mode %d, Switching to mode %d [TEST_MODE]\n",
				_r_mode, TEST_MODE)
		;
		set_radio_mode(TEST_MODE);
	}
}

//-------------------------------------------------------------------------
/**
 * @brief	Gets the current activity of the P25 radio system
 * @details	The P25 radio system can perform either a Transmit activity,
 * 			Receive activity or Standby activity..
 * @return	The current activity of the P25 radio system.
 * @see		TX_ACTIVITY
 * @see		RX_ACTIVITY
 * @see		STANDBY_ACTIVITY
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Controller::get_radio_activity() {
	int res;
	pthread_mutex_lock(&rc_a);
	res = activity;
	pthread_mutex_unlock(&rc_a);
	return res;
}

//-------------------------------------------------------------------------
/**
 * @details	The P25 radio system can be set to perform either a Transmit
 * 			activity, Receive activity or Standby activity.
 * @param[in]	_activity	New activity to be performed by the P25 radio.
 * @see		TX_ACTIVITY
 * @see		RX_ACTIVITY
 * @see		STANDBY_ACTIVITY
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Controller::set_radio_activity(int _activity) {
	pthread_mutex_lock(&rc_a);
	activity = _activity;
	pthread_mutex_unlock(&rc_a);
}

//-------------------------------------------------------------------------
/**
 * @details	The P25 radio system can be in one of two states: Runnning
 * 			state or Shutdown state.
 * @return	The current state of the P25 radio system.
 * @see 	RUNNING_STATE
 * @see		SHUTDOWN_STATE
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Controller::get_radio_state() {
	int res;
	pthread_mutex_lock(&rc_s);
	res = state;
	pthread_mutex_unlock(&rc_s);
	return res;
}

//-------------------------------------------------------------------------
/**
 * @details	The P25 radio system can be set to one of two states: Running
 * 			state or Shutdown state.
 * @param[in]	_state	New state to set the P25 radio to.
 * @see 	RUNNING_STATE
 * @see		SHUTDOWN_STATE
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Controller::set_radio_state(int _state) {
	pthread_mutex_lock(&rc_s);
	state = _state;
	pthread_mutex_unlock(&rc_s);
}

//-------------------------------------------------------------------------
/**
 * @details	The P25 radio system can be in either Live mode or Test mode.
 * @return	The current mode of the P25 radio system.
 * @see 	LIVE_MODE
 * @see 	TEST_MODE
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Controller::get_radio_mode() {
	int res;
	pthread_mutex_lock(&rc_m);
	res = mode;
	pthread_mutex_unlock(&rc_m);
	return res;
}

//-------------------------------------------------------------------------
/**
 * @details	The P25 radio system can be set to either Test mode or Live mode.
 * @param[in]	_mode	New mode to set the P25 radio to.
 * @see 	LIVE_MODE
 * @see 	TEST_MODE
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Controller::set_radio_mode(int _mode) {
	pthread_mutex_lock(&rc_m);
	mode = _mode;
	pthread_mutex_unlock(&rc_m);
}

//-------------------------------------------------------------------------
/**
 * @details	Initializes the Reed Solomon Encoder, sets the initial radio
 * 			state to RUNNING_STATE and the initial radio activity to
 * 			RX_ACTIVITY.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Controller::initialize() {
	// initialize Reed-Solomon encoders
	init_rs();

	pthread_mutex_init(&rc_a, NULL);
	pthread_mutex_init(&rc_s, NULL);
	pthread_mutex_init(&rc_m, NULL);

	set_radio_activity(RX_ACTIVITY);
	set_radio_state(RUNNING_STATE);
}

//-------------------------------------------------------------------------
/**
 * @details		Gracefully Shuts down the Radio System.
 * @param[in]	signum	The signal number
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Controller::shutdown(int signum) {
	/** @todo Change Radio state to SHUTDOWN */
	// set_radio_state(SHUTDOWN_STATE);

	pthread_mutex_destroy(&rc_a);
	pthread_mutex_destroy(&rc_s);
	pthread_mutex_destroy(&rc_m);

	INFO("Exiting Program...\n");
	exit(signum);
}

//-------------------------------------------------------------------------
/**
 * @details	Creates the components of the radio system. Transmitter,
 * 			Receiver, Vocoder, UserInterface. Each component starts its own
 * 			thread except for the Transmitter which is directly controlled
 * 			by the Cotroller.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Controller::start() {
	// setup termination signal interrupt handler
	signal((int) SIGINT, Controller::shutdown);

	initialize();

	INFO("Creating UI object\n");
	UserInterface tUi(this, ui_port);

	INFO("Creating VC object\n");
	Vocoder tVc(this, vc_sport, vc_rxfile);

	INFO("Creating RX object\n");
	Receiver tRx(this, &tVc, &tUi);

	INFO("Creating TX object\n");
	Transmitter tTx(this, &tVc);

	INFO("\nStarting...\n");

	/**
	 * @todo	Change while(true) loop to depend on radio state so the loop
	 * 			will get exited when the radio is shut down.
	 */
	while (get_radio_state() == RUNNING_STATE) {
		// If current radio activity is set to TX, start Transmitting
		if (get_radio_activity() == TX_ACTIVITY) {
			tTx.run();
		} else {
			/** @todo Replace call to usleep with semaphore. This is taking up unnecessary CPU cycles */
			usleep(100);
		}
	}

	// Wait for the other objects to end their threads
	tRx.join();
	tUi.join();
	tVc.join();
}
