//-------------------------------------------------------------------------
/**
 * @file 	receiver.cpp
 * @brief	Implementation of Receiver class functions. The receiver processes
 * 		received bits from the modem and extracts voice data that is played
 * 		back by the Vocoder.
 * @note	none.
 *
 * @author	Cal(it)2
 * @author  	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * File Name:       main.cpp
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 *		1.0 Jul 27 2009 	Anthony Nwokafor    C Model Structure created
 * 		1.1 Feb 22 2011 	Anthony Nwokafor    C++ Model Structure created
 * @endverbatim
 */
//-------------------------------------------------------------------------

#include <iostream>
#include "p25/ecc/golay.h"
#include "p25/ecc/cyclic.h"
#include "p25/ecc/ham.h"
#include "p25/ecc/rs.h"
#include "p25/util/logger.h"
#include "p25/format/bit_manipulate.h"
#include "p25/format/bit_fifo.h"
#include "p25/format/bit_stream.h"
#include "receiver.h"
#include "vocoder.h"
#include "ui.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>

extern UInterface   m_ui;
extern UInterface   mr_ui;

using namespace P25;

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
Receiver::Receiver(Controller* mController, Vocoder* mVocoder, 
    UserInterface*  my_userintf) :
        mController(mController), 
        mVocoder(mVocoder), 
        my_userintf(my_userintf) 
{
    int res;
    pthread_attr_t attr;

    INFO("Loading Receiver\n");
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    res = pthread_create(&rxThread, &attr, Receiver::start_thread, (void*)this);
    if (res) ERROR("Failed to create Receiver thread");
    pthread_attr_destroy(&attr);
	
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Receiver::frameIsMine(void)
{
    /*TODO: Check NAC before proceeding to decode frame*/
    if (m_ui.nac == mr_ui.nac || m_ui.nac == P25_NO_NAC_MODE)
        return 1;

    if (!memcmp(mr_ui.dstId,m_ui.unitId,sizeof(mr_ui.dstId)) ||
        !memcmp(mr_ui.dstId, EVERYONE_UID,sizeof(mr_ui.dstId))){
        if ((mr_ui.tgId == m_ui.tgId) || (mr_ui.tgId == EVERYONE_TGID)){
            return 1;
        }
    }

    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::play_vcw(char* vcw, int len)
{
    if (frameIsMine()) {
        mVocoder->put_in_pb_buf(vcw,len);
    } else {
        printf("Frame is not mine!!\n");
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Receiver::nid_dec(char* nid_cw, char* duid)
{
    mr_ui.nac = bi2de(nid_cw,12);
    *duid = bi2de(nid_cw+12,4);

    //r_ui_update();
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::lsd_dec(char* lsd_cw, char* lsd)
{
    int i;
    char decw;

    // get and decode lsd
    for (i = 0; i < 2; i++) {
        decw = cyclic_16_8_decode(bi2de(lsd_cw+(i*16),16));
        lsd[i] = decw & 0xff;
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::lc_deformat(struct bit_fifo *lc)
{
    char bits[32];
    int lco, p, sf;
    unsigned int i;

    bf_get_n(lc,bits,1);
    p = bi2de(bits,1);
    if (p == 0) {
        bf_get_n(lc,bits,1);
        sf = bi2de(bits,1);

        bf_get_n(lc,bits,6);
        lco = bi2de(bits,6);

        if (lco == LC_GRP_V_CH_USR) {
            // MFID
            bf_get_n(lc,bits,8);
            mr_ui.mfId = bi2de(bits,8);

            // Service Options
            bf_get_n(lc,bits,8);

            // Reserved
            bf_get_n(lc,bits,8);

            // Group Address
            bf_get_n(lc,bits,16);
            mr_ui.tgId = (short)bi2de(bits,16);

            // Source Address
            for (i = 0; i < sizeof(mr_ui.unitId); i++) {
                bf_get_n(lc,bits,8);
                mr_ui.unitId[i] = bi2de(bits,8);
            }
        } else if (lco == LC_UU_V_CH_USR) {
            // MFID
            bf_get_n(lc,bits,8);
            mr_ui.mfId = bi2de(bits,8);

            // Service Options
            bf_get_n(lc,bits,8);

            // Target Address
            for (i = 0; i < sizeof(mr_ui.dstId); i++) {
                bf_get_n(lc,bits,8);
                mr_ui.dstId[i] = bi2de(bits,8);
            }

            // Source Address
            for (i = 0; i < sizeof(mr_ui.unitId); i++) {
                bf_get_n(lc,bits,8);
                mr_ui.unitId[i] = bi2de(bits,8);
            }
        } else if (lco == LC_TELE_INT_V_CH_USR) {

        } else if (lco == LC_SOURCE_ID_EXT) {

        } else if (lco == LC_UU_V_CH_USR_EXT) {

        } else if (lco == LC_CALL_TRM_CAN) {

        } else if (lco == LC_STS_Q) {

        } else if (lco == LC_STA_UPDT) {

        } else if (lco == LC_MSG_UPDT) {

        } else if (lco == LC_CALL_ALRT) {

        } else if (lco == LC_EXT_FNCT_CMD) {

        } else if (lco == LC_CH_ID_UPDT) {

        } else {

        }
    }

    //r_ui_update();
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Receiver::lcw_dec(char* lcw_cw)
{
    int i,decw,eras,eras_count,temp;
    struct bit_fifo *bf;
    char dec_bits[24];
    int   dec_eras[24];
    char bits[32];

    bf = bf_alloc(HEXBIT*12);

    // get bits and send to hamming decoder
    eras_count = 0;
    for (i = 0; i < 24; i++) {
        eras = hamming_10_decode(bi2de(lcw_cw+(i*10),10), &decw);
        dec_bits[i] = decw & 0x3f;
        dec_eras[i] = eras;
        if (eras == 1) eras_count++;
    }

    reverse_bits(dec_bits,24);
    for (i = 0; i < 24/2; i++) {
        temp = dec_eras[i];
        dec_eras[i] = dec_eras[24-i-1];
        dec_eras[24-i-1] = temp;
    }
    decode_rs_24_12(dec_bits,dec_eras,eras_count);
    reverse_bits(dec_bits,24); // we only need the first 12 bits

    for (i = 0; i < 12; i++) {
        de2bi(dec_bits[i],bits,HEXBIT);
        bf_put_n(bf,bits,HEXBIT);
    }

    lc_deformat(bf);

    bf_free(bf);
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Receiver::esw_dec(char* esw_cw)
{
    int decw,eras,eras_count,temp;
    unsigned int i;
    struct bit_fifo *bf;
    char dec_bits[24];
    int   dec_eras[24];
    char bits[32];

    bf = bf_alloc(HEXBIT*16);

    // get bits and send to hamming decoder
    eras_count = 0;
    for (i = 0; i < 24; i++) {
        eras = hamming_10_decode(bi2de(esw_cw+(i*10),10), &decw);
        dec_bits[i] = decw & 0x3f;
        dec_eras[i] = eras;
        if (eras == 1) eras_count++;
    }

    reverse_bits(dec_bits,24);
    for (i = 0; i < 24/2; i++) {
        temp = dec_eras[i];
        dec_eras[i] = dec_eras[24-i-1];
        dec_eras[24-i-1] = temp;
    }
    decode_rs_24_16(dec_bits,dec_eras,eras_count);
    reverse_bits(dec_bits,24); // we only need the first 16 bits

    for (i = 0; i < 16; i++) {
        de2bi(dec_bits[i],bits,HEXBIT);
        bf_put_n(bf,bits,HEXBIT);
    }

    for (i = 0; i < sizeof(mr_ui.mi); i++) {
        bf_get_n(bf,bits,8);
        mr_ui.mi[i] = bi2de(bits,8);
    }

	bf_get_n(bf,bits,8);
	mr_ui.algId = bi2de(bits,8);

	bf_get_n(bf,bits,16);
	mr_ui.keyId = bi2de(bits,16);

    //r_ui_update();

    bf_free(bf);
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Receiver::hdu_dec(char* hdu)
{
    int eras,eras_count,temp;
    unsigned int i;
    long decw;
    struct bit_fifo *bf;
    char dec_bits[36];
    int   dec_eras[36];
    char bits[32];

    bf = bf_alloc(HEXBIT*20);

    // get bits and send to hamming decoder
    eras_count = 0;
    for (i = 0; i < 36; i++) {
        decw = bi2de(hdu+(i*18),18);
        eras = golay_18_6_decode(&decw);
        dec_bits[i] = decw & 0x3f;
        dec_eras[i] = eras;
        if (eras == 1) eras_count++;
    }

    reverse_bits(dec_bits,36);
    for (i = 0; i < 36/2; i++) {
        temp = dec_eras[i];
        dec_eras[i] = dec_eras[36 -i-1];
        dec_eras[36-i-1] = temp;
    }
    decode_rs_36_20(dec_bits,dec_eras,eras_count);
    reverse_bits(dec_bits,36); // reverse all 36 bits but we only use 20 bits

    for (i = 0; i < 20; i++) {
        de2bi(dec_bits[i],bits,HEXBIT);
        bf_put_n(bf,bits,HEXBIT);
    }

    for (i = 0; i < sizeof(mr_ui.mi); i++) {
        bf_get_n(bf,bits,8);
        mr_ui.mi[i] = bi2de(bits,8);
    }

    bf_get_n(bf,bits,8);
    mr_ui.mfId = bi2de(bits,8);

    bf_get_n(bf,bits,8);
    mr_ui.algId = bi2de(bits,8);

    bf_get_n(bf,bits,16);
    mr_ui.keyId = bi2de(bits,16);

    bf_get_n(bf,bits,16);
    mr_ui.tgId = bi2de(bits,16);

    mr_ui.ptt = 1;

    bf_free(bf);
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::ldu_cw_split(char* ldu, int len, char* vc, char* lsd, char* lc)
{
    int i,vc_offset;
    struct bit_fifo *bf;
    
    bf = bf_alloc(len);
    bf_put_n(bf,ldu,len);

    bf_get_n(bf,vc,VC_BLEN*2);

    for (i = 0; i < (LCW_BLEN/40); i++) {
        vc_offset = (VC_BLEN*2)+(i*VC_BLEN);
        bf_get_n(bf,lc+(i*40),40);
        bf_get_n(bf,vc+vc_offset,VC_BLEN);
    }

    bf_get_n(bf,lsd,LSD_BLEN);
    bf_get_n(bf,vc+vc_offset+VC_BLEN,VC_BLEN);
    bf_free(bf);
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::ldu_dec(char* ldu, char type)
{
    int len,res;
    char vc[VC_BLEN*9];
    char lsd[2];
    char lsd_cw[LSD_BLEN];
    char ldu_ci[LCW_BLEN]; // LDU control Information (LCW or ESW)

    len = LDU_VC_BLEN;
    ldu_cw_split(ldu,len,vc,lsd_cw,ldu_ci);

    switch (type) {
        case DUID_LDU1:
            res = lcw_dec(ldu_ci);
            if (res < 0)
                ERROR("LCW decode failed at frame %d: %d\n", frame_id, res);
    	    break;
        case DUID_LDU2:
    	    res = esw_dec(ldu_ci);
            if (res < 0)
                ERROR("ESW decode failed at frame %d: %d\n", frame_id, res);
    	    break;
    }

    play_vcw(vc,sizeof(vc));
    lsd_dec(lsd_cw,lsd);
	
    mr_ui.ptt = 1;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::tdu_dec(char* tdu)
{
	mr_ui.ptt = 0;
	my_userintf->r_ui_update();
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Receiver::tdu_lcw_dec(char* tdu)
{
    int i,j,eras,eras_count,temp;
    long decw;
    struct bit_fifo *bf;
    char dec_bits[24];
    int dec_eras[24];
    char bits[32];

    bf = bf_alloc(HEXBIT*12);

    // get bits and send to hamming decoder
    eras_count = 0;
    j = 0;
    for (i = 0; i < 12; i++) {
        decw = bi2de(tdu+(i*24),24);
        eras = golay_24_12_decode(&decw);
        dec_bits[j] = (decw >> HEXBIT) & 0x3f;
        dec_bits[j+1] = decw  & 0x3f;
        dec_eras[j] = eras;
        dec_eras[j+1] = eras;
        j += 2;
        if (eras == 1) eras_count += 2;
    }

    reverse_bits(dec_bits,24);
    for (i = 0; i < 24/2; i++) {
        temp = dec_eras[i];
        dec_eras[i] = dec_eras[24-i-1];
        dec_eras[24-i-1] = temp;
    }
    decode_rs_24_12(dec_bits,dec_eras,eras_count);
    reverse_bits(dec_bits,24); //reverse all 24 bits but we use only the first 12

    for (i = 0; i < 12; i++) {
        de2bi(dec_bits[i],bits,HEXBIT);
        bf_put_n(bf,bits,HEXBIT);
    }

    lc_deformat(bf);

	mr_ui.ptt = 0;
	my_userintf->r_ui_update();

    bf_free(bf);
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 * @param[in] bits Buffer containing the FS bits to check
 * @return 	Integer value representing success or failure of the FS check
 * @retval 	The value returned is zero or greater (positive) when the number
 * 			of bit differences in the received frame sync is less than the
 * 			FS_ERROR_THRESHOLD and is less than zero (negative) otherwise.
 * @see 	FS_ERROR_THRESHOLD
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Receiver::check_fs(char* bits)
{
    int threshold = FS_ERROR_THRESHOLD;
    int i,sum;

    sum = 0;
    for (i = 0; i < FS_BLEN; i++) {
    	// Sum up all the differences between received FS bits
    	// and expected FSbits
        sum += abs(fs[i] - bits[i]);
    }

    // a
    return threshold - sum;
}

//-------------------------------------------------------------------------
/**
 * @details Gets the specified number of bits from the receiver bit stream
 * 			with the status bits extracted. rx_bit_count tracks the number
 * 			of data bits read since Status bits were last encountered.
 * @param[in] bits The buffer to store the requested bits.
 * @param[in] len The number of bits to read.
 * @return The number of bits read.
 * @see rx_bit_count
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int Receiver::get_bits(char* bits, int len)
{
    int res;
    int bits_read; // Count of total bits read so far
    int bits_to_get;
    char ss[2];

    bits_read = 0;
    while(bits_read < len) {

    	// Calculate number of bits to read to make sure that
    	// Status bits are not included with the data
        bits_to_get = 70 - rx_bit_count;
        if (bits_to_get > (len - bits_read)) 
            bits_to_get = len - bits_read;

	    res = bs_input(bits+bits_read, bits_to_get);
	    bits_read += res;
	    rx_bit_count += res;

	    // Extract Status Bits from the stream.
	    // Currently the status bits are discarded.
        if (rx_bit_count == 70) {
            rx_bit_count = 0;
            bs_input(ss,SS_BLEN);
        }
    }

    return bits_read;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::initialize()
{
    rx_bit_count = 0;
    frame_id = 0;
    memset((char* )&mr_ui,0,sizeof(mr_ui));
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::shutdown()
{
}

//-------------------------------------------------------------------------
/**
 * @details The main receiver process. It decodes one frame at a time by
 * 			checking for a valid Frame Sync. If a valid frame sync is found
 * 			it goes on to decode the Network Identifier (NID) which contains
 * 			the Data Unit ID (DUID) that identifies the frame. Upon successful
 * 			decoding and extraction of the DUID, the appropriate decode
 * 			function for the frame type is called to decode the frame. At
 * 			the end of a full frame decode, the UInterface is updated with
 * 			parameters extracted from the frame.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::run_rx()
{
    int     res;
    char    duid;
    char    bits[MAX_FRAME_SIZE];

    /**
     * @todo: 	Change while(true) to depend on radio state so this thread
     * 			can be shutdown.
     */
    // While radio is in the running state
    while (mController->get_radio_state() == RUNNING_STATE) {
    	// and the current radio activity is receive activity
        if (mController->get_radio_activity() == RX_ACTIVITY) {
        	// Initialize the Receiver State variables
            initialize();
            INFO("Receiver started....\n");

            while((mController->get_radio_activity() == RX_ACTIVITY)) {
                frame_id++;

                res = get_bits(bits, FS_BLEN);
                if (res < FS_BLEN) {
                    ERROR("Invalid FS Length read!!\n");
                    continue;
                }

                //INFO("Checking FS\n");
                res = check_fs(bits);
                if (res < 0 ) {
                    ERROR("check FS failed at frame %d: %d\n", frame_id, res);
                    continue;
                }

                res = get_bits(bits, NID_BLEN);
                if (res < NID_BLEN) {
                    ERROR("Invalid NID Length read!!\n");
                    continue;
                }

                //INFO("Decoding NID\n");
                res = nid_dec(bits, &duid);
                if (res < 0) {
                    ERROR("decode NID failed at frame %d: %d\n", frame_id, res);
                    continue;
                }

                switch (duid) {
                    case DUID_HDU:
                        //INFO("Decoding HDU\n");
                        res = get_bits(bits, HCW_BLEN+HDU_PAD_BLEN);
                        if (res == HCW_BLEN+HDU_PAD_BLEN) {
                            res = hdu_dec(bits);
                            if (res < 0) {
                                ERROR("HDU decode failed at frame %d: %d\n", 
                                    frame_id, res);
                            }
                        } else {
                            ERROR("Invalid HDU Length read!!\n");
                        }
                        break;

                    case DUID_TDU:
                        //INFO("Decoding TDU\n");
                        res = get_bits(bits, TDU_PAD_BLEN);
                        if (res == TDU_PAD_BLEN) {
                            tdu_dec(bits);
                        }
                        break;

                    case DUID_TDU_LCW:
                        //INFO("Decoding TDU_LCW\n");
                        res = get_bits(bits, TDU_LC_BLEN+TDU_LCW_PAD_BLEN);
                        if (res == TDU_LC_BLEN+TDU_LCW_PAD_BLEN) {
                            res = tdu_lcw_dec(bits);
                            if (res < 0) {
                                ERROR("TDU_LCW decode failed at frame %d: %d\n", 
                                frame_id, res);
                            }
                        } else {
                            ERROR("Invalid TDU_LCW Length read!!\n");
                        }
                        break;

                    case DUID_LDU1:
                        //INFO("Decoding LDU1\n");
                        res = get_bits(bits, LDU_VC_BLEN);
                        if (res == LDU_VC_BLEN) {
                            ldu_dec(bits, duid);
                        } else {
                            ERROR("length not equal at frame %d\n", frame_id);
                        }
                        break;

                    case DUID_LDU2:
                        //INFO("Decoding LDU2\n");
                        res = get_bits(bits, LDU_VC_BLEN);
                        if (res == LDU_VC_BLEN) {
                            ldu_dec(bits, duid);
                        } else {
                            ERROR("length not equal at frame %d\n", frame_id);
                        }	
                        break;

                    case DUID_PDU:
                        //INFO("PDU duid not supported yet at frame %d\n", duid);
                        break;

                    case DUID_TSB:
                        //INFO("TSB duid not supported yet at frame %d\n", duid);
                        break;

                    default:
                        INFO("duid %d error at frame %d\n", duid, frame_id);
                        break;
                }

                // update user interface object
                my_userintf->r_ui_update();
            }
            shutdown();
            INFO("Finished Receiving....\n");
        } else { 
			/** @todo Replace with semaphore. This could be using unnecessary CPU cycles. */
            usleep(10);
        }
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Receiver::join()
{
    pthread_join(rxThread, NULL);
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void* Receiver::start_thread(void *obj)
{
    static_cast<Receiver *>(obj)->run_rx();
    return 0;
}

