//-------------------------------------------------------------------------
/**
 * @file 	transmitter.cpp
 * @brief	Implementation of Transmitter class functions to encode and
 * 			transmit P25 frames.
 * @note	none.
 *
 * @author	Cal(it)2
 * @author  	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * File Name:       main.cpp
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 *		1.0 Jul 27 2009 	Anthony Nwokafor    C Model Structure created
 * 		2.0 Feb 22 2011 	Anthony Nwokafor    C++ Model Structure created
 * 		2.1 Mar 31 2011		Anthony Nwokafor	Documentation started
 * @endverbatim
 */
//-------------------------------------------------------------------------

#include <iostream>
#include <pthread.h>
#include "p25/ecc/golay.h"
#include "p25/ecc/cyclic.h"
#include "p25/ecc/ham.h"
#include "p25/ecc/rs.h"
#include "p25/format/bit_fifo.h"
#include "p25/format/bit_stream.h"
#include "p25/format/bit_manipulate.h"
#include "p25/util/logger.h"
#include "transmitter.h"
#include "vocoder.h"

#include <string.h>
#include <stdlib.h>


/**
 * @details Global variable shared with UI class for synchronization of User
 * 			Interface changes.
 */
extern UInterface m_ui;

using namespace P25;

//-------------------------------------------------------------------------
/**
 * @details Transmitter class constructor
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
Transmitter::Transmitter(Controller* mController, Vocoder* mVocoder)
: mController(mController), mVocoder(mVocoder) {

}

//-------------------------------------------------------------------------
/**
 * @details	Encodes some constants based on current configuration data
 * 			in order to speed up P25 frame encoding
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::enc_speedup()
{
	/**
	 * @todo Switch TGID to 0x0000 for Unit to Unit transmission
	 * @date 09/15/2009
	 */
    nid_enc(ldu1_nid,DUID_LDU1,1);
    nid_enc(ldu2_nid,DUID_LDU2,1);

    if (memcmp(m_ui.dstId, EVERYONE_UID, sizeof(m_ui.dstId)) == 0) {
        lcw_enc(LC_GRP_V_CH_USR);
    } else {
        lcw_enc(LC_UU_V_CH_USR);
    }

    esw_enc();
}

//-------------------------------------------------------------------------
/**
 * @details	Formats the Link Control Word
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::lc_format(struct bit_fifo* lc, int lco)
{
    char bits[32];
    unsigned int i;

    switch (lco) {
    case LC_GRP_V_CH_USR:
        // LCO = 0, SF = 0
		de2bi(0x00,bits,8);
		bf_put_n(lc,bits,8);
        // MFID
		de2bi(m_ui.mfId,bits,8);
		bf_put_n(lc,bits,8);
        // Service Options
		de2bi(0x0,bits,8);
		bf_put_n(lc,bits,8);
        // Reserved
		de2bi(0x0,bits,8);
		bf_put_n(lc,bits,8);
        // Group Address
		de2bi(m_ui.tgId,bits,16);
		bf_put_n(lc,bits,16);
        // Source Address
		for (i = 0; i < sizeof(m_ui.unitId); i++) {
			de2bi(m_ui.unitId[i],bits,8);
			bf_put_n(lc,bits,8);
		}
		break;
    case LC_UU_V_CH_USR:
        // LCO = 3, SF = 0
		de2bi(0x03,bits,8);
		bf_put_n(lc,bits,8);
        // MFID
		de2bi(m_ui.mfId,bits,8);
		bf_put_n(lc,bits,8);
        // Service Options
		de2bi(0x0,bits,8);
		bf_put_n(lc,bits,8);
        // Target Address
		for (i = 0; i < sizeof(m_ui.dstId); i++) {
			de2bi(m_ui.dstId[i],bits,8);
			bf_put_n(lc,bits,8);
		}
        // Source Address
		for (i = 0; i < sizeof(m_ui.unitId); i++) {
			de2bi(m_ui.unitId[i],bits,8);
			bf_put_n(lc,bits,8);
		}
		break;
    case LC_TELE_INT_V_CH_USR:
    	break;
    case LC_SOURCE_ID_EXT:
    	break;
    case LC_UU_V_CH_USR_EXT:
    	break;
    case LC_CALL_TRM_CAN:
    	break;
    case LC_STS_Q:
    	break;
    case LC_STA_UPDT:
    	break;
    case LC_MSG_UPDT:
    	break;
    case LC_CALL_ALRT:
    	break;
    case LC_EXT_FNCT_CMD:
    	break;
    case LC_CH_ID_UPDT:
    	break;
    default:
    	break;
    }
}

//-------------------------------------------------------------------------
/**
 * @details	Encodes the Link Control Word
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::lcw_enc(int lco)
{
    unsigned int i;
    long encw;
    char bits[32];
    struct bit_fifo *lc;
    char enc_bits[24];

    lc = bf_alloc(LCW_BLEN);

    lc_format(lc,lco);

	// get bits and send to rs(24,12,13) encoder
    bf_get_n(lc,lcw,72);
    for (i = 0; i < 12; i++) {
        bits[i] = bi2de(lcw+(i*6),6);
    }

    reverse_bits(bits,12);
    encode_rs_24_12(bits,enc_bits);
    reverse_bits(enc_bits,24);

    for (i = 0; i < 24; i++) {
        de2bi(enc_bits[i],lcw+(i*6),6);
    }

    // get rs and send to hamming encoder
    for (i = 0; i < 24; i++) {
        encw = hamming_10_encode(bi2de(lcw+(i*6),6));
        de2bi(encw & 0x3ff, bits,10);
        bf_put_n(lc,bits,10);
    }
    bf_get_n(lc,lcw,LCW_BLEN);
    bf_free(lc);
}

//-------------------------------------------------------------------------
/**
 * @details	Encodes the Encryption Sync Word
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::esw_enc()
{
    unsigned int i;
    long encw;
    char bits[32];
    struct bit_fifo *es;
    char enc_bits[24];
    
    es = bf_alloc(ESW_BLEN);

    for (i = 0; i < sizeof(m_ui.mi); i++) {
        de2bi(m_ui.mi[i],bits,8);
        bf_put_n(es,bits,8);
    }

    de2bi(m_ui.algId,bits,8);
    bf_put_n(es,bits,8);

    de2bi(m_ui.keyId,bits,16);
    bf_put_n(es,bits,16);

    // get bits and send to rs(24,16,9) encoder
    bf_get_n(es,esw,96);
    for (i = 0; i < 16; i++) {
        bits[i] = bi2de(esw+(i*6),6);
    }

    reverse_bits(bits,16);
    encode_rs_24_16(bits,enc_bits);
    reverse_bits(enc_bits,24);

    for (i = 0; i < 24; i++) {
        de2bi(enc_bits[i],esw+(i*6),6);
    }

    // get rs and send to hamming encoder
    for (i = 0; i < 24; i++) {
        encw = hamming_10_encode(bi2de(esw+(i*6),6));
        de2bi(encw & 0x3ff, bits,10);
        bf_put_n(es,bits,10);
    }
    bf_get_n(es,esw,ESW_BLEN);
    bf_free(es);
}

//-------------------------------------------------------------------------
/**
 * @details	Encodes the Network ID (NID)
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::nid_enc(char* nid, char duid, char par_bit)
{
    struct bit_fifo *ni;

    ni = bf_alloc(NID_BLEN);
	memset(nid,0,63);

    de2bi(m_ui.nac,nid,12);
    bf_put_n(ni,nid,12);

    de2bi(duid,nid,4);
    bf_put_n(ni,nid,4);

    // get bits and send to bch encoder
    bf_get_n(ni,nid,16);
    bf_free(ni);

    nid[63] = par_bit;
}

//-------------------------------------------------------------------------
/**
 * @details	Encodes the Header Data Unit (HDU)
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::hdu_enc()
{
    unsigned int i;
    long encw;
    struct bit_fifo *hc;
    char enc_bits[36];
    char bits[32];
    char nid[NID_BLEN];
    char hdu[HDU_BLEN];

    hc = bf_alloc(HDU_BLEN);

	// send FrameSync
    send_to_bitstream((char *)fs,FS_BLEN);
    // send nid
    nid_enc(nid,DUID_HDU,0);
    send_to_bitstream(nid,NID_BLEN);

    for (i = 0; i < sizeof(m_ui.mi); i++) {
        de2bi(m_ui.mi[i],bits,8);
        bf_put_n(hc,bits,8);
    }
    de2bi(m_ui.mfId,bits,8);
    bf_put_n(hc,bits,8);

    de2bi(m_ui.algId,bits,8);
    bf_put_n(hc,bits,8);

    de2bi(m_ui.keyId,bits,16);
    bf_put_n(hc,bits,16);

    de2bi(m_ui.tgId,bits,16);
    bf_put_n(hc,bits,16);

    // get bits and send to rs encoder
    bf_get_n(hc,hdu,120);
    bf_free(hc);

    for (i = 0; i < 20; i++) {
        bits[i] = bi2de(hdu+(i*6),6);
    }

    reverse_bits(bits,20);
    encode_rs_36_20(bits,enc_bits);
    reverse_bits(enc_bits,36);

    for (i = 0; i < 36; i++) {
        de2bi(enc_bits[i],hdu+(i*6),6);
    }

    // get rs and send to golay encoder
    for (i = 0; i < 36; i++) {
        encw = golay_18_6_encode(bi2de(hdu+(i*6),6));
        de2bi(((encw >> 16) & 0x03),bits,2);
        send_to_bitstream(bits,2);
        de2bi(encw & 0xffff, bits,16);
        send_to_bitstream(bits,16);
    }

    memset(bits,0,HDU_PAD_BLEN);
    send_to_bitstream(bits,HDU_PAD_BLEN);
}

//-------------------------------------------------------------------------
/**
 * @details		Encodes the Logical Data Unit (both 1 and 2)
 * @param type	Type of LDU (1 or 2)
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::ldu_enc(char type)
{
    unsigned int i;
    long encw;
    char bits[32];
    char lsd[2];
    char vcw[VC_BLEN];
    char* ldu_nid;
    char* cntrlW;

    switch (type) {
    case DUID_LDU1:
    	ldu_nid = ldu1_nid;
    	cntrlW = lcw;
        lsd[0] = 0xDE;
        lsd[1] = 0xAD;
    	break;
    case DUID_LDU2:
    	ldu_nid = ldu2_nid;
    	cntrlW = esw;
        lsd[0] = 0xBE;
        lsd[1] = 0xEF;
    	break;
    default:
    	ERROR("%d is not a valid LDU type", type);
    	return;
    }

    // send FrameSync
    send_to_bitstream((char *)fs,FS_BLEN);
    // send nid
    send_to_bitstream(ldu_nid,NID_BLEN);
	// send first vcw
	get_vcw(vcw);
	send_to_bitstream(vcw,VC_BLEN);

    //send_to_bitstream(vcw,VC_BLEN);
    for (i = 0; i < 6; i++) {
		get_vcw(vcw);
		send_to_bitstream(vcw,VC_BLEN);
        send_to_bitstream(cntrlW+(i*40),40);
    }
	get_vcw(vcw);
    send_to_bitstream(vcw,VC_BLEN);

    //encode lsd
    for (i = 0; i < 2; i++) {
        encw = cyclic_16_8_encode(lsd[i]);
        de2bi(encw & 0xffff,bits,16);
        send_to_bitstream(bits,16);
    }

	get_vcw(vcw);
    send_to_bitstream(vcw,VC_BLEN);
}

//-------------------------------------------------------------------------
/**
 * @details		Encodes the Terminator Data Unit (TDU)
 * @param type	Type of TDU ("Short TDU" or "Long TDU")
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::tdu_enc(char type)
{
    unsigned int i;
    long encw;
    struct bit_fifo *tc;
    char enc_bits[24];
    char bits[32];
    char nid[NID_BLEN];
    char tdu[TDU_LCW_BLEN];
    

    // send FrameSync
    send_to_bitstream((char *)fs,FS_BLEN);

    switch (type) {
    case DUID_TDU:
        // send nid
		nid_enc(nid,DUID_TDU,0);
        send_to_bitstream(nid,NID_BLEN);

		memset(bits,0,TDU_PAD_BLEN);
		send_to_bitstream(bits,TDU_PAD_BLEN);
		break;
    case DUID_TDU_LCW:
        tc = bf_alloc(TDU_LCW_BLEN);

        // send nid
		nid_enc(nid,DUID_TDU_LCW,0);
        send_to_bitstream(nid,NID_BLEN);

		de2bi(0x03,bits,8);
		bf_put_n(tc,bits,8);

		de2bi(m_ui.mfId,bits,8);
		bf_put_n(tc,bits,8);

		de2bi(0x0,bits,8);
		bf_put_n(tc,bits,8);

		for (i = 0; i < sizeof(m_ui.dstId); i++) {
			de2bi(m_ui.dstId[i],bits,8);
			bf_put_n(tc,bits,8);
		}

		for (i = 0; i < sizeof(m_ui.unitId); i++) {
			de2bi(m_ui.unitId[i],bits,8);
			bf_put_n(tc,bits,8);
		}

		// get bits and send to rs(24,12,13) encoder
		bf_get_n(tc,tdu,72);
        bf_free(tc);

		for (i = 0; i < 12; i++) {
			bits[i] = bi2de(tdu+(i*6),6);
		}

		reverse_bits(bits,12);
		encode_rs_24_12(bits,enc_bits);
		reverse_bits(enc_bits,24);
                
		for (i = 0; i < 24; i++) {
			de2bi(enc_bits[i],tdu+(i*6),6);
		}

		// get rs and send to golay encoder
		for (i = 0; i < 12; i++) {
			encw = golay_24_12_encode(bi2de(tdu+(i*12),12));
			de2bi(((encw >> 16) & 0xff),bits,8);
			send_to_bitstream(bits,8);
			de2bi(encw & 0xffff, bits,16);
			send_to_bitstream(bits,16);
		}

		memset(bits,0,TDU_LCW_PAD_BLEN);
		send_to_bitstream(bits,TDU_LCW_PAD_BLEN);
                INFO("Transmitted TDU\n");
        break;
    }
}

//-------------------------------------------------------------------------
/**
 * @details	Prepares Transmitter class global variables (status symbol,
 * 			bit counter, session-wide constants based on current config.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::initialize()
{
    // Set status symbols
    stat_sym[0] = 0;
    stat_sym[1] = 0;
    bitCount = 0;
    enc_speedup();
}

//-------------------------------------------------------------------------
/**
 * @details	Flushes bit stream output in preparation for next TX
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::shutdown()
{
    bs_output_flush();
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::get_vcw(char* vcw)
{
    int res = 0;

    while(res != VC_BLEN && (mController->get_radio_activity() == TX_ACTIVITY)) {
        res = mVocoder->receive_vc_data(vcw);
    }

	if (res != VC_BLEN) {
		// Radio Mode has changed to RX set Voice Code Words to 0
    	memset(vcw,0,VC_BLEN);
	}

    //INFO("Got VCW\n");
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::send_to_bitstream(char* bits, int len)
{
    int ret, numtoSend, offset;

    offset = 0;
    while ( len ) {
        if ((len + bitCount) > 70 ) {
            numtoSend = 70 - bitCount;
        } else {
            numtoSend = len;
        }

        ret = bs_output(bits+offset,numtoSend);
        if (ret < 0) {
            INFO("Couldn't send to DSP\n");
            break;
        }
        bitCount += ret;
        len -= ret;
        offset += ret;

        if (bitCount == 70) {
            ret = bs_output(stat_sym,SS_BLEN);
            bitCount = 0;
        }
    }
}

//-------------------------------------------------------------------------
/**
 * @details Starts the frame transmission process and continues running
 * 			while the radio activity is still TX activity. The HDU is encoded
 * 			and put in the bit stream followed by successive pairs of
 * 			encoded LDU1 and LDU2 frames. Once the radio activity is no
 * 			longer TX activity a TDU frame is encoded and put in the bit
 * 			stream.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void Transmitter::run()
{
	// Check current radio activity
    if(mController->get_radio_activity() == TX_ACTIVITY) {
        INFO("Transmitter started....\n");
        initialize();
        hdu_enc();

        while(mController->get_radio_activity() == TX_ACTIVITY) {
            ldu_enc(DUID_LDU1);
            /** @todo May want to check if radio activity is still TX
             * 	here before continuing to send LDU2
             */
            ldu_enc(DUID_LDU2);
        }

        tdu_enc(DUID_TDU_LCW);
        shutdown();
        INFO("Transmitter Ended....\n");
    }
}
