//-------------------------------------------------------------------------
/**
 * @file 	transmitter.h
 * @brief	Transmitter Class Definition
 *
 * @author	Cal(it)2
 * @author  Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
//-------------------------------------------------------------------------

#ifndef _TRANSMITTER_H
#define _TRANSMITTER_H

#include "p25/config.h"
#include "controller.h"
#include "vocoder.h"

class Transmitter
{
private:
	/** @brief	Pre-encoded NID for LDU1 frames.*/
	char	ldu1_nid[NID_BLEN];

	/** @brief	Pre-encoded NID for LDU2 frames.*/
	char	ldu2_nid[NID_BLEN];

	/** @brief	Pre-encoded link control bytes for LDU1 frames.*/
	char	lcw[LCW_BLEN];

	/** @brief	Pre-encoded encryption sync bytes for LDU1 frames.*/
	char	esw[ESW_BLEN];

	/** @brief	Holds status symbols.*/
	char	stat_sym[SS_BLEN];
	
	/** @brief	Counts the number of bits put on the outgoing channel.*/
	int		bitCount;

	/** @brief	Path to UNIX domain socket for communication with modem.*/
	int		tx_sock;
	
	/** @brief	Pointer to Controller object that created instance of this class. */
	Controller*		mController;

	/** @brief	Pointer to Vocoder instance created by Controller.*/
	Vocoder*		mVocoder;
	
public:
	/** @brief Constructor for Transmitter Class. */
	Transmitter(Controller* _mController, Vocoder* _mVocoder);

	/** @brief Encodes reused P25 frame components. */
	void	enc_speedup();

	/**
	 * @brief 	Prepares the parts of the Link Control for the specified
	 * 			Link Control Option (LCO).
	 */
	void	lc_format(struct bit_fifo* lc, int lco);

	/** @brief Get one Voice Code Word (VCW) from the Vocoder. */
	void	get_vcw(char* vcw);

	/** @brief Puts the specified bits in the bit stream. */
	void	send_to_bitstream(char* bits, int len);

	/**
	 * @brief 	Encodes the Link Control Word (LCW) for the specified Link
	 * 			Control Option.
	 */
	void	lcw_enc(int lco);

	/** @brief Encodes the Encryption Synchronization Word. */
	void	esw_enc(void);

	/** @brief Encodes the Network Identifier (NID). */
	void	nid_enc(char* nid, char duid, char par_bit);

	/** @brief Encodes the Header Data Unit (HDU) frame. */
	void	hdu_enc();

	/**
	 * @brief 	Encodes the Logical Data Unit (LDU) frames which carry
	 * 			the voice codewords.
	 */
	void	ldu_enc(char type);

	/** @brief Encodes the Termination Data Unit (TDU) frame. */
	void	tdu_enc(char type);

	/** @brief Prepares the Transmitter for transmission. */
	void	initialize();

	/** @brief Begin encoding and transmitting P25 frames. */
	void	run();

	/** @brief Gracefully ends transmissions and shuts down the Transmitter. */
	void	shutdown();

};


#endif
