//-------------------------------------------------------------------------
/**
 * @file        ui.cpp
 * @brief       UserInterface Class Implementation that enables control of
 *				the P25 waveform from an external GUI.
 *
 * @author      Cal(it)2
 * @author      Anthony Nwokafor <aanwokafor@ucsd.edu>
 * @author      Jeffrey Cuenco <jcuenco@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 *              Rev Date                Author              Description of Change
 * @endverbatim
 */
//-------------------------------------------------------------------------

#include <iostream>
#include <netinet/in.h>
#include "ui.h"
#include "p25/util/logger.h"
#include "p25/util/sys.h"
#include "p25/comm/tcplib.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/**
 * @brief		Stores the user configurable p25 waveform properties available on the UI
 * @details		This is updated through messages from the java GUI over a TCP/IP socket
 */
UInterface      m_ui;

/**
 * @brief		Stores the p25 waveform properties decoded from received p25 frames
 * @details		This is also sent to the java GUI over a TCP/IP Socket
 */
UInterface      mr_ui;

//-------------------------------------------------------------------------
/**
 * @details		Reverse byte endian of a double variable
 * @param d		Input double variable
 * @return		Byte endian-reversed double variable
 * @author		Jeffrey Cuenco <jcuenco@ucsd.edu>
 */
//-------------------------------------------------------------------------
double swap(double d)
{
   double a;
   unsigned char *dst = (unsigned char *)&a;
   unsigned char *src = (unsigned char *)&d;

   dst[0] = src[7];
   dst[1] = src[6];
   dst[2] = src[5];
   dst[3] = src[4];
   dst[4] = src[3];
   dst[5] = src[2];
   dst[6] = src[1];
   dst[7] = src[0];

   return a;
}

//-------------------------------------------------------------------------
/**
 * @details	UserInterface class constructor
 *
 * @param mController	Pointer to Controller class instance
 * @param uiv_port		Graphical User Interface (GUI) communication port
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
UserInterface::UserInterface(Controller* mController, int uiv_port)
: mController(mController), ui_port(uiv_port) {
	int res;
	pthread_attr_t attr;
		
	INFO("Loading User Interface\n");
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	pthread_attr_setstacksize(&attr,131072);
	res = pthread_create(&uiThread, &attr, UserInterface::start_thread, (void*)this);
    	if (res) ERROR("Failed to create User Interface thread");
	pthread_attr_destroy(&attr);
}


static inline void print_ui(const UInterface & ui_tmp)
{
    INFO("ui.ptt = 0x%0X\n", ui_tmp.ptt);
    INFO("ui.nac = 0x%03X\n", ntohs(ui_tmp.nac));
    INFO("ui.tgId = 0x%04X\n", ntohs(ui_tmp.tgId));
    INFO("ui.keyId = 0x%04X\n", ntohs(ui_tmp.keyId));
    INFO("ui.algId = 0x%02X\n", ui_tmp.algId);
    INFO("ui.mfId = 0x%0X\n", ui_tmp.mfId);
    INFO("ui.ebit = 0x%0X\n", ui_tmp.ebit);
    INFO("ui.unitId = 0x%02X%02X%02X\n", ui_tmp.unitId[0], ui_tmp.unitId[1],
            ui_tmp.unitId[2]);
    INFO("ui.dstId = 0x%02X%02X%02X\n", ui_tmp.dstId[0], ui_tmp.dstId[1],
            ui_tmp.dstId[2]);
    INFO("ui.mi = 0x%02X%02X%02X%02X%02X%02X%02X%02X%02X\n", ui_tmp.mi[0],
            ui_tmp.mi[1], ui_tmp.mi[2], ui_tmp.mi[3], ui_tmp.mi[4],
            ui_tmp.mi[5], ui_tmp.mi[6], ui_tmp.mi[7], ui_tmp.mi[8]);
}

//-------------------------------------------------------------------------
/**
 * @details	Updates p25 waveform properties given input waveform property 
 *			struct.
 *
 * @param ui_cntrl	P25 waveform property struct
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterface::update(UInterface ui_cntrl)
{
    m_ui.cfreq = swap(ui_cntrl.cfreq);
    m_ui.nac = ntohs(ui_cntrl.nac);
    m_ui.tgId = ntohs(ui_cntrl.tgId);
    m_ui.keyId = ntohs(ui_cntrl.keyId);
    m_ui.algId = ui_cntrl.algId;
    m_ui.mfId = ui_cntrl.mfId;
    m_ui.ebit = ui_cntrl.ebit;
    memcpy(m_ui.unitId,ui_cntrl.unitId,sizeof(m_ui.unitId));
    memcpy(m_ui.dstId,ui_cntrl.dstId,sizeof(m_ui.dstId));
    memcpy(m_ui.mi,ui_cntrl.mi,sizeof(m_ui.mi));
    m_ui.ptt = ui_cntrl.ptt;
}

//-------------------------------------------------------------------------
/**
 * @details	Updates cache of received p25 waveform properties decoded from 
 *			received p25 frames and sends a copy to the GUI.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterface::r_ui_update()
{
    char buf[300];
    UInterface r_ui_cntrl;

    buf[0] = 0x0A;
    buf[1] = 0xA0;

    r_ui_cntrl.cfreq = swap(mr_ui.cfreq);
    r_ui_cntrl.nac = htons(mr_ui.nac);
    r_ui_cntrl.tgId = htons(mr_ui.tgId);
    r_ui_cntrl.keyId = htons(mr_ui.keyId);
    r_ui_cntrl.algId = mr_ui.algId;
    r_ui_cntrl.mfId = mr_ui.mfId;
    r_ui_cntrl.ebit = mr_ui.ebit;
    memcpy(r_ui_cntrl.unitId,mr_ui.unitId,sizeof(r_ui_cntrl.unitId));
    memcpy(r_ui_cntrl.dstId,mr_ui.dstId,sizeof(r_ui_cntrl.dstId));
    memcpy(r_ui_cntrl.mi,mr_ui.mi,sizeof(r_ui_cntrl.mi));
    //memset(r_ui_cntrl.tfpath,0,sizeof(r_ui_cntrl.tfpath));
    //r_ui_cntrl.tfmode = 0;
    r_ui_cntrl.ptt = mr_ui.ptt;

    memcpy(buf+2,(char* )&r_ui_cntrl,sizeof(UInterface));
    send_data(buf,2+sizeof(UInterface));

    print_ui(r_ui_cntrl);
}

//-------------------------------------------------------------------------
/**
 * @details	Prepares the UserInterface instance for GUI communication.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterface::initialize()
{
    memset((char* )&m_ui,0,sizeof(UInterface));
    ui_connFd = 0;
    ui_sock = 0;
    // wait for ui connection
    // Don't start TX or RX until a User Interface is detected
    ui_sock = start_tcp_server(ui_port);
    INFO("Waiting for UI connection on port %d [sock = %d]\n",ui_port,ui_sock);
    do {
        ui_connFd = get_tcp_server_connection(ui_sock);
        usleep(100);
    } while (!ui_connFd);
    ui_connected = 1;   

    INFO("UI Connected\n");
}

//-------------------------------------------------------------------------
/**
 * @details	Gracefully ends GUI communication and shuts down the 
 *			UserInterface instance.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterface::shutdown()
{
    close_tcp_server_connection(ui_connFd);
    stop_tcp_server(ui_sock);
    ui_connected = 0;
    INFO("Switching to Standby activity\n");
    mController->set_radio_activity(STANDBY_ACTIVITY);
}

//-------------------------------------------------------------------------
/**
 * @details	Receives data from GUI via TCP socket connection.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int UserInterface::receive_data(char* buf, int len) 
{
    int res;

    res =  read_tcp_server_stream(ui_connFd,buf,len);
    if (res <= 0) ui_connected = 0;
    return res;
}

//-------------------------------------------------------------------------
/**
 * @details	Sends data to GUI via TCP socket connection.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterface::send_data(char* buf, int len)
{
	if (ui_connected) {
        send_tcp_client_data(ui_connFd,buf,len);
    }
}

//-------------------------------------------------------------------------
/**
 * @details	Begin listening for GUI connections.  When connected to a GUI,
 *			any p25 waveform properties received from the GUI are cached
 *			and propagated to the waveform, and any p25 waveform properties
 *			received from decoded frames are relayed to the GUI.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterface::run_ui()
{
    int res;
    char command[2];
    UInterface ui_cntrl;
    double old_cfreq = 0.0;
    
   
    INFO("User Interface Loaded\n");
	
    while(mController->get_radio_state() == RUNNING_STATE) {
        initialize();
        while (ui_connected) {
            /**
             * @todo Check if this is a blocking call. If so
             * remove call to usleep() at the bottom.
             */
            res = receive_data(command,2);

            if ((command[0] == p25_cntrl_header[0]) && (command[1] == p25_cntrl_header[1])) {
                res = receive_data((char* )&ui_cntrl,sizeof(UInterface));

                if (res < sizeof(UInterface)) {
                	if (res > 0) {
                		INFO("Incorrect UInterface packet size received:  %d instead of %d\n",
                				res, sizeof(UInterface));
                	}
                	continue;
                }

                ui_cntrl.cfreq = swap(ui_cntrl.cfreq);
                update(ui_cntrl);

                // Let GUI configure whether TX/RX modes will read-from/
                // write-to a file (test mode) or from the VC55 serial
                // vocoder (default mode).
				if (ui_cntrl.tfmode == 0xF5) {
					//INFO("File test mode set.\n");
					//INFO("File to RX to or TX from:  %s\n", ui_cntrl.tfpath);
				} else {
					//INFO("VC55 serial vocoder mode set.\n");
				}

				// If old value of carrier frequency is different than new
				// one, send the new frequency.
				if (old_cfreq != ui_cntrl.cfreq) {
					INFO("Received new frequency change: %5.4f MHz\n", 
						ui_cntrl.cfreq / 1e6);
					res = set_fm_fc(ui_cntrl.cfreq);
				}
				old_cfreq = ui_cntrl.cfreq;

				if (m_ui.ptt == 0x01) {
					INFO("Switching to TX_ACTIVITY\n");
					mController->set_radio_activity(TX_ACTIVITY);
				} else {
					INFO("Switching to RX_ACTIVITY\n");
					mController->set_radio_activity(RX_ACTIVITY);
				}
            }

            /** @todo This is not needed if receive_data is a blocking call. */
            usleep(10);
        }
        shutdown();
    }
}

//-------------------------------------------------------------------------
/**
 * @details	Waits for main loop thread termination.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterface::join()
{
	pthread_join(uiThread, NULL);
}
	
//-------------------------------------------------------------------------
/**
 * @details	Starts main loop thread.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void* UserInterface::start_thread(void *obj)
{
	static_cast<UserInterface *>(obj)->run_ui();
	return 0;
}
