//-------------------------------------------------------------------------
/**
 * @file 	vocoder.cpp
 * @brief	Contains implementation of the Vocoder class
 * @note	none.
 *
 * @author	Cal(it)2
 * @author  Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 *		1.0 Jul 27 2009 	Anthony Nwokafor    C Model Structure created
 * 		1.1 Feb 22 2011 	Anthony Nwokafor    C++ Model Structure created
 * @endverbatim
 */
//-------------------------------------------------------------------------

#include <iostream>
#include <termios.h>
#include <fcntl.h>
#include "p25/format/bit_fifo.h"
#include "p25/util/logger.h"
#include "vocoder.h"

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <stdlib.h>
#include <netinet/in.h>

#define VC_PKT_SIZE VC_BLEN/8
#define TRACE printf


// tick-tock timer macros
#include <sys/time.h>
#include <iostream>
#define TICK() struct timeval tp1, tp2;						\
{															\
	gettimeofday(&tp1, NULL);								\
}

#define _TOCK(fmt, thresh)	{								\
	gettimeofday(&tp2, NULL);								\
	if (((tp2.tv_usec - tp1.tv_usec) / 1e3) > thresh) 		\
		std::cout << fmt << " time:  " << 					\
			(tp2.tv_usec - tp1.tv_usec) / 1e3 << " ms."     \
				<< std::endl;								\
}

#define TOCK(fmt)		_TOCK(fmt, 0)


void dumpvc(const char *buf, int len)
{
	if (len == VC_PKT_SIZE) {
		TRACE("\nContents of VC packet:\n");
		TRACE("%02hhx %02hhx %02hhx %02hhx %02hhx %02hhx\n",
				buf[0], buf[1], buf[2], buf[3], buf[4], buf[5]);
		TRACE("%02hhx %02hhx %02hhx %02hhx %02hhx %02hhx\n",
				buf[6], buf[7], buf[8], buf[9], buf[10], buf[11]);
		TRACE("%02hhx %02hhx %02hhx %02hhx %02hhx %02hhx\n\n",
				buf[12], buf[13], buf[14], buf[15], buf[16], buf[17]);
	}
	else
		TRACE("ERROR:  invalid packet, dump() expects %d-byte frames\n",
				VC_PKT_SIZE);
}

/*
 * @param Opens a file and returns a pointer to a dynamically-allocated
 *        buffer containing the complete file data, as well as the size
 *        of the buffer (size variable passed by reference).
 *
 */
void *loadfile(const char *filename, long *size, const char *fmode)
{
    FILE *fp;
    long lSize;
    char *buffer;

    /* open file in binary mode */
    fp = fopen ( filename , fmode );
    if( !fp ) perror(filename);

    /* get filesize */
    fseek( fp , 0L , SEEK_END);
    lSize = ftell( fp );
    rewind( fp );

    /* allocate memory for entire content */
    buffer = (char *)calloc( 1, lSize+1 );
    if( !buffer ) fclose(fp),fputs("memory alloc fails",stderr),exit(1);

    /* copy the file into the buffer */
    if( 1!=fread( buffer , lSize, 1 , fp) )
      fclose(fp),free(buffer),fputs("entire read fails",stderr),exit(1);

    /* do your work here, buffer is a string contains the whole text */
    *size = lSize;

    /* close file */
    fclose(fp);

    return buffer;
}

/*
 * @param Reads data from file into buf.
 *
 */
int Vocoder::readfile(char *buf, int len)
{
	int res;

	if (fbufpos+len <= fbufsize) {
		res = len;
		memcpy(buf, fbuf+fbufpos, res);
		fbufpos += res;
	}
	else {
		res = fbufsize-fbufpos;
		memcpy(buf, fbuf+fbufpos, res);
		fbufpos += res;
	}

	return res;
}

/*
 * @param
 *
 */
Vocoder::Vocoder(Controller* _mController, char* _vc_sport, char* _vc_rxfile)
:mController(_mController), vc_sport(_vc_sport), vc_rxfile(_vc_rxfile)
{
    int res;
    pthread_attr_t attr;
	struct sched_param sched_param;
	sched_param.sched_priority = 99;  // highest priority

    pthread_attr_init(&attr);
	pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
	pthread_attr_setschedparam(&attr, &sched_param);
	pthread_attr_setstacksize(&attr,131072);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    res = pthread_create(&vcThread, &attr, Vocoder::start_thread, (void*)this);
    if (res) ERROR("Failed to create Vocoder thread");
    pthread_attr_destroy(&attr);
	
}

void Vocoder::initialize()
{
    pb_bf = bf_alloc(PLAY_BACK_BUF_SIZE);
    vc_bf = bf_alloc(VC_BLEN);
    vc_sock = 0;
    vc_bf_level = 0;
    vc_pb_bf_level = 0;

    pthread_mutex_init (&vc_vc_bf_m, NULL);
    pthread_mutex_init (&vc_pb_bf_m, NULL);

    if (mController->get_radio_mode() == LIVE_MODE) {

		fdSerial = open(vc_sport, O_RDWR | O_NOCTTY | O_NDELAY);
		if (fdSerial == -1) {
			ERROR("Failed to open Serial Port %s\n",vc_sport);
			exit(-1);
		}

		fcntl(fdSerial, F_SETFL, 0);

		tcgetattr(fdSerial, &port_settings);

		cfsetispeed(&port_settings, B115200);
		cfsetospeed(&port_settings, B115200);

		port_settings.c_cflag &= ~PARENB;
		port_settings.c_cflag &= ~CSTOPB;
		port_settings.c_cflag &= ~CSIZE;
		port_settings.c_cflag |= CS8;
		port_settings.c_cflag |= (CLOCAL | CREAD);

		port_settings.c_cflag &= ~CRTSCTS;				// disable hardware flow control
		port_settings.c_iflag &= ~(IXON | IXOFF | IXANY);		// disable software flow control
		port_settings.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);	// use raw input		
		port_settings.c_oflag &= ~OPOST;				// use raw output

		//port_settings.c_cc[VMIN] = 20;
		//port_settings.c_cc[VTIME] = 0;

		tcsetattr(fdSerial, TCSANOW, &port_settings);

		INFO("Established serial connection to VC55 on port %s\n",vc_sport);
    } else {
    	INFO("Radio in Test Mode\n");
    	INFO("Opening %s for reading, Opening %s for writing\n",vc_sport, vc_rxfile);

    	// Open files
	//fbuf = (char *)loadfile(vc_sport, &fbufsize, "rb+");
	//fbufpos = 0;

    	test_mode_file_tx = fopen(vc_sport, "rb+");
    	test_mode_file_rx = fopen(vc_rxfile, "wb+");

    	if (!test_mode_file_tx) {
    		fprintf(stderr, "Unable to open file %s", vc_sport);
    		exit(-1);
    	}
    	if (!test_mode_file_rx) {
    		fprintf(stderr, "Unable to open file %s", vc_rxfile);
    		exit(-1);
    	}
    }
}

void Vocoder::shutdown()
{
    bf_free(pb_bf);
    bf_free(vc_bf);

    pthread_mutex_destroy (&vc_vc_bf_m);
    pthread_mutex_destroy (&vc_pb_bf_m);

    if (mController->get_radio_mode() == TEST_MODE) {
    	fclose(test_mode_file_tx);
    	//free(fbuf);
    	fclose(test_mode_file_rx);
    }
}
/**
 * @details
 * @param[in] buf Buffer to store Voice Code Word. Size must be VC_BLEN bytes.
 * @return Number of Voice Code Word bytes in buffer.
 * @retval Value should be equivalent to VC_BLEN.
 */
int Vocoder::receive_vc_data(char *buf)
{
	int i, res;
	char b;

	res = 0;

	// Get VC packet
	pthread_mutex_lock(&vc_vc_bf_m);
	res = getvc(buf);
	pthread_mutex_unlock(&vc_vc_bf_m);
	//INFO("$\n");
	
	if (res > 0) {
		// reset bit FIFO for each VC packet
		bf_reset(vc_bf);
		
		// Format VC bytes so that each bit is represented by a byte
		for (i = 0; i < VC_BLEN; i++) {
			// Extract each bit from each byte
			b = (buf[i/8] >> (7 - i%8)) & 0x01;
			// Put bit in bit FIFO
			bf_put_n(vc_bf, &b, 1);
		}
		// Extract bits from bit FIFO
		res = bf_get_n(vc_bf,buf,VC_BLEN);
	}

	return res;
}

void Vocoder::send_vc_data()
{
	int i;
	char b, buf[VC_PKT_SIZE];

	memset(buf, 0, VC_PKT_SIZE);
	pthread_mutex_lock(&vc_pb_bf_m);
	if(vc_pb_bf_level) {

		for (i = 0; i < VC_BLEN; i++) {
			bf_get_n(pb_bf, &b, 1);
			buf[i/8] |= b<<(7 - i%8);
		}
			
        putvc(buf,VC_PKT_SIZE);
        vc_pb_bf_level -= VC_BLEN;
        //DBG("-\n");
	}
	pthread_mutex_unlock(&vc_pb_bf_m);

}

void Vocoder::put_in_pb_buf(char *vcw, int len)
{
    int offset;

	offset = 0;
	while(len >= VC_BLEN) {
		pthread_mutex_lock(&vc_pb_bf_m);
		if (vc_pb_bf_level < PLAY_BACK_BUF_SIZE) {
				bf_put_n(pb_bf,vcw+offset,VC_BLEN);
				len -= VC_BLEN;
				offset += VC_BLEN;
				vc_pb_bf_level += VC_BLEN;
		    //DBG("+\n");
		} else {
		    len -= VC_BLEN;
		    offset += VC_BLEN;
		    DBG("*\n");
		}
		pthread_mutex_unlock(&vc_pb_bf_m);
	}
}

void Vocoder::read_vocoder()
{
	char buf[VC_PKT_SIZE];

	pthread_mutex_lock(&vc_vc_bf_m);
	getvc(buf);
	pthread_mutex_unlock(&vc_vc_bf_m);
		//DBG("*\n");
}

int Vocoder::getvc(char* buf)
{
	unsigned char szBuff[2];
	int res, serror, bytesLeft;

	bool dumpTX = false;
	bool hdr_lock = false;

	if (mController->get_radio_mode() == LIVE_MODE) {
		// Read vc55 every 20ms which corresponds to 18 bytes + 2 header bytes				

		// Continue to receive packets one by one until we "lock" onto a header
		while (!hdr_lock) {
			// Read in first byte of header, 0x56
			res = read(fdSerial, (char *)szBuff, 1);
			if (res != 1) {
				if (res < 0) {				
					// Error occurred.  Report to user.
					serror = errno;					
					TRACE("ERROR: Failed to read 1st header byte from %s:  errno = %d (%s)\n", 
						vc_sport, serror, strerror(serror));
				}
				return 0;
			}
			if (szBuff[0] != 0x56) {
				// Wrong header
				TRACE("ERROR: Wrong 1st header byte from %s:  0x%02X\n", vc_sport, szBuff[0]);
				continue;
			}

			// Read in second byte of header, 0xF0 if VC55 is configured for Full Rate.  If the second
			// header byte received is 0x0F, then the VC55 is configured for Half Rate, which is currently
			// not supported by this implementation.
			res = read(fdSerial, (char *)szBuff+1, 1);
			if (res != 1) {
				if (res < 0) {				
					// Error occurred.  Report to user.
					serror = errno;					
					TRACE("ERROR: Failed to read 2nd header byte from %s:  errno = %d (%s)\n", 
						vc_sport, serror, strerror(serror));
				}
				return 0;
			}
			if (szBuff[1] != 0xF0) {
				if (szBuff[1] == 0x0F) {
					// VC55 configured for Half Rate mode.  Inform  user that board needs to be
					// reconfigured for Full Rate.
					TRACE("ERROR: Half-rate mode detected:  0x%02X\n", szBuff[1]);
					return 0;
				}
				else {
					// Wrong header
					TRACE("ERROR: Wrong 2nd header byte from %s:  0x%02X\n", vc_sport, szBuff[1]);
					continue;
				}				
			}

			hdr_lock = true;
		}

		// Reads in 18 vocoder bytes.
		res = read(fdSerial, buf, VC_PKT_SIZE);
		if (res != VC_PKT_SIZE) {
			if (res < 0) {				
				// Error occurred.  Report to user.
				serror = errno;					
				TRACE("ERROR: could not read data from %s:  errno = %d (%s)\n", 
					vc_sport, serror, strerror(serror));
				return 0;
			}
			else {
				bytesLeft = VC_PKT_SIZE - res;
				while (bytesLeft > 0) {
					res = read(fdSerial, buf+(VC_PKT_SIZE-bytesLeft), bytesLeft);
					if (res < 0) {				
						// Error occurred.  Report to user.
						serror = errno;					
						TRACE("ERROR: could not read data from %s:  errno = %d (%s)\n", 
							vc_sport, serror, strerror(serror));
						return 0;
					}		
					bytesLeft -= res;
				}
				res = VC_PKT_SIZE;
			}
		}
	} else {
		dumpTX = true;

		res = fread(buf, 1, VC_PKT_SIZE, test_mode_file_tx);
		//res = readfile(buf, VC_PKT_SIZE);

		//INFO("read %d bytes\n",res);
		if (feof(test_mode_file_tx)) {
		//if (fbufpos == fbufsize) {
			memset(buf+res,0,VC_PKT_SIZE-res);
			fseek(test_mode_file_tx, 0, SEEK_SET);  // go back to beginning of file
			//fbufpos = 0;
			mController->set_radio_activity(RX_ACTIVITY);
		}
		
		usleep(18000);
	}

	if (dumpTX && res == VC_PKT_SIZE && mController->get_radio_activity() == TX_ACTIVITY) {
		// pktcount++;
		// Dump entire packet including header
		dumpvc(buf,res);
	}

	return res;
}

void Vocoder::putvc(char* buf, int len)
{
    int res;
    unsigned char nbuf[len+2];
    bool dumpRX = false;
	
    if (mController->get_radio_mode() == LIVE_MODE) {
        // Set header block for return to VC55, 0x56 is sync byte, and
    	// 0xF0 is for full-rate (7200bps) operation.
        nbuf[0] = 0x56;
        nbuf[1] = 0xF0;
        memcpy(nbuf+2,buf,len);

        res = write(fdSerial,nbuf,len+2);
        if (res != len+2) {
            // error occurred.  Report to user.
            TRACE("ERROR: error occurred writing data to serial port!\n");
        }
    } else {
	dumpRX = true;

    	res = fwrite(buf,1,len,test_mode_file_rx);
    	if (res != len)
    		ERROR("Error writing data to file\n");
    }

    if (dumpRX && mController->get_radio_activity() == RX_ACTIVITY) {
        // Dump entire packet including header
        dumpvc(buf, len);
    }
}

void Vocoder::run_vc()
{
    initialize();

	/** @todo	Change while(true) to depend on radio state. */
    while(mController->get_radio_state() == RUNNING_STATE) {
		if (mController->get_radio_activity() == RX_ACTIVITY) {
			// Use VC55 for 20 ms timing if we're in live mode
			if (mController->get_radio_mode() == LIVE_MODE)
				read_vocoder();
			else
				usleep(20000);

			// Send Voice Code Words
			send_vc_data();
		} else {
			/** @todo Replace call to usleep with semaphore. This is taking up unnecessary CPU cycles */
			usleep(10);	
		}
    }
    shutdown();
}


void Vocoder::join()
{
	pthread_join(vcThread, NULL);
}

void* Vocoder::start_thread(void *obj)
{
	static_cast<Vocoder *>(obj)->run_vc();
	return 0;
}
