/**
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <openssl/des.h>
#include <semaphore.h>

#include "p25/format/byte_fifo.h"
#include "p25/util/logger.h"
#include "p25_cryptography.h"

static char LFSR[64];

void dump(const char *str, unsigned char *buf, int len) {
	int i=0;

	printf("\n==== %s ====",str);
	for (i=0; i < len; i++) {
		if (i%8 == 0)
			printf("\n");

		printf("%02x ",buf[i]);
	}
	printf("\n");
}

/**
 * @details Puts packetizer data into crypt->packetizer buffer.
 *
 * @author Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
int send_crypto_pktizer(char *buf, int len) {
	// @todo: Add mutexes
	int ret;

	//INFO("Entered send_crypto_pktizer [Try Put = %d]\n", len);

	// If space in crypto -> pktizer FIFO is less than the requested len, wait
	// for the pktizer process to free up more space
	while (bf_space(bf_crypto_pktizer) < len) {
		//INFO("sem_wait on sem_crypto_pktizer_empty\n");
		sem_wait(&sem_crypto_pktizer_empty);
		//INFO("returned from sem_wait on sem_crypto_pktizer_empty\n");
	}

	// Put buf in crypto->pktizer FIFO
	ret = bf_put_n(bf_crypto_pktizer, buf, len);

	// Inform pktizer process of new data
	// @todo: might want to do this only when the fifo size was 0 prior to
	// inserting len bytes
	sem_post(&sem_crypto_pktizer_notempty);

	//INFO("Leaving send_crypto_pktizer [Put = %d]\n", ret);

	return ret;
}

/**
 * @details Receives packetizer data from packetizer->crypto buffer.
 *
 * @author Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
int recv_pktizer_crypto(char *buf, int len) {
	// @todo: Add mutexes
	int ret;

	//INFO("Entered recv_pktizer_crypto [Req = %d]\n", len);
	// If number of bytes in pktizer -> cryto FIFO is less than the requested
	// len, wait for more bytes to be added by the pktizer process
	while ((bf_size(bf_pktizer_crypto) < len) && !bf_pktizer_crypto_accept_less) {
		//INFO("sem_wait on sem_pktizer_crypto_notempty\n");
		sem_wait(&sem_pktizer_crypto_notempty);
		//INFO("returned from sem_wait on sem_pktizer_crypto_notempty\n");
	}

	// Get len bytes from pktizer->crypto FIFO
	ret = bf_get_n(bf_pktizer_crypto, buf, len);
	if (ret != len)
		INFO("Requested %d, Got %d",len,ret);

	if (bf_pktizer_crypto_accept_less)
		bf_pktizer_crypto_accept_less = 0;


	// Inform pktizer process that some space has been created in 
	// pktizer -> crypto FIFO
	sem_post(&sem_pktizer_crypto_empty);

	//INFO("Leaving recv_pktizer_crypto [Got = %d]\n", ret);

	return ret;
}

/**
 * @details Puts voice code words into crypto->vocoder buffer.
 *
 * @author Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
int send_crypto_vocoder(char *buf, int len) {
	// @todo: Add mutexes
	int ret;

	//INFO("Entered send_crypto_vocoder [Try Put = %d]\n", len);

	// If space in crypto -> vocoder FIFO is less than the requested len, wait
	// for the vocoder process to free up more space
	while (bf_space(bf_crypto_vocoder) < len) {
		//INFO("sem_wait on sem_crypto_vocoder_empty\n");
		sem_wait(&sem_crypto_vocoder_empty);
		//INFO("returned from sem_wait on sem_crypto_vocoder_empty\n");
	}


	// Put buf in crypto -> vocoder FIFO
	ret = bf_put_n(bf_crypto_vocoder, buf, len);

	// Inform vocoder process of new data
	// @todo: might want to do this only when the fifo size was 0 prior to
	// inserting len bytes
	sem_post(&sem_crypto_vocoder_notempty);

	//INFO("Leaving send_crypto_vocoder [Put = %d]\n", ret);

	return ret;
}

/**
 * @details Gets voice code words from vocoder->crypto buffer.
 *
 * @author Anthony Nwokafor <aanwokafor@ucsd.edu>
*/
int recv_vocoder_crypto(char *buf, int len) {
	// @todo: Add mutexes
	int ret;

	//INFO("Entered recv_vocoder_crypto [Req = %d]\n", len);

	// If number of bytes in vocoder -> cryto FIFO is less than the requested 
	// len wait for more bytes to be added by the vocoder process
	while ((bf_size(bf_vocoder_crypto) < len) && !bf_vocoder_crypto_accept_less) {
		//INFO("sem_wait on sem_vocoder_crypto_notempty\n");
		sem_wait(&sem_vocoder_crypto_notempty);
		//INFO("returned from sem_wait on sem_vocoder_crypto_notempty\n");
	}

	// Get len bytes from vocoder -> crypto FIFO
	ret = bf_get_n(bf_vocoder_crypto, buf, len);
	if (ret != len)
		INFO("Requested %d, Got %d",len,ret);

	if (bf_vocoder_crypto_accept_less)
			bf_vocoder_crypto_accept_less = 0;

	// Inform vocoder process that some space has been created in 
	// vocoder -> crypto FIFO
	sem_post(&sem_vocoder_crypto_empty);

	//INFO("Leaving recv_vocoder_crypto [Got = %d]\n", ret);

	return ret;
}

/**
 * @details Where the actual OFB encryption takes place.
 * @see DES_ofb64_encrypt()
 * @author Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
void xcrypt(P25_DES_SCHED_STATE *dState, char *buf) {
	// do one interation of ofb encryption
	
	char out[BLK_SZ];
	
	// Encryption. 
	DES_ofb64_encrypt((unsigned char *)buf, (unsigned char *)out, BLK_SZ,
			&dState->sched, &dState->mi, &dState->mi_idx);
	
	// use input buffer as output buffer. So contents of buffer will be replaced
	// by the plain text equivalent.
	memcpy(buf,out,BLK_SZ);

	dState->word_count++;
	//dump("mi",(unsigned char *)dState->mi,sizeof(dState->mi));
}



/**
 * @details Resets cryptography state variables.
 *
 * @author Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
void reset_dState(P25_DES_SCHED_STATE *dState) {
	dState->done = 0;
	dState->word_count = 0;
	dState->mi_idx = 0;
}

/**
 * @details Initializes cryptography process.
 *
 * @author Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
void init_dState(P25_DES_SCHED_STATE *dState, char *key, char *mi) {
	DES_cblock desKey;

	reset_dState(dState);

	memset(&dState->mi, 0, sizeof(dState->mi));
	memcpy(dState->mi, mi, sizeof(dState->mi));

	/* Prepare the key for use with DES_ofb64_encrypt */
	memset(&dState->sched, 0, sizeof(dState->sched));
	memcpy(desKey, key, sizeof(desKey));
	DES_set_odd_parity(&desKey);
	DES_set_key_checked(&desKey, &dState->sched);
}

void init_lfsr(char *iv) {
	int stages = sizeof(LFSR);
	int i,idx;

	memcpy(LFSR, iv, sizeof(LFSR));

	// expand bits of iv into 1 byte per bit
	for (i=0; i < stages; i++) {
		idx = stages-1-i;
		LFSR[idx] = (iv[i/8] & (1 << (idx%8))) >> (idx%8);
	}
}

void lfsr_new_mi(char *mi) {
	int stages = sizeof(LFSR);
	char TMP_LFSR[stages];
	char *lfsr;
	char *tmp_lfsr;
	char *old_lfsr;
	int i,idx;

	tmp_lfsr = TMP_LFSR;
	lfsr = LFSR;

	for (i=0; i < stages; i++) {
		memcpy(tmp_lfsr+1,lfsr,stages-1);
		tmp_lfsr[0] = lfsr[14]^lfsr[26]^lfsr[37]^lfsr[45]^lfsr[61]^lfsr[63];

		old_lfsr = lfsr;
		lfsr = tmp_lfsr;
		tmp_lfsr = old_lfsr;
	}

	// expand bits of iv into 1 byte per bit
	memset(mi, 0, BLK_SZ);
	for (i=0; i < stages; i++) {
		idx = stages-1-i;
		mi[i/8] = mi[i/8] | (LFSR[idx] << (idx%8));
	}

//	dump("mi",(unsigned char *)mi,BLK_SZ);
}

/**
 * @details Initializes all vocoder and packetizer buffers,
 * 			semaphores for state machine, and sets state
 *			variables disallowing threads to read less
 *			than the requested amount of data from buffers.
 *
 * @author Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
int crypto_init() {
	bf_pktizer_crypto = bf_alloc((8+8+4)*2); //mi+lc+lsd
	bf_crypto_pktizer = bf_alloc((8+8+4)*2);
	bf_vocoder_crypto = bf_alloc(11*18*2);
	bf_crypto_vocoder = bf_alloc(11*18*2);

	sem_init(&sem_pktizer_crypto_empty, 0, 0);
	sem_init(&sem_pktizer_crypto_notempty, 0, 0);
	sem_init(&sem_crypto_pktizer_empty, 0, 0);
	sem_init(&sem_crypto_pktizer_notempty, 0, 0);

	sem_init(&sem_vocoder_crypto_empty, 0, 0);
	sem_init(&sem_vocoder_crypto_notempty, 0, 0);
	sem_init(&sem_crypto_vocoder_empty, 0, 0);
	sem_init(&sem_crypto_vocoder_notempty, 0, 0);

	sem_init(&sem_pktizer_crypt_processData, 0, 0);
	sem_init(&sem_pktizer_ProcessData, 0, 0);

	sem_init(&sem_vocoder_crypt_processData, 0, 0);
	sem_init(&sem_vocoder_ProcessData, 0, 0);

	sem_init(&sem_crypt_ProcessData, 0, 0);

	sem_init(&sem_pktizer_crypto_finished, 0, 0);
	sem_init(&sem_vocoder_crypto_finished, 0, 0);

	pthread_mutex_init (&mtx_crypto_reset, NULL);

	crypto_reset_done = 0;

    bf_pktizer_crypto_accept_less = 0;
    bf_crypto_pktizer_accept_less = 0;
    bf_vocoder_crypto_accept_less = 0;
    bf_crypto_vocoder_accept_less = 0;

    pktizer_crypto_finished = 1;
    vocoder_crypto_finished = 1;

	return (0);
}

/**
 * @details Resets all vocoder and packetizer buffers.
 * 			Posts all semaphores that might be waiting on
 *			additional data.
 *
 * @author Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
void crypto_reset() {
	pthread_mutex_lock(&mtx_crypto_reset);

	if (!crypto_reset_done) {
//		INFO("Reseting Encryption Device States\n");
		bf_reset(bf_crypto_pktizer);
		bf_reset(bf_pktizer_crypto);
		bf_reset(bf_vocoder_crypto);
		bf_reset(bf_crypto_vocoder);

		bf_pktizer_crypto_accept_less = 1;
		bf_crypto_pktizer_accept_less = 1;
		bf_vocoder_crypto_accept_less = 1;
		bf_crypto_vocoder_accept_less = 1;

		sem_post(&sem_crypto_pktizer_notempty);
		sem_post(&sem_pktizer_crypto_notempty);
		sem_post(&sem_crypto_vocoder_notempty);
		sem_post(&sem_vocoder_crypto_notempty);

		crypto_reset_done = 1;
	}
	pthread_mutex_unlock(&mtx_crypto_reset);
}

