/****************************************************************************
Copyright 2007 Virginia Polytechnic Institute and State University
This file is part of the OSSIE __COMP_NAME__.
OSSIE __COMP_NAME__ is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
"(at your option) any later version
OSSIE __COMP_NAME__ is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with OSSIE __COMP_NAME__; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

/**
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 */

#include <string>
#include <iostream>
#include <stdio.h>
#include <pthread.h>
#include <netinet/in.h>
#include "UserInterfaceDevice.h"

/* user includes */
#include "p25/config.h"
#include "p25/format/packets.h"
#include "p25/util/logger.h"
#include "ui.h"
#include <fstream>

#define SCANNING
#define SCAN_PERIOD		1 	//In micro seconds

double current_cfreq;

//-------------------------------------------------------------------------
/**
 * @details UI class constructor for OSSIE implementation
 *
 * @author
 */
//-------------------------------------------------------------------------
UserInterfaceDevice_i::UserInterfaceDevice_i(const char *uuid, omni_condition *condition) :
    Resource_impl(uuid), component_running(condition)
{
    rx_data_in = new standardInterfaces_i::realChar_p("ui_in_rx");
    tx_data_in = new standardInterfaces_i::realChar_p("ui_in_tx");
    tx_data_out = new standardInterfaces_i::realChar_u("ui_out_tx");
    rx_data_out = new standardInterfaces_i::realChar_u("ui_out_rx");
    modem_data_out = new standardInterfaces_i::realChar_u("ui_out_mod");
    modem_data_in = new standardInterfaces_i::realChar_p("ui_in_mod");

    // create UI class object instance (user_code)
    _ui = new UI();

    sem_init(&receiver_pkt_stop, 0, 0);

//    start();
}

//-------------------------------------------------------------------------
/**
 * @details UI class destructor for OSSIE implementation
 *
 * @author
 */
//-------------------------------------------------------------------------
UserInterfaceDevice_i::~UserInterfaceDevice_i(void)
{
    delete rx_data_in;
    delete tx_data_out;
    delete rx_data_out;
    delete tx_data_in;

    delete _ui;
}

//-------------------------------------------------------------------------
/**
 * @details Static function for omni thread
 *
 * @author
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::Run( void * data )
{
    ((UserInterfaceDevice_i*)data)->ProcessData();
}

//-------------------------------------------------------------------------
/**
 * @details Static function for omni thread
 *
 * @author
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::scanning_Run( void * data )
{
	((UserInterfaceDevice_i*)data)->scanning_ProcessData();
}

//-------------------------------------------------------------------------
/**
 * @details Static function for omni thread
 *
 * @author
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::run_ui(void *arg)
{
	((UserInterfaceDevice_i*)arg)->tcp_task();
}

//-------------------------------------------------------------------------
/**
 * @details Get port data
 *
 * @author
 */
//-------------------------------------------------------------------------
CORBA::Object_ptr UserInterfaceDevice_i::getPort( const char* portName ) throw (
    CORBA::SystemException, CF::PortSupplier::UnknownPort)
{
    DEBUG(3, UserInterfaceDevice, "getPort() invoked with " << portName)

    CORBA::Object_var p;

    p = rx_data_in->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = tx_data_in->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = tx_data_out->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = rx_data_out->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = modem_data_out->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = modem_data_in->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    /*exception*/
    throw CF::PortSupplier::UnknownPort();
}

//-------------------------------------------------------------------------
/**
 * @details Starts threads for UI component
 *
 * @author
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::start() throw (CORBA::SystemException,
    CF::Resource::StartError)
{
    DEBUG(3, UserInterfaceDevice, "start() invoked")
	omni_mutex_lock  l(processing_mutex);
	if( false == thread_started )
	{
		thread_started = true;
		// Create the thread for the writer's processing function
		processing_thread = new omni_thread(Run, (void *) this, omni_thread::PRIORITY_NORMAL);

		// Start the thread containing the writer's processing function
		processing_thread->start();
	}
    	uiPort = 8002;

    omni_thread::stacksize(2*1024*1024);

	omni_mutex_lock  ll(ui_mutex);
	if( false == ui_thread_started )
	{
		ui_thread_started = true;
		// Create the thread for the writer's processing function
		ui_thread = new omni_thread(run_ui, (void *) this, omni_thread::PRIORITY_NORMAL);

		// Start the thread containing the writer's processing function
		ui_thread->start();
	}
    INFO("User Interface thread loaded.\n");

    omni_mutex_lock  lll(scanning_processing_mutex);
	if( false == scanning_thread_started )
	{
		scanning_thread_started = true;
		// Create the thread for the writer's processing function
		scanning_thread = new omni_thread(scanning_Run, (void *) this, omni_thread::PRIORITY_NORMAL);

		// Start the thread containing the writer's processing function
		scanning_thread->start();
	}
    INFO("Scanning thread started\n");

    scan_running = true;
}

//-------------------------------------------------------------------------
/**
 * @details Stops all threads
 *
 * @author
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::stop() throw (CORBA::SystemException, CF::Resource::StopError)
{
    DEBUG(3, UserInterfaceDevice, "stop() invoked")
	omni_mutex_lock l(processing_mutex);
	thread_started = false;

	/* user code */
    omni_mutex_lock ll(ui_mutex);
    ui_thread_started = false;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::releaseObject() throw (CORBA::SystemException,
    CF::LifeCycle::ReleaseError)
{
    DEBUG(3, UserInterfaceDevice, "releaseObject() invoked")

    component_running->signal();
}

//-------------------------------------------------------------------------
/**
 * @details Initialize UI component
 *
 * @author
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::initialize() throw (CF::LifeCycle::InitializeError,
    CORBA::SystemException)
{
    DEBUG(3, UserInterfaceDevice, "initialize() invoked");
    sem_init(&sem_rx_running, 0, 0);
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::query( CF::Properties & configProperties ) throw (CORBA::SystemException, CF::UnknownProperties)
{
	unsigned int i, j;

	if( configProperties.length() == 0 )
	{
		configProperties.length( propertySet.length() );
		for( i = 0; i < propertySet.length(); i++ )
		{
			configProperties[i].id = CORBA::string_dup( propertySet[i].id );
			configProperties[i].value = propertySet[i].value;
		}
		return;
	} else {
		for( i = 0; i < configProperties.length(); i++ ) {
			for( j = 0; j < propertySet.length(); j++ ) {
				if( strcmp(configProperties[i].id, propertySet[j].id) == 0 ) {
					configProperties[i].value = propertySet[j].value;
				}
			}
		}
	} // end if-else
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::configure(const CF::Properties& props)
throw (CORBA::SystemException,
    CF::PropertySet::InvalidConfiguration,
    CF::PropertySet::PartialConfiguration)
{
    DEBUG(3, UserInterfaceDevice, "configure() invoked")

    static int init = 0;
	unsigned int i, j, k;

    if( init == 0 ) {
        if( props.length() == 0 ) {
            std::cout << "configure called with invalid props.length - " << props.length() << std::endl;
            return;
        }
        propertySet.length(props.length());
        for( j=0; j < props.length(); j++ ) {
            propertySet[j].id = CORBA::string_dup(props[j].id);
            propertySet[j].value = props[j].value;
        }
        init = 1;
    }

    std::cout << "props length : " << props.length() << std::endl;

    for ( i = 0; i <props.length(); i++)
    {
        std::cout << "Property id : " << props[i].id << std::endl;

        if (strcmp(props[i].id, "DCE:ee2fdaa2-9600-11e0-84df-0000e80000b1") == 0)
        {
            CORBA::Short simple_temp;
            props[i].value >>= simple_temp;
            simple_0_value = simple_temp;
            for( k = 0; k < propertySet.length(); k++ ) {
                if( strcmp(propertySet[k].id, props[i].id) == 0 ) {
                    propertySet[k].value = props[i].value;
                    break;
                }
            }
        }

    }
}

//-------------------------------------------------------------------------
/**
 * @details Receives control message from Rx component.  If the header
 * 			indicates ACK, the Rx component is in standby mode.
 * 			If header indicates UPDATE, an update UI has been requested
 * 			by the Rx component.
 * 			For scanning, if the header indicates SCAN_HALT, a valid
 * 			frame sync has been received.  If the header indicates
 * 			SCAN_CONTINUE, a TDU has been received.
 *
 * @author	Justyn Bell <jwbell@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::ProcessData()
{
    DEBUG(3, UserInterfaceDevice, "ProcessData() invoked");

    PortTypes::CharSequence *I_in_0(NULL);
    CORBA::UShort I_in_0_length;

   	ModemPacket mod_pkt;
	UIControlPacket *tmp_ui;
	PortTypes::CharSequence_var I_out_0;

    while(continue_processing()) {
        rx_data_in->getData(I_in_0);
        I_in_0_length = I_in_0->length();

    	tmp_ui = ((UIControlPacket *)(I_in_0->get_buffer()));

        /*insert code here to do work*/
    	switch (tmp_ui->hdr) {
    	case UPDATE:
    		DBG("Received UPDATE from Rx device\n");
			/* Since we're implementing scanning, we want the GUI to display the current frequency, no matter
				which frequency it should pick up.  The Rx device handles updating the ui struct, but the cfreq
				does not come from the Rx device.  We have to explicitly set it here, within the UI device */
			DBG("Sending current cfreq to GUI of %5.4f Mhz\n", current_cfreq/1e6);
			tmp_ui->ui.cfreq = current_cfreq;	
			_ui->ui_update(tmp_ui->ui);
          	_ui->r_ui_update(tmp_ui->ui);
          	break;
    	case ACK:
        	sem_post(&receiver_pkt_stop);
        	INFO("Received ACK message from RX device\n");
        	break;
    	case SCAN_HALT:
    		DBG("Received a SCAN_HALT from Rx Device\n");
    		// Stop scanning after we get an FS
    		scan_running = false;
    		//_ui->ui_update(tmp_ui->ui);
    		//_ui->r_ui_update(tmp_ui->ui);
    		_ui->stop_gui_scan();
    		break;
    	case SCAN_CONTINUE:
    		DBG("Received a SCAN_CONTINUE from Rx device\n");
			/* The following is only necesary if the user wants scanning to start
				automatically after a TDU is received*/
    		//scan_running = true;
    		//sem_post(&sem_rx_running);
    		// Start scanning again after we get a TDU
    		break;
		case FLUSH:
			mod_pkt.hdr = BS_INPUT_FLUSH;
			I_out_0 = new PortTypes::CharSequence(sizeof(ModemPacket),sizeof(ModemPacket), (CORBA::Char *)&mod_pkt, 0);
			modem_data_out->pushPacket(I_out_0);
			break;	
    	default:
        	ERROR("Unknown header!!!!! Value of pkt.hdr = 0x%02X\n", tmp_ui->hdr);
        	break;
    	}
        rx_data_in->bufferEmptied();
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
bool UserInterfaceDevice_i::continue_processing()
{
	omni_mutex_lock l(processing_mutex);
	return thread_started;
}

//-------------------------------------------------------------------------
/**
 * @details Receives information from GUI component, and calls
 * 			setPttMode with state information.
 *
 * @author Jeffrey Cuenco <jcuenco@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::tcp_task(void)
{
    int res;
    char command[2];
    UInterface ui_cntrl;
    SUInterface s_ui_cntrl;

    INFO("User Interface Loaded\n");

    while(ui_continue_processing()) {
    	_ui->ui_initialize(uiPort);
        while (_ui->get_ui_connected()) {
            res = _ui->receive_ui_data(command,2);
            if ((command[0] == p25_cntrl_header[0]) && (command[1] == p25_cntrl_header[1])) {
                res = _ui->receive_ui_data((char *)&ui_cntrl,sizeof(UInterface));

                if (res < (signed) sizeof(UInterface)) {
                	if (res > 0) {
                		INFO("Incorrect UInterface packet size received:  %d instead of %d\n",
                				res, sizeof(UInterface));
                	}
                	continue;
                }

                _ui->ui_update(ui_cntrl);
				current_cfreq = (_ui->get_ui()).cfreq;
                INFO("Got frequency from GUI, %5.4f\n", ((_ui->get_ui()).cfreq)/1e6); 
				INFO("Set current_cfreq to %5.4f\n", current_cfreq);

                if (_ui->get_ui().ptt == 0x01) {
                    INFO("Switching to TX_ACTIVITY\n");
                    setPttMode(_ui->get_ui(), TX_ACTIVITY);
                } else {
                    INFO("Switching to RX_ACTIVITY\n");
                    setPttMode(_ui->get_ui(), RX_ACTIVITY);
                }
            }

            else if ((command[0] == p25_scan_header[0]) && (command[1] == p25_scan_header[1])) {
            	res = _ui->receive_ui_data((char *)&s_ui_cntrl,sizeof(SUInterface));

            	if (res < (signed) sizeof(SUInterface)) {
            		if (res > 0) {
            			ERROR("Incorrect SUInterface packet size received:  %d instead of %d\n",
            					res, sizeof(SUInterface));
            		}
            		continue;
            	}
            	_ui->scan_ui_update(s_ui_cntrl);

            	if ((_ui->get_s_ui()).state == G_SCAN_CONTINUE) {
                	scan_running = true;
                	sem_post(&sem_rx_running);
            	}
            	else if ((_ui->get_s_ui()).state == G_SCAN_HALT) {
                	scan_running = false;
            	}
            }
            else {
            	ERROR("Read 0x%X and 0x%X from GUI\n", command[0], command[1]);
            }
        }
        setPttMode(_ui->get_ui(), RX_ACTIVITY);
        scan_running = false;
        _ui->ui_shutdown();
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
bool UserInterfaceDevice_i::ui_continue_processing()
{
	omni_mutex_lock ll(ui_mutex);
	return ui_thread_started;
}

//-------------------------------------------------------------------------
/**
 * @details 	Main UI component thread.  Controls the state of the Rx and Tx
 * 				components.  If ptt is 0 (RX_ACTIVITY), a STOP control message
 * 				is sent to the Tx component, and once an ACK is received, a START
 * 				is sent to the Rx component.
 * 				If ptt is 1 (TX_ACTIVITY), a STOP control message is sent to the Rx
 * 				component, and once an ACK is received, the Tx frequency (from GUI)
 * 				is sent to the modem, followed by a START control message to the Tx
 * 				component
 * @param ui 	GUI information
 * @param mode 	Ptt mode
 *
 * @author 		Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::setPttMode(UInterface ui, int mode)
{
	UIControlPacket ui_pkt;
	UIControlPacket *rui_pkt;
	ModemPacket mod_pkt;

	// @todo: (09/30/2011) - Add mutexes around entire function call to prevent multiple
	// calls (trigger happy people) during clean up states

	memcpy((char *)&ui_pkt.ui, (char *)&ui, sizeof(UInterface));

	PortTypes::CharSequence_var I_out_0;
    PortTypes::CharSequence *ui_data_in(NULL);
	
	rui_pkt = NULL;

	switch (mode) {
	case RX_ACTIVITY:
		// send stop packet to TX
		INFO("Sending STOP pkt to Transmitter\n");
		ui_pkt.hdr = STOP;
		I_out_0 = new PortTypes::CharSequence(sizeof(UIControlPacket), sizeof(UIControlPacket), (CORBA::Char *)&ui_pkt, 0);
		tx_data_out->pushPacket(I_out_0);

		// Wait for ACK from TX
		do {
			tx_data_in->getData(ui_data_in);
			rui_pkt = (UIControlPacket *)ui_data_in->get_buffer();	
		} while (rui_pkt->hdr != ACK);
		tx_data_in->bufferEmptied();
		INFO("Got ACK pkt from Transmitter\n");

		// Send center frequency to modem
		mod_pkt.hdr = FREQ_CHANGE;
		mod_pkt.freq = (_ui->get_ui()).cfreq;
		INFO("Sending new freq to modem\n");
		I_out_0 = new PortTypes::CharSequence(sizeof(ModemPacket),sizeof(ModemPacket), (CORBA::Char *)&mod_pkt, 0);
		modem_data_out->pushPacket(I_out_0);

		// Send buffer flush to Modem
		mod_pkt.hdr = BS_OUTPUT_FLUSH;
		INFO("Sending buffer flush to modem\n");
		I_out_0 = new PortTypes::CharSequence(sizeof(ModemPacket),sizeof(ModemPacket), (CORBA::Char *)&mod_pkt, 0);
		modem_data_out->pushPacket(I_out_0);

		// send start packet to RX
		INFO("Sending START pkt to Receiver\n");
		ui_pkt.hdr = START;
		I_out_0 = new PortTypes::CharSequence(sizeof(UIControlPacket), sizeof(UIControlPacket), (CORBA::Char *)&ui_pkt, 0);
		rx_data_out->pushPacket(I_out_0);
			// For scanning purposes
		rx_running = true;
		//sem_post(&sem_rx_running);
		break;
	case TX_ACTIVITY:
		// send stop packet to RX
		INFO("Sending STOP pkt to Receiver\n");
		ui_pkt.hdr = STOP;
		I_out_0 = new PortTypes::CharSequence(sizeof(UIControlPacket), sizeof(UIControlPacket), (CORBA::Char *)&ui_pkt, 0);
		rx_data_out->pushPacket(I_out_0);

		// Wait for ACK from RX
		sem_wait(&receiver_pkt_stop);
		INFO("Got ACK pkt from Receiver\n");
		rx_running = false;

		// Send center frequency to modem
		mod_pkt.hdr = FREQ_CHANGE;
		mod_pkt.freq = (_ui->get_ui()).cfreq;
		INFO("Sending new freq to modem\n");
		I_out_0 = new PortTypes::CharSequence(sizeof(ModemPacket),sizeof(ModemPacket), (CORBA::Char *)&mod_pkt, 0);
		modem_data_out->pushPacket(I_out_0);

		// Send start to TX
		INFO("Sending START to Tx component\n");
		ui_pkt.hdr = START;
		I_out_0 = new PortTypes::CharSequence(sizeof(UIControlPacket), sizeof(UIControlPacket), (CORBA::Char *)&ui_pkt, 0);
		tx_data_out->pushPacket(I_out_0);

		break;
	default:
		INFO("Unknown Mode!\n");
		break;
	}
}

//-------------------------------------------------------------------------
/**
 * @details	Main scanning thread.  Once GUI sends a scan requested, the
 * 			thread scans through the array of frequencies until a SCAN_HALT
 * 			is received.  Once a halt is received it waits on that channel
 * 			until a the user starts scanning once again from the GUI
 *
 * @author  Justyn Bell <jwbell@ucsd.edu>
 */
//-------------------------------------------------------------------------
void UserInterfaceDevice_i::scanning_ProcessData()
{
    DEBUG(3, UserInterfaceDevice, "ProcessData() invoked");

	PortTypes::CharSequence_var buffer_var;

    UIControlPacket ui_pkt;
	ModemPacket mod_pkt;
	
	mod_pkt.hdr = FREQ_CHANGE;

    short numf, i;
    double freqs[15];

    while(continue_processing())
    {
    	sem_wait(&sem_rx_running);

    	// Get number of frequencies to scan, and put the
    	//   relevant frequencies in freqs
    	numf = _ui->get_freqs(freqs); i = 0;

    	while(rx_running && scan_running)
    	{
    		// Send frequency change to RX device
    		DBG("Sending frequency change to modem device\n");

			mod_pkt.freq = freqs[i];

			buffer_var = new PortTypes::CharSequence(sizeof(ModemPacket), sizeof(ModemPacket), (CORBA::Char *)&mod_pkt, 0);
			modem_data_out->pushPacket(buffer_var);

			
			// Update local UI object with current frequency as well as global current frequency for GUI
			current_cfreq = freqs[i];
    		ui_pkt.ui.cfreq = freqs[i];
    		// Update global UI with local UI object
    		_ui->ui_update(ui_pkt.ui);
    		// Update GUI information with local UI object
    		_ui->r_ui_update(ui_pkt.ui);

    		sleep(SCAN_PERIOD);
    		// Only scan the valid frequencies from GUI, not all 15
    		if (++i > numf-1) i = 0;
    	}
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
bool UserInterfaceDevice_i::scanning_continue_processing()
{
	omni_mutex_lock lll(scanning_processing_mutex);
	return scanning_thread_started;
}


