/**
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 */

#include <iostream>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

//#include "p25/util/logger.h"
#include "p25/config.h"
#include "ui.h"

extern "C" {
    #include "p25/comm/tcplib.h"
}

const char id[] = "UI";

#define INFO(fmt, args...)							\
	fprintf(stdout, "%20s: [info] %s() " fmt, id, __FUNCTION__, ##args);

#define ERROR(fmt, args...)							\
	fprintf(stderr, "%20s: [error] %s() " fmt, id, __FUNCTION__, ##args);

//-------------------------------------------------------------------------
/**
 * @details		Reverse byte endian of a double variable
 * @param d		Input double variable
 * @return		Byte endian-reversed double variable
 * @author		Jeffrey Cuenco <jcuenco@ucsd.edu>
 */
//-------------------------------------------------------------------------
double swap(double d)
{
   double a;
   unsigned char *dst = (unsigned char *)&a;
   unsigned char *src = (unsigned char *)&d;

   dst[0] = src[7];
   dst[1] = src[6];
   dst[2] = src[5];
   dst[3] = src[4];
   dst[4] = src[3];
   dst[5] = src[2];
   dst[6] = src[1];
   dst[7] = src[0];

   return a;
}

/**
 * @details
 */
UI::UI(void)
{
}

/**
 * @details
 */
UI::~UI(void)
{
}

/**
 * @details Reverse byte endians of P25 variables, and update private User Interface struct.
 */
void UI::ui_update(UInterface ui_cntrl)
{
	ui.cfreq = swap(ui_cntrl.cfreq);
    ui.nac = ntohs(ui_cntrl.nac);
    ui.tgId = ntohs(ui_cntrl.tgId);
    ui.keyId = ntohs(ui_cntrl.keyId);
    ui.algId = ui_cntrl.algId;
    ui.mfId = ui_cntrl.mfId;
    ui.ebit = ui_cntrl.ebit;
    memcpy(ui.unitId,ui_cntrl.unitId,sizeof(ui.unitId));
    memcpy(ui.dstId,ui_cntrl.dstId,sizeof(ui.dstId));
    memcpy(ui.mi,ui_cntrl.mi,sizeof(ui.mi));
    ui.ptt = ui_cntrl.ptt;
	ui.rxEncState = ui_cntrl.rxEncState;
}

/**
 * @details Reverse byte endians of scanning variables, and update private Scanning struct.
 * @param s_ui_cntrl 	Received scanning struct.
 */
void UI::scan_ui_update(SUInterface s_ui_cntrl)
{
	double tmp_freqs[15];

	// As of now, we don't really do anything with the state, since we assume when we get a scan struct
	//	from the GUI, it always means start scanning.
	s_ui.state = s_ui_cntrl.state;
	s_ui.snum = ntohs(s_ui_cntrl.snum);

	if (s_ui.snum > 15)
		ERROR("The number of frequencies to scan has exceeded maximum.\n");

	memcpy(tmp_freqs,s_ui_cntrl.freqs,sizeof(s_ui.freqs));

	for (int i = 0; i < s_ui.snum; i++)
		s_ui.freqs[i] = swap(tmp_freqs[i]);
	for (int j = s_ui.snum; j < 15; j++)
		s_ui.freqs[j] = 0;
}

/**
 * @details Gets a list of relevant frequencies to scan from the GUI. The list can only contain 15 frequencies
 * 				so the rest of the list is filled with 0, and the number of elements of the list filled with 
 *				valid frequencies is returned.
 * @param buf	Contains list of frequencies to be scanned.
 * @return 		Returns the number of valid frequencies received from the GUI to scan.
 */
short UI::get_freqs(double *buf)
{
	short i = s_ui.snum;
	for (int j = 0; j < i; j++)
		buf[j] = s_ui.freqs[j];
	return i;
}

/** 
 * @details Updates GUI P25 information. Fills values of sctruct, then sends user interface struct to GUI preceeded by 
 *				a p25 header via an established TCP protocol connection.
 * @param r_ui User Interface struct that contains data used to update GUI.
 */
void UI::r_ui_update(UInterface r_ui)
{
    char buf[sizeof(UInterface) + 2];
    UInterface r_ui_cntrl;

	// P25 header.  Used for the GUI to validate this is indeed a User Interface struct packet
    buf[0] = 0x0A;
    buf[1] = 0xA0;

	r_ui_cntrl.cfreq = swap(r_ui.cfreq);
    r_ui_cntrl.nac = htons(r_ui.nac);
    r_ui_cntrl.tgId = htons(r_ui.tgId);
    r_ui_cntrl.keyId = htons(r_ui.keyId);
    r_ui_cntrl.algId = r_ui.algId;
    r_ui_cntrl.mfId = r_ui.mfId;
    r_ui_cntrl.ebit = r_ui.ebit;
    memcpy(r_ui_cntrl.unitId,r_ui.unitId,sizeof(r_ui_cntrl.unitId));
    memcpy(r_ui_cntrl.dstId,r_ui.dstId,sizeof(r_ui_cntrl.dstId));
    memcpy(r_ui_cntrl.mi,r_ui.mi,sizeof(r_ui_cntrl.mi));
    r_ui_cntrl.ptt = r_ui.ptt;
	r_ui_cntrl.rxEncState = r_ui.rxEncState;

	memcpy(buf+2,(char *)&r_ui_cntrl,sizeof(UInterface));
    send_ui_data(buf,2+sizeof(UInterface));
}

/**
 * @details Updates GUI scan information. Similar to the r_ui_update function, the information is preceeded by a p25
 *				scan header and data is sent via a TCP connection to the GUI.  In this application, only the header and
 *				one byte is sent to the GUI since it is assumed when the GUI receives a scan packet, it always means
 *				a call has been received and the scanning should stop.
 */
void UI::stop_gui_scan()
{
	char buf[3];

	// P25 scan header. User for the GUI to validate this is indeed a Scan struct packet received from the OSSIE
	//		User Interface side.
	buf[0] = 0x0B;
	buf[1] = 0xB0;

	buf[2] = G_SCAN_HALT;
	send_ui_data(buf,sizeof(buf));
}

/**
 * @details Starts a TCP server connection with the GUI and waits until a connection is established.
 * @see ui_connected.
 */
void UI::ui_initialize(int ui_port)
{
    memset((char *)&ui,0,sizeof(UInterface)-1);
    ui_connFd = 0;
    ui_sock = 0;
	
    // Wait for ui connection
    // Don't start TX or RX until a User Interface is detected
    ui_sock = start_tcp_server(ui_port);
    INFO("Waiting for UI connection on port %d [sock = %d]\n",ui_port,ui_sock);
    do {
        ui_connFd = get_tcp_server_connection(ui_sock);
        usleep(10);
    } while (!ui_connFd);
	
    ui_connected = 1;
	INFO("UI Connected\n");
}

/**
 * @brief Terminates connection with the GUI and updates state variables.
 * @see ui_connected.
 */
void UI::ui_shutdown()
{
    close_tcp_server_connection(ui_connFd);
    stop_tcp_server(ui_sock);
    ui_connected = 0;
 //   set_rc_mode(RX_MODE);
}

/** 
 * @details Receives GUI data via established TCP connection. 
 * @param buf Buffer which contains data read from TCP connection.
 * @len Length of data to be read in bytes into the buffer
 * @return Returns the actual number of bytes read upon success, returns negative upon TCP error.
 */
int UI::receive_ui_data(char *buf, int len) 
{
    int res;

    res =  read_tcp_server_stream(ui_connFd,buf,len);
    if (res <= 0) ui_connected = 0;
    return res;
}

/**
 * @details Sends GUI information via established TCP connection.
 * @param buf Buffer which contains data to be sent to TCP connection.
 * @param len Length of data in bytes to be sent to TCP connection.
 */
void UI::send_ui_data(char *buf, int len)
{
	if (ui_connected) {
        send_tcp_client_data(ui_connFd,buf,len);
    }
}

/**
 * @details
 */
UInterface UI::get_ui(void)
{
	return ui;
}

/**
 * @details
 */
SUInterface UI::get_s_ui(void)
{
	return s_ui;
}

/**
 * @details
 */
int UI::get_ui_connected(void)
{
	return ui_connected;
}
