//-------------------------------------------------------------------------
/**
 * @file 	controller.h
 * @brief	Controller class definition
 *
 * @author	Cal(it)2
 * @author  	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
//-------------------------------------------------------------------------

#ifndef _CONTROLLER_H
#define _CONTROLLER_H

#include "p25/config.h"

/**
 * @brief	Controller class definitions.
 */
class Controller
{
private:
	/**
	 * @brief 	Graphical User Interface communication port.
	 * @details Current implementation uses a TCP socket.
	 */
	int ui_port;

	/**
	 * @brief 	VC55 communication port location.
	 * @details Current implementation uses a serial port.
	 *          Doubles as a path to a TX file in filetest mode.
	 */
	char* vc_sport;

	/**
	 * @brief 	Filetest mode RX file location.
	 * @details Path to file that receiver writes to in filetest mode.
	 */
	char* vc_rxfile;

	/**
	 * @brief Keeps track of the current radio activity.
	 */
	char activity;
	
	/**
	 * @brief Keeps track of the current radio state in any mode
	 */
	char state;

	/**
	 * @brief Keeps track of the current radio mode
	 */
	char mode;

public:
	/**
	 * @brief Constructor for Controller Class.
	 */
	Controller(int _ui_port, char* _vc_sport, char* _vc_rxfile, int _r_mode);

	/**
	 * @brief Gets the current activity of the P25 radio system.
	 */
	int get_radio_activity();

	/**
	 * @brief Sets the current activity of the P25 radio system.
	 */
	void set_radio_activity(int _activity);

	/**
	 * @brief Gets the current state of the P25 radio system.
	 */
	int get_radio_state();

	/**
	 * @brief Sets the current state of the P25 radio system.
	 */
	void set_radio_state(int _state);

	/**
	 * @brief Gets the current mode of the P25 radio system.
	 */
	int get_radio_mode();

	/**
	 * @brief Sets the current mode of the P25 radio system.
	 */
	void set_radio_mode(int _mode);

	/**
	 * @brief Sets up the initial state of the radio system
	 */
	void initialize();

	/**
	 * @brief Handles termination signals from the user
	 */
	static void shutdown(int signum);

	/**
	 * @brief Entry point of the P25 GPP program
	 */
	void start();
};

#endif

