#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "p25/format/byte_fifo.h"
#include "p25/util/logger.h"

/** 
 * @details 
 *
 * @param size	Expected size in bytes of the fifo, the allocated
 * 		 fifo size might be larger
 *
 * @return 		Pointer to a byte_fifo structure, or NULL if failed
 */		 
struct byte_fifo *bf_alloc(int size)
{
	int nbytes;
	struct byte_fifo *bf;

	if (size == 0)
		size = BF_SIZE_DFT;
	if (size > BF_SIZE_MAX)
		size = BF_SIZE_MAX;

	//nbytes = size/8 + 1;
	nbytes = size + 1;
	bf = (struct byte_fifo *) malloc(sizeof(struct byte_fifo) + nbytes);
	if (bf == NULL) {
		ERROR("alloc byte fifo failed, size = %d\n",size);
		return NULL;
	}

	bf->byte_size = nbytes;
	bf->head_index = 0;
	bf->tail_index = bf->byte_size - 1;
	bf->bytes = (char *) (bf + 1);

	INFO("allocated fifo, size = %d\n",size);
	return bf;
}

/** 
 * @details
 *
 * @param bf The byte fifo to be freed
 */		 
void bf_free(struct byte_fifo *bf)
{
	free(bf);
}

/* 
 * @details 
 *
 * @param bf The byte fifo to be reset
 */		 
void bf_reset(struct byte_fifo *bf)
{
	bf->head_index = 0;
	bf->tail_index = bf->byte_size - 1;
}

#define incr_index_offset(idx, nbytes)	\
{					\
	(idx) = ((idx)+1) % (nbytes);	\
}

/**
 * @details 
 *
 * @return Return 1 if byte fifo is empty, otherwise
 * 			  return 0
 */
int bf_empty(struct byte_fifo *bf)
{
	int index;

	index = bf->tail_index;

	incr_index_offset(index, bf->byte_size);
	return index == bf->head_index;
}

/**
 * @details
 *
 * @return Return 1 if byte fifo is full, otherwise
 * 			 return 0
 */
int bf_full(struct byte_fifo *bf)
{
	return bf->tail_index == bf->head_index;
}

/**
 * @details
 *
 * @return	Return the number of bytes in a byte fifo
 */
int bf_size(struct byte_fifo *bf)
{
	int size;
	size = (bf->head_index - bf->tail_index) -1;
	return (size >= 0) ? size : size + bf->byte_size; 
}

/**
 * @details
 *
 * @ return	Return the number of bytes can be put in the byte fifo
 */
int bf_space(struct byte_fifo *bf)
{
	return (bf->byte_size - 1) - bf_size(bf);
}

/**
 * @details
 */
void __bf_put(struct byte_fifo *bf, char b)
{
	bf->bytes[bf->head_index] = b;
	incr_index_offset(bf->head_index, bf->byte_size);
}

/**
 * @details
 */
char __bf_get(struct byte_fifo *bf)
{
	incr_index_offset(bf->tail_index, bf->byte_size);
	return bf->bytes[bf->tail_index];
}

/** 
 * @details
 *
 * @param bf   	The destination byte fifo
 * @param bytes Bytes array
 * @param n 	Number of bytes to be put into fifo
 *
 * @return		The number of bytes put into the fifo
 */
int bf_put_n(struct byte_fifo *bf, char *bytes, int n)
{
	int i;

	for (i = 0; i < n && !bf_full(bf); i++)
		__bf_put(bf, bytes[i]);
	if (i != n)
		ERROR("n = %d, put = %d\n", n, i);
	return i;	
}

/** 
 * @details
 *
 * @param bf   	The source byte fifo
 * @param bytes Bytes array
 * @param n 	Number of bytes to get from fifo
 * 
 * @return 		The number of bytes gotten from the fifo
 */
int bf_get_n(struct byte_fifo *bf, char *bytes, int n)
{
	int i;

	for (i = 0; i < n && !bf_empty(bf); i++)
		bytes[i] = __bf_get(bf);
	if (i != n)
		ERROR("n = %d, got = %d\n", n, i);
	return i;	
}

/*
 * @details Dump the contents of a byte fifo to stdout
 */
void bf_dump(struct byte_fifo *bf)
{
	struct byte_fifo bf1;
	unsigned char b;
	int id;

	memcpy(&bf1, bf, sizeof(bf1));
	printf("size=%d, space=%d", bf_size(bf), bf_space(bf));
	printf(" bs=%d, hi=%d, ti=%d", bf->byte_size,
		bf->head_index, bf->tail_index);
	id = 0;
	while (!bf_empty(&bf1)) {
		if (id % 16 == 0)
			printf("\n%06d  ", id);
		else if (id % 8 ==0)
			printf("  ");

		b = __bf_get(&bf1);
		printf("%02x  ", b);
		id++;
	}
	printf("\n");
}
