/****************************************************************************
Copyright 2007 Virginia Polytechnic Institute and State University
This file is part of the OSSIE __COMP_NAME__.
OSSIE __COMP_NAME__ is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
"(at your option) any later version
OSSIE __COMP_NAME__ is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with OSSIE __COMP_NAME__; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

/**
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 */

#ifndef P25_RECEIVER_IMPL_H
#define P25_RECEIVER_IMPL_H

#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#include "ossie/cf.h"
#include "ossie/PortTypes.h"
#include "ossie/Resource_impl.h"
#include "ossie/debug.h"

#include "standardinterfaces/realChar.h"
#include "standardinterfaces/realChar_u.h"
#include "standardinterfaces/realChar_p.h"

/* user includes */
#include "p25/config.h"
#include "p25/controller.h"
#include "p25/format/packets.h"

extern "C" {
    #include "p25/format/bit_fifo.h"
}

/** \brief
 *
 *
 */
class p25_Receiver_i : public virtual Resource_impl
{
  public:
    /** @brief Initializing constructor */
    p25_Receiver_i(const char *uuid, omni_condition *sem);

    /** @brief Destructor */
    ~p25_Receiver_i(void);


    /** @brief Static function for omni thread */
    static void Run( void * data );

    /** @brief Static function for omni thread */
    static void ui_Run( void * data );

    /** @brief Static function for omni thread */
    static void rx_Run( void * data );

    /** @brief */
    void start() throw (CF::Resource::StartError, CORBA::SystemException);

    /** @brief */
    void stop() throw (CF::Resource::StopError, CORBA::SystemException);

    /** @brief */
    CORBA::Object_ptr getPort( const char* portName )
        throw (CF::PortSupplier::UnknownPort, CORBA::SystemException);

    /** @brief */
    void releaseObject() throw (CF::LifeCycle::ReleaseError, CORBA::SystemException);

    /** @brief */
    void initialize() throw (CF::LifeCycle::InitializeError, CORBA::SystemException);

	/** @brief Query properties from .prf.xml */
	void query(CF::Properties&)
		throw (CORBA::SystemException,
				CF::UnknownProperties);

    /** @brief Configures properties read from .prf.xml */
    void configure(const CF::Properties&)
        throw (CORBA::SystemException,
            CF::PropertySet::InvalidConfiguration,
            CF::PropertySet::PartialConfiguration);

	/** @brief Main Rx processing thread*/
	void run_rx();


  private:
    /** @brief Disallow default constructor */
    p25_Receiver_i();

    /** @brief Disallow copy constructor */
    p25_Receiver_i(p25_Receiver_i&);

    /// Main signal processing method
    void ProcessData();
	bool continue_processing();
	volatile bool thread_started;
	omni_mutex processing_mutex;

    /// UI signal processing method
    void ui_ProcessData();
	bool ui_continue_processing();
	volatile bool ui_thread_started;
	omni_mutex ui_processing_mutex;

	volatile bool rx_thread_started;
	omni_mutex rx_processing_mutex;
   
    omni_condition *component_running;  ///< for component shutdown
    omni_thread *processing_thread;     ///< for component writer function
    omni_thread *ui_processing_thread;     ///< for component writer function
    omni_thread *rx_processing_thread;     ///< for component writer function
    
    // list components provides and uses ports
        standardInterfaces_i::realChar_p *modem_data;
        standardInterfaces_i::realChar_u *voc_out_data;
        standardInterfaces_i::realChar_p *voc_in_data;
        standardInterfaces_i::realChar_u *ui_dataOut;
        standardInterfaces_i::realChar_p *ui_dataIn;
        standardInterfaces_i::realChar_u *enc_dataOut;
        standardInterfaces_i::realChar_p *enc_dataIn;

	/* user includes */

	/**
	 * @details Variable shared for synchronization of User Interface changes.
	 */
	UInterface m_ui;

	/**
	 * @details Variable shared for synchronization of User Interface changes.
	 */
	UInterface mr_ui;


	/**
	 * @brief 	Tracks the frame number for the current frame
	 *  		being processed by the Receiver
	 */
	int frame_id;

	/**
	 * @brief 	Tracks the number of bits pulled from the receiver bit stream
	 * @details	It is used in computing the locations of the Status bits (SS)
	 * 			in the receiver bit stream.
	 * @see	Receiver::get_bits
	 */
	int rx_bit_count; //used to be static


	/** @brief */
	int frameIsMine(void);

	/** @brief */
	void play_vcw(char* vcw, int len);

	/** @brief */
	int nid_dec(char* nid_cw, char* duid);

	/** @brief */
	void lsd_dec(char* lsd_cw, char* lsd);

	/** @brief */
	int is_lc_protected(struct bit_fifo *lc);

	/** @brief */
	void lc_deformat(struct bit_fifo *lc);

	/** @brief */
	void es_deformat(struct bit_fifo *es);

	/** @brief */
	int lcw_dec(char* lcw_cw, char *lc);

	/** @brief */
	int esw_dec(char* esw_cw, char *es);

	/** @brief */
	int hdu_dec(char* hdu, char *mi, char *algid, char *kid, char *tgid);

	/** @brief */
	void ldu_cw_split(char* ldu, int len, char* vc, char* lsd, char* lc);

	/** @brief */
	void ldu_dec(char* ldu, char *vc, char *lc_es, char *lsd, char type);

	/** @brief */
	void ldu2_dec(char* ldu2);

	/** @brief */
	void tdu_dec(char* tdu);

	/** @brief */
	int tdu_lcw_dec(char* tdu);

	/** @brief */
	int check_fs(char* bits);

	/** @brief gets the specified number of bits from the receiver bit stream */
	int get_bits(char* bits, int len);

	/** @brief */
	void rx_initialize();

	/** @brief */
	void rx_shutdown();

	/** @brief */
	Controller* mController;

	/** @brief RX buffer variables */
	struct bit_fifo 	*rx_bf;
	pthread_mutex_t		vc_rx_bf_m;
	sem_t 				m_bf_full;
	sem_t				m_bf_empty;

	/** @brief Rx state machine variables */
	sem_t 	sem_ui_rx_ack;
	sem_t 	sem_run_rx;
	bool crypto_rx_ack;
	bool vocoder_rx_ack;
	bool rx_stop_requested;

	/** @brief Encrypted or non-encrypted frame mode */
	bool encrypted_frame_mode;

	/** @brief Gets data from Rx buffer */
	int bs_input(char *buf, unsigned int len);

	/** @brief Sends data from Rx component to Cryptography component */
	void sendbuf_rx_crypto(short hdr,char *data = NULL, int len = 0);

	/** @brief Receives data from Cryptography component */
	int recvbuf_rx_crypto(char *buf, int len);

	/** @brief Sends control message to UI component */
	void sendstate_rx_ui(short hdr);
	
	/** @brief Sends data from Rx component to Vocoder component */
	void sendbuf_rx_vox(short hdr, char *data = NULL, int len = 0);
};
#endif
