/****************************************************************************
Copyright 2007 Virginia Polytechnic Institute and State University
This file is part of the OSSIE __COMP_NAME__.
OSSIE __COMP_NAME__ is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
"(at your option) any later version
OSSIE __COMP_NAME__ is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with OSSIE __COMP_NAME__; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

/**
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 */

#include <string>
#include <iostream>
#include "p25_Transmitter.h"

/* user includes */
#include <stdio.h>
#include <pthread.h>
#include <semaphore.h>
#include <fcntl.h>
#include "p25/config.h"
#include "p25/controller.h"
#include "p25/util/logger.h"
#include "p25/format/packets.h"

extern "C" {
#include "p25/ecc/golay.h"
#include "p25/ecc/cyclic.h"
#include "p25/ecc/ham.h"
#include "p25/ecc/rs.h"
#include "p25/format/bit_fifo.h"
#include "p25/format/bit_manipulate.h"
}


using namespace P25;

//-------------------------------------------------------------------------
/**
 * @details Transmitter class constructor for OSSIE implementation
 *
 * @author
 */
//-------------------------------------------------------------------------
p25_Transmitter_i::p25_Transmitter_i(const char *uuid, omni_condition *condition) :
    Resource_impl(uuid), component_running(condition)
{
    modem_data = new standardInterfaces_i::realChar_u("mod_out");
	voc_in_data = new standardInterfaces_i::realChar_p("voc_in");
    voc_out_data = new standardInterfaces_i::realChar_u("voc_out");
    ui_dataIn = new standardInterfaces_i::realChar_p("ui_in");
    ui_dataOut = new standardInterfaces_i::realChar_u("ui_out");
    enc_dataOut = new standardInterfaces_i::realChar_u("enc_out");
    enc_dataIn = new standardInterfaces_i::realChar_p("enc_in");
//    start();
}

//-------------------------------------------------------------------------
/**
 * @details Transmitter class destructor for OSSIE implementation
 *
 * @author
 */
//-------------------------------------------------------------------------
p25_Transmitter_i::~p25_Transmitter_i(void)
{
    delete voc_in_data;
    delete voc_out_data;
    delete modem_data;
    delete ui_dataIn;
    delete ui_dataOut;
    delete enc_dataOut;
    delete enc_dataIn;
}

//-------------------------------------------------------------------------
/**
 * @details Static function for omni thread
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::Run( void * data )
{
    ((p25_Transmitter_i*)data)->ProcessData();
}

//-------------------------------------------------------------------------
/**
 * @details Static function for omni thread
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::ui_Run( void * data )
{
    ((p25_Transmitter_i*)data)->ui_ProcessData();
}

//-------------------------------------------------------------------------
/**
 * @details Static function for omni thread
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::tx_Run( void * data )
{
    ((p25_Transmitter_i*)data)->run();
}

//-------------------------------------------------------------------------
/**
 * @details Get port data
 *
 * @author
 */
//-------------------------------------------------------------------------
CORBA::Object_ptr p25_Transmitter_i::getPort( const char* portName ) throw (
    CORBA::SystemException, CF::PortSupplier::UnknownPort)
{
    DEBUG(3, p25_Transmitter, "getPort() invoked with " << portName)

    CORBA::Object_var p;

    p = voc_in_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = voc_out_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = modem_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = ui_dataIn->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = ui_dataOut->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = enc_dataOut->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = enc_dataIn->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    /*exception*/
    throw CF::PortSupplier::UnknownPort();
}

//-------------------------------------------------------------------------
/**
 * @details	Instantiates Controller, sets stack size, allocates
 *			buffers and starts all threads for Tx component
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::start() throw (CORBA::SystemException,
    CF::Resource::StartError)
{
    DEBUG(3, p25_Transmitter, "start() invoked");

	// Instantiate Controller
	mController = new Controller(0, NULL, NULL, 2);
	mController->start();

	tx_bf = bf_alloc(PLAY_BACK_BUF_SIZE);

	omni_thread::stacksize(2*1024*1024);

	omni_mutex_lock  l(processing_mutex);
	if( false == thread_started )
	{
		thread_started = true;
		// Create the thread for the writer's processing function
		processing_thread = new omni_thread(Run, (void *) this, omni_thread::PRIORITY_HIGH);

		// Start the thread containing the writer's processing function
		processing_thread->start();

		INFO("Started tx_in port pushPacket processing thread.\n");
	}

	omni_mutex_lock  ll(ui_processing_mutex);
	if( false == ui_thread_started )
	{
		ui_thread_started = true;
		// Create the thread for the writer's processing function
		ui_processing_thread = new omni_thread(ui_Run, (void *) this, omni_thread::PRIORITY_NORMAL);

		// Start the thread containing the writer's processing function
		ui_processing_thread->start();

		INFO("Started ui_in port pushPacket processing thread.\n");
	}

//	omni_mutex_lock  lll(tx_processing_mutex);
	if( false == tx_thread_started )
	{
		tx_thread_started = true;
		// Create the thread for the writer's processing function
		tx_processing_thread = new omni_thread(tx_Run, (void *) this, omni_thread::PRIORITY_HIGH);

		// Start the thread containing the writer's processing function
		tx_processing_thread->start();

		INFO("Started TX main processing thread.\n");
	}
}

//-------------------------------------------------------------------------
/**
 * @details	Stops all threads, frees buffers and cleans up Controller
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::stop() throw (CORBA::SystemException, CF::Resource::StopError)
{
    DEBUG(3, p25_Transmitter, "stop() invoked")
	omni_mutex_lock l(processing_mutex);
	thread_started = false;

	omni_mutex_lock ll(ui_processing_mutex);
	ui_thread_started = false;

//	omni_mutex_lock lll(tx_processing_mutex);
	tx_thread_started = false;

	bf_free(tx_bf);

	delete mController;
}


void p25_Transmitter_i::releaseObject() throw (CORBA::SystemException,
    CF::LifeCycle::ReleaseError)
{
    DEBUG(3, p25_Transmitter, "releaseObject() invoked")

    component_running->signal();
}

//-------------------------------------------------------------------------
/**
 * @details Initializes all semaphores and mutexes for Tx component, and puts
 * 			the component in an initialized state
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::initialize() throw (CF::LifeCycle::InitializeError,
    CORBA::SystemException)
{
	int result;
	CF::StringSequence initErrSeq;

    DEBUG(3, p25_Transmitter, "initialize() invoked");

	result = pthread_mutex_init(&vc_tx_bf_m, NULL);
	if (result != 0) {
		initErrSeq.length(1);
		initErrSeq[0] = CORBA::string_dup("TX pthread_mutex_init() failure\n");
		throw CF::LifeCycle::InitializeError(initErrSeq);
	}

	sem_init(&m_bf_full, 0, 0);
	sem_init(&m_bf_empty, 0, 0);
	sem_init(&sem_run_tx, 0, 0);
    sem_init(&sem_ui_tx_ack, 0, 0);
    sem_init(&sem_voc_ack, 0, 0);
    sem_init(&sem_enc_ack, 0, 0);

	tx_stop_requested = false;

	//Initialize Reed-Solomon encoders
	init_rs();
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::query( CF::Properties & configProperties ) throw (CORBA::SystemException, CF::UnknownProperties)
{
	if( configProperties.length() == 0 )
	{
		configProperties.length( propertySet.length() );
		for( int i = 0; i < propertySet.length(); i++ )
		{
			configProperties[i].id = CORBA::string_dup( propertySet[i].id );
			configProperties[i].value = propertySet[i].value;
		}
		return;
	} else {
		for( int i = 0; i < configProperties.length(); i++ ) {
			for( int j = 0; j < propertySet.length(); j++ ) {
				if( strcmp(configProperties[i].id, propertySet[j].id) == 0 ) {
					configProperties[i].value = propertySet[j].value;
				}
			}
		}
	} // end if-else
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------8
void p25_Transmitter_i::configure(const CF::Properties& props)
throw (CORBA::SystemException,
    CF::PropertySet::InvalidConfiguration,
    CF::PropertySet::PartialConfiguration)
{
    DEBUG(3, p25_Transmitter, "configure() invoked")

    static int init = 0;
    if( init == 0 ) {
        if( props.length() == 0 ) {
            std::cout << "configure called with invalid props.length - " << props.length() << std::endl;
            return;
        }
        propertySet.length(props.length());
        for( int j=0; j < props.length(); j++ ) {
            propertySet[j].id = CORBA::string_dup(props[j].id);
            propertySet[j].value = props[j].value;
        }
        init = 1;
    }

    std::cout << "props length : " << props.length() << std::endl;

    for ( int i = 0; i <props.length(); i++)
    {
        std::cout << "Property id : " << props[i].id << std::endl;

    }
}

//-------------------------------------------------------------------------
/**
 * @details Reads packetized data in from Vocoder component.  If the packet
 * 			header indicates data, data is placed into the Tx component buffer
 * 			for later processing.
 * 			If the header indicates a Vocoder ACK message, a stop has already
 * 			been sent, and the Vocoder component has completed all processing
 *
 * @author		Jeffrey Cuenco <jcuenco@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::ProcessData()
{
    DEBUG(3, p25_Transmitter, "ProcessData() invoked")


    PortTypes::CharSequence *I_in_0(NULL);
    CORBA::UShort I_in_0_length;

	int res, len, bf_isfull, bufsize;

	VocoderPacket *v_pkt;

    while(continue_processing())
    {
        voc_in_data->getData(I_in_0);
        I_in_0_length = I_in_0->length();

        /*insert code here to do work*/
        if (I_in_0_length > 0) {
        	/* get buffer and buffer length */
        	v_pkt = (VocoderPacket *)(I_in_0->get_buffer());
			len = v_pkt->len;

			//INFO("!bf_isfull, continuing\n");

			//check header of vocoder packet
			switch (v_pkt->hdr) {
			case VOCODER_PKT_CNTRL_ACK:
				DBG("Received vocoder ack message\n");
				sem_post(&sem_voc_ack);
				DBG("Posted sem_voc_ack\n");
				break;

			case VOCODER_PKT_CNTRL_DATA:
				DBG("Got VOCODER_PKT_CNTRL_DATA from OveroVocoderDevice\n");

				// Block on semaphore if buffer is full or does not have room to receive
				// the complete payload.
				pthread_mutex_lock(&vc_tx_bf_m);
				bf_isfull = bf_full(tx_bf) || (len > bf_space(tx_bf));
				pthread_mutex_unlock(&vc_tx_bf_m);

				while (bf_isfull) {
					//INFO("bf_isfull, waiting\n");
					if (sem_wait(&m_bf_full) < 0) {
						ERROR("error in sem_wait\n");
						continue;
					}
					pthread_mutex_lock(&vc_tx_bf_m);
					bf_isfull = bf_full(tx_bf) || (len > bf_space(tx_bf));
					pthread_mutex_unlock(&vc_tx_bf_m);
				}

				// Put payload into buffer.
				pthread_mutex_lock(&vc_tx_bf_m);
				res = bf_put_n(tx_bf,v_pkt->data,len);
				DBG("Put %d into tx_buf from vocoder\n", res);
				bufsize = bf_size(tx_bf);
				pthread_mutex_unlock(&vc_tx_bf_m);

				// Notify the consumer that the buffer is no longer empty
				// provided a nonzero payload was inserted into the buffer.
				if (bufsize > 0) {
					//INFO("bf_isempty, posting\n");
					if (sem_post(&m_bf_empty) < 0) {
						ERROR("error in sem_post\n");
					}
				}
				break;

			default:
				ERROR("Got unexpected VocoderPacket header %x\n", v_pkt->hdr);
				break;
			}
		}
       voc_in_data->bufferEmptied();
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
bool p25_Transmitter_i::continue_processing()
{
	omni_mutex_lock l(processing_mutex);
	return thread_started;
}

//-------------------------------------------------------------------------
/**
 * @details Reads packetized control data from the UI component.  If the header indicates
 * 			START, the Tx component switches frequency to the received frequency
 * 			from the UI component, sends a start message to the Vocoder component,
 * 			switches into a state of active processing, and starts the main
 * 			processing thread by posting the sem_run_tx semaphore.
 * 			If the header indicates STOP, the Tx component switches to standby,
 * 			sends a stop message to the Vocoder component, notifies the main processing
 * 			thread that a stop has been received, and waits until an ACK message
 * 			has been received from both the Vocoder component and the main Tx
 * 			processing thread.  It then sends an ACK message back to the UI device
 * 			as a notification that all processing has been completed.
 *
 * @author	Justyn Bell <jwbell@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::ui_ProcessData()
{
    DEBUG(3, Transmitter, "ui_ProcessData() invoked");

    float old_cfreq = 0.0;
    char tmp_buf[16];
    UIControlPacket uicPkt;

    PortTypes::CharSequence_var buffer_var;
    PortTypes::CharSequence *I_in_1(NULL);
    CORBA::UShort I_in_1_length;

    while(ui_continue_processing())
    {
    	//INFO("ui_dataIn->getData(I_in_1)\n");
        ui_dataIn->getData(I_in_1);
        DBG("Got data from ui device\n");
        I_in_1_length = I_in_1->length();

        if (I_in_1_length == sizeof(UIControlPacket)) {
        	for (int i = 0; i < I_in_1_length; i++)
        		((char *)&m_ui)[i] = (*I_in_1)[i];


            old_cfreq = m_ui.ui.cfreq;

            switch(m_ui.hdr) {
            case START:
            	if (mController->get_radio_activity() == STANDBY_ACTIVITY) {
            		/*
					// If old value of carrier frequency is different than new
					// one, send the new frequency to modem.
					if (old_cfreq != m_ui.ui.cfreq) {
							DBG("Received new frequency change: %5.4f MHz\n",
									m_ui.ui.cfreq / 1e6);

							memcpy(tmp_buf, "NEW_FREQ", 8);
							memcpy(tmp_buf+8, &m_ui.ui.cfreq, 8);
							bs_output(tmp_buf, 16);
					}
					*/
					sendstate_tx_voc(VOCODER_PKT_CNTRL_START, m_ui.ui.algId);
					DBG("Sent start to vocoder device, with algid %x\n", m_ui.ui.algId);

					mController->set_radio_activity(TX_ACTIVITY);
					INFO("Switching Transmitter Device to TX_ACTIVITY\n");

					sem_post(&sem_run_tx);
					DBG("Post sem_run_tx\n");
            	}

				break;
            case STOP:
            	if (mController->get_radio_activity() == TX_ACTIVITY) {
					INFO("Switching Transmitter Device to STANDBY_ACTIVITY\n");
					mController->set_radio_activity(STANDBY_ACTIVITY);

					tx_stop_requested = true;

					DBG("Got TX_CNTRL_PKT_STOP.  Sending stop message to vocoder component\n");
					sendstate_tx_voc(VOCODER_PKT_CNTRL_STOP);

					DBG("Waiting on acks from voc and enc\n");
					//sem_wait(&sem_voc_ack);
					if (encrypted_frame_mode)
						sem_wait(&sem_enc_ack);

	            	sem_wait(&sem_ui_tx_ack);
					DBG("Vocoder and cryptography done, sending ack to ui device\n");
            	}

				// Send ack to UI
				uicPkt.hdr = ACK;
				buffer_var = new PortTypes::CharSequence(sizeof(uicPkt), sizeof(uicPkt), (CORBA::Char *)&uicPkt, 0);
				ui_dataOut->pushPacket(buffer_var);
				DBG("Sent ACK to UI\n");

        		break;
            default:
            	ERROR("Got unexpected packet header.  Value = %X\n", m_ui.hdr);
            	break;
            }

        } else
            	ERROR("I_in_1_length != sizeof(UIControlPacket");

        ui_dataIn->bufferEmptied();
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
bool p25_Transmitter_i::ui_continue_processing()
{
	omni_mutex_lock ll(ui_processing_mutex);
	return ui_thread_started;
}

//-------------------------------------------------------------------------
/**
 * @details	Encodes some constants based on current configuration data
 * 			in order to speed up P25 frame encoding
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::enc_speedup()
{
	struct bit_fifo *lc_tmp = bf_alloc(LCW_BLEN);

	/**
	 * @todo Switch TGID to 0x0000 for Unit to Unit transmission
	 * @date 09/15/2009
	 */
    nid_enc(ldu1_nid,DUID_LDU1,1);
    nid_enc(ldu2_nid,DUID_LDU2,1);

    if (memcmp(m_ui.ui.dstId, EVERYONE_UID, sizeof(m_ui.ui.dstId)) == 0) {
		INFO("using group voice\n");
    	lc_format(lc_tmp, LC_GRP_V_CH_USR);
    } else {
		INFO("using user -> user voice\n");
    	lc_format(lc_tmp, LC_UU_V_CH_USR);
    }

    bf_export(lc_tmp, lc, 72);
    bf_reset(lc_tmp);
    bf_import(lc_tmp, lc, 72);

    lcw_enc(lc_tmp);

    esw_enc();
}

//-------------------------------------------------------------------------
/**
 * @details	Formats the Link Control Word
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::lc_format(struct bit_fifo* lc, int lco)
{
    char bits[32];
    unsigned int i;

    switch (lco) {
		case LC_GRP_V_CH_USR:
			//if (encrypted_frame_mode) {
				// P=0, LCO = 0, SF = 0
			//	de2bi(0x80,bits,8);
			//	bf_put_n(lc,bits,8);
			//	INFO("Setting P = 1, encrypting LC\n");
			//} else {
				// P=0, LCO = 0, SF = 0
				de2bi(0x00,bits,8);
				bf_put_n(lc,bits,8);
				INFO("Setting P = 0, not encrypting LC\n");
			//}
			// MFID
			de2bi(m_ui.ui.mfId,bits,8);
			bf_put_n(lc,bits,8);
			// Service Options
			de2bi(0x0,bits,8);
			bf_put_n(lc,bits,8);
			// Reserved
			de2bi(0x0,bits,8);
			bf_put_n(lc,bits,8);
			// Group Address
			de2bi(m_ui.ui.tgId,bits,16);
			bf_put_n(lc,bits,16);
			// Source Address
			for (i = 0; i < sizeof(m_ui.ui.unitId); i++) {
				de2bi(m_ui.ui.unitId[i],bits,8);
				bf_put_n(lc,bits,8);
			}
			break;
		case LC_UU_V_CH_USR:
			//if (encrypted_frame_mode) {
				// P=1, LCO = 3, SF = 0
			//	de2bi(0x83,bits,8);
			//	bf_put_n(lc,bits,8);
			//	INFO("Setting P = 1, encrypting LC\n");
			//} else {
				//P=0, LCO = 3, SF = 0
				de2bi(0x03,bits,8);
				bf_put_n(lc,bits,8);
				INFO("Setting P = 0, not encrypting LC\n");
			//}	
			// MFID
			de2bi(m_ui.ui.mfId,bits,8);
			bf_put_n(lc,bits,8);
			// Service Options
			de2bi(0x0,bits,8);
			bf_put_n(lc,bits,8);
			// Target Address
			for (i = 0; i < sizeof(m_ui.ui.dstId); i++) {
				de2bi(m_ui.ui.dstId[i],bits,8);
				bf_put_n(lc,bits,8);
                INFO("destinationID[%d] = %d\n",i,m_ui.ui.dstId[i]);
			}
			// Source Address
			for (i = 0; i < sizeof(m_ui.ui.unitId); i++) {
				de2bi(m_ui.ui.unitId[i],bits,8);
				bf_put_n(lc,bits,8);
			}
			break;
		case LC_TELE_INT_V_CH_USR:
			break;
		case LC_SOURCE_ID_EXT:
			break;
		case LC_UU_V_CH_USR_EXT:
			break;
		case LC_CALL_TRM_CAN:
			break;
		case LC_STS_Q:
			break;
		case LC_STA_UPDT:
			break;
		case LC_MSG_UPDT:
			break;
		case LC_CALL_ALRT:
			break;
		case LC_EXT_FNCT_CMD:
			break;
		case LC_CH_ID_UPDT:
			break;
		default:
			break;
    }
}



//-------------------------------------------------------------------------
/**
 * @details	Encodes the Link Control Word
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::lcw_enc(struct bit_fifo *lc)
{
    unsigned int i;
    long encw;
    char bits[32];
    char enc_bits[24];

	// get bits and send to rs(24,12,13) encoder
    bf_get_n(lc,lcw,72);
    for (i = 0; i < 12; i++) {
        bits[i] = bi2de(lcw+(i*6),6);
    }

    reverse_bits(bits,12);
    encode_rs_24_12(bits,enc_bits);
    reverse_bits(enc_bits,24);

    for (i = 0; i < 24; i++) {
        de2bi(enc_bits[i],lcw+(i*6),6);
    }

    // get rs and send to hamming encoder
    for (i = 0; i < 24; i++) {
        encw = hamming_10_encode(bi2de(lcw+(i*6),6));
        de2bi(encw & 0x3ff, bits,10);
        bf_put_n(lc,bits,10);
    }
    bf_get_n(lc,lcw,LCW_BLEN);
}

//-------------------------------------------------------------------------
/**
 * @details	Encodes the Encryption Sync Word
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::esw_enc()
{
    unsigned int i;
    long encw;
    char bits[32];
    struct bit_fifo *es;
    char enc_bits[24];

    es = bf_alloc(ESW_BLEN);

    for (i = 0; i < sizeof(m_ui.ui.mi); i++) {
        de2bi(m_ui.ui.mi[i],bits,8);
        bf_put_n(es,bits,8);
    }

    de2bi(m_ui.ui.algId,bits,8);
    bf_put_n(es,bits,8);

    de2bi(m_ui.ui.keyId,bits,16);
    bf_put_n(es,bits,16);

    // get bits and send to rs(24,16,9) encoder
    bf_get_n(es,esw,96);
    for (i = 0; i < 16; i++) {
        bits[i] = bi2de(esw+(i*6),6);
    }

    reverse_bits(bits,16);
    encode_rs_24_16(bits,enc_bits);
    reverse_bits(enc_bits,24);

    for (i = 0; i < 24; i++) {
        de2bi(enc_bits[i],esw+(i*6),6);
    }

    // get rs and send to hamming encoder
    for (i = 0; i < 24; i++) {
        encw = hamming_10_encode(bi2de(esw+(i*6),6));
        de2bi(encw & 0x3ff, bits,10);
        bf_put_n(es,bits,10);
    }
    bf_get_n(es,esw,ESW_BLEN);
    bf_free(es);
}

//-------------------------------------------------------------------------
/**
 * @details	Encodes the Network ID (NID)
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::nid_enc(char* nid, char duid, char par_bit)
{
    struct bit_fifo *ni;

    ni = bf_alloc(NID_BLEN);
	memset(nid,0,63);

    de2bi(m_ui.ui.nac,nid,12);
    bf_put_n(ni,nid,12);

    de2bi(duid,nid,4);
    bf_put_n(ni,nid,4);

    // get bits and send to bch encoder
    bf_get_n(ni,nid,16);
    bf_free(ni);

    nid[63] = par_bit;
}

//-------------------------------------------------------------------------
/**
 * @details	Encodes the Header Data Unit (HDU)
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::hdu_enc()
{
    unsigned int i;
    long encw;
    struct bit_fifo *hc;
    char enc_bits[36];
    char bits[32];
    char nid[NID_BLEN];
    char hdu[HDU_BLEN];

    hc = bf_alloc(HDU_BLEN);

	// send FrameSync
    send_to_bitstream((char *)fs,FS_BLEN);
    // send nid
    nid_enc(nid,DUID_HDU,0);
    send_to_bitstream(nid,NID_BLEN);

    for (i = 0; i < sizeof(m_ui.ui.mi); i++) {
        de2bi(m_ui.ui.mi[i],bits,8);
        bf_put_n(hc,bits,8);
    }

    de2bi(m_ui.ui.mfId,bits,8);
    bf_put_n(hc,bits,8);

    de2bi(m_ui.ui.algId,bits,8);
    bf_put_n(hc,bits,8);

    de2bi(m_ui.ui.keyId,bits,16);
    bf_put_n(hc,bits,16);

    de2bi(m_ui.ui.tgId,bits,16);
    bf_put_n(hc,bits,16);

    // get bits and send to rs encoder
    bf_get_n(hc,hdu,120);
    bf_free(hc);

    for (i = 0; i < 20; i++) {
        bits[i] = bi2de(hdu+(i*6),6);
    }

    reverse_bits(bits,20);
    encode_rs_36_20(bits,enc_bits);
    reverse_bits(enc_bits,36);

    for (i = 0; i < 36; i++) {
        de2bi(enc_bits[i],hdu+(i*6),6);
    }

    // get rs and send to golay encoder
    for (i = 0; i < 36; i++) {
        encw = golay_18_6_encode(bi2de(hdu+(i*6),6));
        de2bi(((encw >> 16) & 0x03),bits,2);
        send_to_bitstream(bits,2);
        de2bi(encw & 0xffff, bits,16);
        send_to_bitstream(bits,16);
    }

    memset(bits,0,HDU_PAD_BLEN);
    send_to_bitstream(bits,HDU_PAD_BLEN);
}

//-------------------------------------------------------------------------
/**
 * @details		Encodes the Logical Data Unit (both 1 and 2)
 * @param type	Type of LDU (1 or 2)
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::ldu_enc(char type, char *lsd)
{
    unsigned int i;
    long encw;
    char bits[32];

    char vcw[VC_BLEN];
    char* ldu_nid;
    char* cntrlW;

    switch (type) {
		case DUID_LDU1:
			ldu_nid = ldu1_nid;
			cntrlW = lcw;
			break;
		case DUID_LDU2:
			ldu_nid = ldu2_nid;
			cntrlW = esw;
			break;
		default:
			ERROR("%d is not a valid LDU type", type);
			return;
    }

    // send FrameSync
    send_to_bitstream((char *)fs,FS_BLEN);
    // send nid
    send_to_bitstream(ldu_nid,NID_BLEN);
	// send first vcw
	get_vcw(vcw);
	send_to_bitstream(vcw,VC_BLEN);

    //send_to_bitstream(vcw,VC_BLEN);
    for (i = 0; i < 6; i++) {
		get_vcw(vcw);
		send_to_bitstream(vcw,VC_BLEN);
        send_to_bitstream(cntrlW+(i*40),40);
    }
	get_vcw(vcw);
    send_to_bitstream(vcw,VC_BLEN);

    //encode lsd
    for (i = 0; i < 2; i++) {
        encw = cyclic_16_8_encode(lsd[i]);
        de2bi(encw & 0xffff,bits,16);
        send_to_bitstream(bits,16);
    }

	get_vcw(vcw);
    send_to_bitstream(vcw,VC_BLEN);
}

//-------------------------------------------------------------------------
/**
 * @details		Encodes the Terminator Data Unit (TDU)
 * @param type	Type of TDU ("Short TDU" or "Long TDU")
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::tdu_enc(char type)
{
    unsigned int i;
    long encw;
    struct bit_fifo *tc;
    char enc_bits[24];
    char bits[32];
    char nid[NID_BLEN];
    char tdu[TDU_LCW_BLEN];


    // send FrameSync
    send_to_bitstream((char *)fs,FS_BLEN);

    switch (type) {
    case DUID_TDU:
        // send nid
		nid_enc(nid,DUID_TDU,0);
        send_to_bitstream(nid,NID_BLEN);

		memset(bits,0,TDU_PAD_BLEN);
		send_to_bitstream(bits,TDU_PAD_BLEN);
		break;
    case DUID_TDU_LCW:
        tc = bf_alloc(TDU_LCW_BLEN);

        // send nid
		nid_enc(nid,DUID_TDU_LCW,0);
        send_to_bitstream(nid,NID_BLEN);

		de2bi(0x03,bits,8);
		bf_put_n(tc,bits,8);

		de2bi(m_ui.ui.mfId,bits,8);
		bf_put_n(tc,bits,8);

		de2bi(0x0,bits,8);
		bf_put_n(tc,bits,8);

		for (i = 0; i < sizeof(m_ui.ui.dstId); i++) {
			de2bi(m_ui.ui.dstId[i],bits,8);
			bf_put_n(tc,bits,8);
		}

		for (i = 0; i < sizeof(m_ui.ui.unitId); i++) {
			de2bi(m_ui.ui.unitId[i],bits,8);
			bf_put_n(tc,bits,8);
		}

		// get bits and send to rs(24,12,13) encoder
		bf_get_n(tc,tdu,72);
        bf_free(tc);

		for (i = 0; i < 12; i++) {
			bits[i] = bi2de(tdu+(i*6),6);
		}

		reverse_bits(bits,12);
		encode_rs_24_12(bits,enc_bits);
		reverse_bits(enc_bits,24);

		for (i = 0; i < 24; i++) {
			de2bi(enc_bits[i],tdu+(i*6),6);
		}

		// get rs and send to golay encoder
		for (i = 0; i < 12; i++) {
			encw = golay_24_12_encode(bi2de(tdu+(i*12),12));
			de2bi(((encw >> 16) & 0xff),bits,8);
			send_to_bitstream(bits,8);
			de2bi(encw & 0xffff, bits,16);
			send_to_bitstream(bits,16);
		}

		memset(bits,0,TDU_LCW_PAD_BLEN);
		send_to_bitstream(bits,TDU_LCW_PAD_BLEN);
                DBG("Transmitted TDU\n");
        break;
    }
}

//-------------------------------------------------------------------------
/**
 * @details	Prepares Transmitter class global variables (status symbol,
 * 			bit counter, session-wide constants based on current config.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::tx_initialize()
{
	if (m_ui.ui.algId != ALGID_UNENCRYPTED) 
	    	encrypted_frame_mode = true;

    // Set status symbols
    stat_sym[0] = 0;
    stat_sym[1] = 0;
    bitCount = 0;
    enc_speedup();
}

//-------------------------------------------------------------------------
/**
 * @details	Flushes bit stream output in preparation for next TX
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::tx_shutdown()
{
	DBG("Resetting tx_bf.\n");
	bf_reset(tx_bf);

    // switched from TX to RX mode, stop waiting on any semaphores
    if ((sem_post(&m_bf_full) < 0) || (sem_post(&m_bf_empty) < 0)) {
    	ERROR("error in sem_post()\n");
    }

	DBG("Waiting on sem_voc_ack\n");
	sem_wait(&sem_voc_ack);

    DBG("Resetting tx_bf.\n");
    bf_reset(tx_bf);

    sem_post(&sem_ui_tx_ack);
    tx_stop_requested = false;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::get_vcw(char* vcw)
{
    int res = 0;

    //INFO("Get VCW\n");

    //while(res != VC_BLEN && (mController->get_radio_activity() == TX_ACTIVITY)) {
        res = receive_vc_data(vcw);
    //}

    // TODO test pkt header, not length
	if (res != VC_BLEN) {
		// Radio Mode has changed to RX set Voice Code Words to 0
    	memset(vcw,0,VC_BLEN);
	}
    //INFO("Got VCW\n");
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::send_to_bitstream(char* bits, int len)
{
    int ret, numtoSend, offset;

    offset = 0;
    while ( len ) {
        if ((len + bitCount) > 70 ) {
            numtoSend = 70 - bitCount;
        } else {
            numtoSend = len;
        }

        ret = bs_output(bits+offset,numtoSend);
        if (ret < 0) {
            DBG("Couldn't send to DSP\n");
            break;
        }
        bitCount += ret;
        len -= ret;
        offset += ret;

        if (bitCount == 70) {
            ret = bs_output(stat_sym,SS_BLEN);
            bitCount = 0;
        }
    }
}

//-------------------------------------------------------------------------
/**
 * @details Starts the frame transmission process and continues running
 * 			while the radio activity is still TX activity. The HDU is encoded
 * 			and put in the bit stream followed by successive pairs of
 * 			encoded LDU1 and LDU2 frames. Once the radio activity is no
 * 			longer TX activity a TDU frame is encoded and put in the bit
 * 			stream.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Transmitter_i::run()
{
	char lsd[2], lc_tmp[9];
	struct bit_fifo *lc_bf = bf_alloc(LCW_BLEN);

	CryptoPacket c_pkt;

	PortTypes::CharSequence_var buffer_var;

	// Check current radio activity
	while (mController->get_radio_state() == RUNNING_STATE)
	{
		// And if radio is in transmit activity
		if (mController->get_radio_activity() == TX_ACTIVITY) {

			DBG("Transmitter started...., algid %x\n", m_ui.ui.algId);
			tx_initialize();

		    // If ALGID field is != ALGID_UNENCRYPTED, then data is encrypted
		    if (m_ui.ui.algId != ALGID_UNENCRYPTED) {
		    	encrypted_frame_mode = true;

		    	// Send Encryption Device a CRYPTO_PKT_CNTRL_START message
		    	sendbuf_tx_crypto(CRYPTO_PKT_CNTRL_START);

		    	do {
		    		recvbuf_tx_crypto((char *)&c_pkt, sizeof(c_pkt));
		    	} while (c_pkt.hdr != CRYPTO_PKT_MI_HDU);
				memcpy(m_ui.ui.mi,c_pkt.data,c_pkt.len);
		    	DBG("Got MI_HDU\n");

		    	// Send LC in preparation for LDU1
				sendbuf_tx_crypto(CRYPTO_PKT_LC, lc, sizeof(lc)-1);
		    } else {
		    	encrypted_frame_mode = false;
		    	DBG("Encrypted_frame_mode = false\n");
		    }

			hdu_enc();

			while(mController->get_radio_activity() == TX_ACTIVITY)
			{
				// Setting dummy LSD data for LDU1
				lsd[0] = 0xDE;
				lsd[1] = 0xAD;

				if (encrypted_frame_mode) {
					// Send LC and LSD to crypto process
					sendbuf_tx_crypto(CRYPTO_PKT_LSD, lsd, sizeof(lsd));

					// Receive encrypted LC and LSD from crypto process
					do {
						recvbuf_tx_crypto((char *)&c_pkt, sizeof(c_pkt));
					} while (c_pkt.hdr != CRYPTO_PKT_LC);
					memcpy(lc_tmp,c_pkt.data,c_pkt.len);
					DBG("Got LC\n");

					do {
						recvbuf_tx_crypto((char *)&c_pkt, sizeof(c_pkt));
					} while (c_pkt.hdr != CRYPTO_PKT_LSD);
					memcpy(lsd,c_pkt.data,c_pkt.len);
					DBG("Got LSD1\n");

					// Perform FEC on encrypted LC.
					bf_import(lc_bf, lc_tmp, 72);
					lcw_enc(lc_bf);
				}

				ldu_enc(DUID_LDU1, lsd);

				/** @todo May want to check if radio activity is still TX
				 * 	here before continuing to send LDU2
				 */
				// Setting dummy LSD data for LDU2
				lsd[0] = 0xBE;
				lsd[1] = 0xEF;

				if (encrypted_frame_mode) {
					// Receive updated encrypted MI from crypto process
					do {
						recvbuf_tx_crypto((char *)&c_pkt, sizeof(c_pkt));
					} while (c_pkt.hdr != CRYPTO_PKT_MI_LDU);
					memcpy(m_ui.ui.mi,c_pkt.data,c_pkt.len);
					DBG("Got MI_LDU2\n");

					esw_enc();

					// Send LSD to crypto process
					DBG("Sending LSD to encryption device\n");
					sendbuf_tx_crypto(CRYPTO_PKT_LSD, lsd, sizeof(lsd));

					do {
						recvbuf_tx_crypto((char *)&c_pkt, sizeof(c_pkt));
					} while (c_pkt.hdr != CRYPTO_PKT_LSD);
					memcpy(lsd,c_pkt.data,c_pkt.len);
					DBG("Got LSD2\n");

					// Send Encryption Device LC
					sendbuf_tx_crypto(CRYPTO_PKT_LC, lc, sizeof(lc)-1);
				}

				ldu_enc(DUID_LDU2, lsd);

				DBG("lsd2 = 0x%02X%02X\n", lsd[0], lsd[1]);
			}

			if (encrypted_frame_mode) {
				// Send Encryption Device CRYPTO_PKT_LSD
				sendbuf_tx_crypto(CRYPTO_PKT_LSD, lsd, sizeof(lsd));
				sendbuf_tx_crypto(CRYPTO_PKT_LSD, lsd, sizeof(lsd));

				// Send Encryption Device CRYPTO_PKT_CNTRL_STOP
				sendbuf_tx_crypto(CRYPTO_PKT_CNTRL_STOP);
				DBG("Sent CRYPTO_STOP to encryption device\n");

				do {
					recvbuf_tx_crypto((char *)&c_pkt, sizeof(c_pkt));
				} while(c_pkt.hdr != CRYPTO_PKT_CNTRL_ACK);
				DBG("Got ack from encryption device\n");

				sem_post(&sem_enc_ack);
			}

			tdu_enc(DUID_TDU_LCW);
			tx_shutdown();

		} else {
			DBG("Waiting on sem_run_tx\n");
			sem_wait(&sem_run_tx);
		}
	}
}

//-------------------------------------------------------------------------
/**
 * @details		Gets data from Tx buffers
 * @param buf	Contains contents pulled from Tx buffer
 * @param len	Number of elements to be placed in buf
 * @return 		Number of elements actually read from Tx buffer
 * @author		Jeffrey Cuenco <jcuenco@ucsd.edu>
 *
 */
//-------------------------------------------------------------------------
int p25_Transmitter_i::receive_vc_data(char *buf, int len)
{
    int res, bf_isempty, bufspace;

    // Block on semaphore if buffer is empty.
    pthread_mutex_lock(&vc_tx_bf_m);
    bf_isempty = bf_empty(tx_bf) || (bf_size(tx_bf) < len);
    pthread_mutex_unlock(&vc_tx_bf_m);

    while (bf_isempty && !tx_stop_requested) {
		//INFO("bf_isempty, waiting\n");
		if (sem_wait(&m_bf_empty) < 0) {
			DBG("error in sem_wait\n");
			continue;
		}

		pthread_mutex_lock(&vc_tx_bf_m);
		bf_isempty = bf_empty(tx_bf) || (bf_size(tx_bf) < len);
		pthread_mutex_unlock(&vc_tx_bf_m);
    }
    //INFO("!bf_isempty, continuing\n");

    // Retrieve requested data from buffer.
    res = 0;
    if (len > 0) {
		pthread_mutex_lock(&vc_tx_bf_m);
		res = bf_get_n(tx_bf, buf, len);
		bufspace = bf_space(tx_bf);
		pthread_mutex_unlock(&vc_tx_bf_m);
    }

    // Notify the producer that the buffer is no longer full
    // provided a nonzero data size was retrieved from the buffer.
    if (bufspace > 0) {
		//INFO("bf_isfull, posting\n");
		if (sem_post(&m_bf_full) < 0) {
			ERROR("error in sem_post\n");
			return 0;
		}
    }
    return res;
}

//-------------------------------------------------------------------------
/**
 * @details	Pushes data to Modem component.
 * @param 	buf	Data to be sent to Modem component.
 * @param	len	Length of data to be sent to Modem component in bytes.
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Transmitter_i::bs_output(char *buf, unsigned int len)
{
	PortTypes::CharSequence_var buffer_var;

    if (len && (buf != NULL)) {
		//INFO("sending data to Modem Device.\n");
    	buffer_var = new PortTypes::CharSequence(len, len, (CORBA::Char *)buf, 0);

		//mhalOutPort->usedObject->pushPacket(1, buffer_var);
	    modem_data->pushPacket(buffer_var);
    }

    return len;
}


//-------------------------------------------------------------------------
/**
 * @details		Sends control information and data to Cryptography component.
 * @param hdr	Header of packet.
 * @param data	Defaults to NULL if only control message needs to be sent.
 * 				Contains data based on header otherwise.
 * @param len	Defaults to 0 if only control message needs to be sent.
 * 				Contains length of data otherwise.
 * @author		Justyn Bell <jwbell@ucsd.edu>
 *
 */
//-------------------------------------------------------------------------
int p25_Transmitter_i::sendbuf_tx_crypto(short hdr, char *data, int len)
{
	PortTypes::CharSequence_var buffer_var;
	CryptoPacket c_pkt;

	c_pkt.hdr = hdr;
	c_pkt.len = ((data == NULL) ? 0 : len);

	memcpy((char *)&c_pkt.data, data, len);
	DBG("Sending cntrl 0x%04X, len %d\n", c_pkt.hdr, c_pkt.len);

	buffer_var = new PortTypes::CharSequence(sizeof(CryptoPacket),
			sizeof(CryptoPacket), (CORBA::Char *)&c_pkt, 0);
	enc_dataOut->pushPacket(buffer_var);

	return 0;
}

//-------------------------------------------------------------------------
/**
 * @details		Receives data send from Cryptography component.
 * @param buf	Buffer containing received data.
 * @param len	Length of data requested.
 * @return		Returns len.
 * @retvalue 	Returns len if successful, 0 if fail.
 * @author		Jeffrey Cuenco <jcuenco@ucsd.edu>
 *
 */
//-------------------------------------------------------------------------
int p25_Transmitter_i::recvbuf_tx_crypto(char *buf, int len)
{
    PortTypes::CharSequence *enc_data(NULL);
    CORBA::UShort enc_data_length;

    int res;
    char *data;

    DBG("Receiving encrypted payload of size %d bytes...\n", len);

    enc_dataIn->getData(enc_data);
    enc_data_length = enc_data->length();

    data = (char *)(enc_data->get_buffer());

    if (enc_data_length != len) {
    	ERROR("Enc_data_length = %d, len = %d\n", enc_data_length, len);
    	res = 0;
    }
    else {
    	memcpy(buf, data, len);
    	res = len;
    }

    enc_dataIn->bufferEmptied();

    return res;
}

//-------------------------------------------------------------------------
/**
 * @details		Sends control information and data to Vocoder device.
 * @param hdr	Header of packet.
 * @param data	Defaults to 0.  AlgId as received from UI device otherwise.
 * @author		Justyn Bell <jwbell@ucsd.edu>
 *
 */
//-------------------------------------------------------------------------
int p25_Transmitter_i::sendstate_tx_voc(short hdr, char data)
{
	PortTypes::CharSequence_var buffer_var;
	VocoderPacket v_pkt;
	int VPSize = sizeof(VocoderPacket);

	v_pkt.hdr = hdr;
	v_pkt.len = ((data == '0') ? 0 : sizeof(data));
	v_pkt.data[0] = data;

	buffer_var = new PortTypes::CharSequence(VPSize, VPSize, (CORBA::Char *)&v_pkt, 0);
	voc_out_data->pushPacket(buffer_var);

	return 0;
}

