/**
 * ============================================================================
 * @file	ambe_intf.c
 * @brief	This file contains implementation of the processing functions for
 * 			the AMBE library. This file contains modifications from the test
 * 			file provided by DVSI
 * @note	none.
 *
 * @author	DVSI
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */

/******************************************************************************
 * DVSI CONFIDENTIAL PROPRIETARY
 ******************************************************************************
 * DVSI AMBE+2 Vocoder
 * (c) Copyright, Digital Voice Systems Inc., 2001, 2002
 * All Rights Reserved
 *
 * DVSI claims certain rights, including patent rights under
 * U.S. Patents #6,199,037 B1, #6,161,089, #5,870,045, #5,826,222,
 * #5,754,974, #5,715,365, #5,701,390, #5,649,050, #5,630,011,
 * #5,247,579, and #5,226,084, and under other U.S. and foreign
 * patents and patents pending, in the AMBE^(TM) speech coding
 * technology embodied in this software.  Any use of this software
 * or technology requires a separate written license from DVSI.
 *
 * AMBE is a registered trademark of Digital Voice Systems, Inc.
 * AMBE+ and AMBE+2 are trademarks of Digital Voice Systems, Inc.
 ******************************************************************************
 *****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sched.h>

#include "common.h"
#include "ambe_intf.h"
#include "arm_ambe.h"
#include "log_arm.h"
#include "ticktock.h"
#include "byte_fifo.h"

unsigned int exprcv_bytes = 0;
unsigned int rcv_bytes = 0;

/**
 * ----------------------------------------------------------------------------
 * @name	usage
 * @brief	Print Usage Information
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
void usage( char *name ) {
	static char usagestr[] =
			"%s [options] infile outfile\n"
			"Options are:\n"
			" [-enc] [enc_options] <speech_input_file> <bits_output_file> : run the encoder only\n"
			" [-dec] [dec_options] <bits_input_file> <speech_output_file> : run the decoder only\n"
			" [enc_options] [dec_options] <speech_input_file> <speech_output_file> : run encoder and decoder in loopback\n"
			" [enc_options] are one of:\n"
			"        -p25         : full-rate mode (default (i.e. no arg) is 3600 bps)\n"
			"        -m24         : 2400 bps mode  (default (i.e. no arg) is 3600 bps)\n"
			"        -m71         : 7100 bps mode  (default (i.e. no arg) is 3600 bps)\n"
			"        -rep         : run repeater (if -p25 arg preceeds -rep,\n"
			"                                 repeater converts to 3600 bps,\n"
			"                                 otherwise repeater converts to 7200 bps)\n"
			"	 -vad level   : voice activation level (-90,3 dBm0 default=-90)\n"
			"        -nofec       : no FEC encode (default is FEC encoder on)\n"
			"        -nsoff       : turn off noise suppressor (default is on)\n"
			"        -toneoff     : disable encoder tone detection (default is on)\n"
			"        -sdbits bits     : number of soft decision bits  per channel bit (1<= bits <=8) (default=1)\n"
			"        -steal       : steal 6 bits from the vocoder (3600 bps only)\n"
			"        -dtx         : enable voice activity based DTX (3600 bps only, default is off)\n"
			"        -cnirate K   : set CNI update rate to every K frames (K=-1 is once per voice burst, \n"
			"                                 K=0 is every frame, K=1 is every other frame, etc...)\n"
			"        -info        : print out system and VAD/DTX information for each frame\n"
			"	 -agc         : enable automatic gain control (default= disabled)\n"
			" [dec_options] are one of:\n"
			"        -p25         : full-rate mode (default (i.e. no arg) is 3600 bps)\n"
			"        -m24         : 2400 bps mode  (default (i.e. no arg) is 3600 bps)\n"
			"        -m71         : 7100 bps mode  (default (i.e. no arg) is 3600 bps)\n"
			"        -rep         : run repeater (if -p25 arg preceeds -rep,\n"
			"                                 repeater converts to half-rate,\n"
			"                                 otherwise repeater converts to 7200 bps)\n"
			"        -nofec       : no FEC decode (default is FEC decoder on)\n"
			"        -sdbits bits     : number of soft decision bits  per channel bit (1<= bits <=8) (default=1)\n"
			"        -steal bits  : steal 0-6 bits from the vocoder (3600 bps only)\n"

			;
	fprintf(stderr,usagestr, name);
	exit ( 3 );
}

/**
 * ----------------------------------------------------------------------------
 * @name	byte_swap
 * @brief	Swap Bytes
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
void byte_swap(short *buf, short nsamples) {
	short i;
	DBG("swapping endian here\n");
	for(i=0;i<nsamples;i++) {
		buf[i] = ((buf[i]&0x00ff)<<8) | ((buf[i]>>8)&0x00ff);
	}
}

/**
 * ----------------------------------------------------------------------------
 * @name	get_speech
 * @brief	Read specified number of speech samples from a file
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
short  get_speech( short *buf, short  nsamples, FILE *fpi ) {
	unsigned int n;

	n = fread( buf, SIZEINT, nsamples, fpi );

	if ( n < (unsigned int) nsamples ) {
		short i;
		short *pbuf = buf + n;

		for ( i = 0; i < (short)(nsamples-n); i++ )
			*pbuf++ = 0;
	}

#ifdef BIG_ENDIAN
	//byte_swap(buf,nsamples);
#endif

	return( (short) n );
}

/**
 * ----------------------------------------------------------------------------
 * @name	put_speech
 * @brief	Write nsamples speech samples in buf to the output file fpo.
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
static short put_speech( const short *buf, short nsamples, FILE *fpo ) {
	short n;

#ifdef BIG_ENDIAN
	//byte_swap((short *)buf,nsamples);
#endif

	n = (short) fwrite( buf, SIZEINT, nsamples, fpo );

	return ( n );
}

/**
 * ----------------------------------------------------------------------------
 * @name	put_bits
 * @brief
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
static short put_bits( const short *buf, short nbits, FILE *fpo ) {
	unsigned char cbuf[PACKED_BITS_SZ];
	short i;

	for ( i = 0; i < nbits; i++ ) {
		cbuf[i] = (unsigned char) buf[i];
	}

	return fwrite( cbuf, SIZEBIT, nbits, fpo );
}

/**
 * ----------------------------------------------------------------------------
 * @name	get_bits
 * @brief
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
short get_bits( short *buf, short nbits, FILE *fpi ) {
	unsigned char cbuf[PACKED_BITS_SZ];
	short i;
	short nbits_read;

	nbits_read = fread( cbuf, SIZEBIT, nbits, fpi );

	for ( i = 0; i < nbits_read; i++ )
		buf[i] = (short) cbuf[i];

	for ( /* i = nbits_read */; i < nbits; i++ )
		buf[i] = 0;

	return nbits_read;
}

/**
 * ----------------------------------------------------------------------------
 * @name	enc_cvt_sdbits
 * @brief	Convert binary encoder channel data output to desired soft
 * 			decision size
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
static void enc_cvt_sdbits( short *cbuf, short len, short sd_bits ) {
	short i;
	for ( i = 0; i < len; i++ )
		cbuf[i] *= ((1 << sd_bits) - 1);
}

/**
 * ----------------------------------------------------------------------------
 * @name	dec_cvt_sdbits
 * @brief	Convert decoder channel data input to desired soft decision level
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
static void dec_cvt_sdbits( short *t, short len, short sd_bits ) {
	short sd_shift;
	short i;

	sd_shift = sd_bits - 1;

	for ( i = 0; i < len; i++ )
		t[i] = t[i] >> sd_shift;
}

/**
 * ----------------------------------------------------------------------------
 * @name	get_options
 * @brief	Set default options and read in command line options
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
void get_options( PC_OPTIONS *o, int argc, char *argv[] ) {
	// set default options and read in command line args
	o->verbose 		=  0;
	o->mode 		= APCO_HR_MODE;
	o->fec_mode 	= APCO_HR_MODE;			
	o->rep_mode 	= -1;
	o->run_encoder 	=  1;
	o->run_decoder 	=  1;
	o->enc_ws 		= 2*AMBE_WS_NOM;
	o->dec_ws	 	= 2*AMBE_WS_NOM;
	o->vad_level   	= VAD_LEVEL;
	o->nsprsn_enabled   =  1;
	o->tone_det_enabled =  1;
	o->tone_on		=  0;
	o->tone_amp 	= TONE_AMP;
	o->tone_index  	= TONE_INDEX;
	o->sd_bits 		= SD_BITS;
	o->sd_dist 		=  0;
	o->hd_errs 		=  0;
	o->dtx	 		=  0;
	o->bit_steal 	=  0;
	o->cnirate 		=  0;
	o->agc_on		=  0;
	o->use_fec 		=  1;
	o->compare		= COMPARE_OFF;
	o->fpc_name 	= NULL;
	o->fpc 			= NULL;
	o->fpi_name 	= NULL;
	o->fpi 			= NULL;
	o->fpo_name 	= NULL;
	o->fpo 			= NULL;

	if ( ambe_read_args( o, argc, argv ) < 0 )
		usage( argv[0] ); 
	// make sure important options have the proper values
	if ( !o->fpi ) {
		ERROR("get_options: no input file specified.\n");
		usage( argv[0] );
	}

	if ( o->mode == APCO_MODE ) {
		// no DTX or bit-stealing in APCO mode
		o->dtx = 0;
		o->bit_steal = 0;
	}
	// an output file is now optional
	// for instance if you use the -cmp file option you might not
	// want to write the output file, in order to save time

	/*
	if ( !o->fpo ) {
		fprintf( stdout, "get_options: no output file specified.\n" );
		usage( argv[0] );
	}
	 */

	if ( o->vad_level < -90 )
		o->vad_level = -90;
	else if ( o->vad_level > 3 )
		o->vad_level = 3;

	if ( o->sd_bits < 1 )
		o->sd_bits = 1;
	else if ( o->sd_bits > SD_BITS_MAX )
		o->sd_bits = SD_BITS_MAX;

}

/**
 * ----------------------------------------------------------------------------
 * @name	getnextline
 * @brief
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
int getnextline( FILE *fp, char *line ) {
	char c;

	// find the first non-space character
	do {
		if (feof(fp))
			return(0);	// no non-space chars left in the file
		c = fgetc( fp );
	} while (isspace(c));

	// start writing out the charcters until '\n' or EOF is found
	do {
		*line++ = c;

		if (feof(fp)) {
			*line++ = 0;
			return(1);	// end-of-file but we have non-space line to process still
		}

		c = fgetc( fp );

	} while (c != '\n');		// end-of-line

	*line++ = 0;
	return(1);
}

/**
 * ----------------------------------------------------------------------------
 * @name	get_ttlbits
 * @brief
 * @author	DVSI
 * ----------------------------------------------------------------------------
 */
int get_ttlbits(short rep_mode, short mode, short use_fec) {
	short dttlbits;

	if (rep_mode >= 0) {
#ifdef BUILD_REPEATER
		if (rep_mode == APCO_MODE) {
			dttlbits  = APCO_SRCBITS;
			if (use_fec)
				dttlbits += APCO_FECBITS;
		} else {
			dttlbits  = APCO_HR_SRCBITS_MAX;
			if (use_fec)
				dttlbits += APCO_HR_FECBITS;
		}
#endif /* BUILD_REPEATER */
	} else {
		if (mode == APCO_MODE) {
			dttlbits  = APCO_SRCBITS;
			if (use_fec)
				dttlbits += APCO_FECBITS;
		} else {
			dttlbits  = APCO_HR_SRCBITS_MAX;
			if (use_fec)
				dttlbits += APCO_HR_FECBITS;
		}
	}

	return dttlbits;

}

/**
 * ----------------------------------------------------------------------------
 * @func	process_test_input
 * @brief  	Processes test data to be sent to DSP using options specified in po
 * @details	Encoder or Decoder will be run depending on the options
 * @arg		po
 *             Options struct containing parameters related to encoding/
 *             decoding.
 * @see		process_test_output
 * @author	DVSI
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
int process_test_input(PC_OPTIONS * po) {
	static short speech_buf[SPEECH_BUF_SZ];
	static short chan_buf[CHAN_BUF_SZ];
	static short frame_count = 0;
	short dttlbits;
	int ret;

	// Tracks expected bytes to be received from DSP based on how many bytes
	// Sent to DSP
	exprcv_bytes = 0;

	dttlbits = get_ttlbits(po->rep_mode, po->mode, po->use_fec);

	for (;;) {
		if (po->run_encoder) {
			// read speech input from file
			ret = get_speech(speech_buf, (short)po->enc_ws, po->fpi);
			usleep(2000);
			if (ret <= 0) {
				if (frame_count%9) {
					// Make sure number of frames is on the LDU1/LDU2 boundary
					DBG("+++++++++++++ Sending extra frame\n");
					memset(speech_buf, 0, SPEECH_BUF_SZ);
					ret = (short)po->enc_ws;
				} else {
					INFO("Done reading speech from %s\n",po->fpi_name);
					break;
				}
			}

			frame_count++;

			// This is useful for files that are not exact multiples of 320
			// bytes. Since, the speech_buf is padded with zeros we can send
			// the entire buffer. ret is the number of bytes read
			if (ret != (short)po->enc_ws) ret = po->enc_ws;

			// Send voice data to DSP for encoding
			ARMAMBE_SendBuffer((char *)&speech_buf, ret*sizeof(short), 0);
			exprcv_bytes += dttlbits*sizeof(short);  
		}

		if (po->run_decoder) {
			// read channel input from file if decoder only
			if (!po->run_encoder) {
				if (get_bits(chan_buf, dttlbits, po->fpi) != dttlbits) {
					INFO("Done reading encoded bits from %s\n", po->fpi_name);
					break;
				}
				// Send encoded voice data top DSP for decoding
				ARMAMBE_SendBuffer((char *)&chan_buf, dttlbits*sizeof(short), 0);
				exprcv_bytes += po->enc_ws*sizeof(short);  
			} else {
				exprcv_bytes += po->enc_ws*sizeof(short) - dttlbits*sizeof(short);  
			}
		}
		//INFO("*\n");
		INFO("EXPECTED RECEIVED BYTES = %d\n",exprcv_bytes);

		if (frame_count%9 == 0)
			frame_count = 0;
	}

	return (0);
}

/**
 * ----------------------------------------------------------------------------
 * @func	process_test_output
 * @brief  	Processes test data received from DSP using options specified in po
 * @details	Encoder or Decoder will be run depending on the options
 * @arg		po
 *             Options struct containing parameters related to encoding/
 *             decoding.
 * @see		process_test_input
 * @author	DVSI
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
int process_test_output(PC_OPTIONS * po) {
	long compare_count = 0;
	static short speech_buf[SPEECH_BUF_SZ];
	static short chan_buf[CHAN_BUF_SZ];
	static short cmp_speech_buf[SPEECH_BUF_SZ];
	static short cmp_chan_buf[CHAN_BUF_SZ];
	short compare_fail = 0;
	short echanbits;
	short ws;
	short i;
	int ret;

	rcv_bytes = 0;
	echanbits = get_ttlbits(po->rep_mode, po->mode, po->use_fec);

	for (;;) {
		if (po->run_encoder) {
			if (!po->run_decoder) {
				// If we are not running decoder after encoding then compare
				// output to known encoded bits of the test vectors and write
				// output to file

				TICK();

				// Receive encoded voice bits from DSP
				ret = ARMAMBE_ReceiveBuffer((char *)&chan_buf, echanbits*sizeof(short));
				TOCK("Elapsed Encode");
				if (ret != echanbits*sizeof(short))
					break;

				rcv_bytes += ret;  

				enc_cvt_sdbits(chan_buf, echanbits, po->sd_bits);

				// Compare to known encoded bits
				if (po->compare > 0) {
					compare_fail = 0;
					if (get_bits(cmp_chan_buf, echanbits, po->fpc) == 0) {
						INFO("EOF on %s\n", po->fpc_name);
						return 1;
					}
					for (i = 0; i < echanbits; i++) {
						if (cmp_chan_buf[i] != chan_buf[i]) {
							compare_fail = 1;
							break;
						}
					}
					compare_count += i;
					if (compare_fail) {
						ERROR("output differs from %s at char %ld\n", po->fpc_name, 2*compare_count);
						if (po->compare == COMPARE_FAIL)
							exit(0); // stop processing vectors
						else if (po->fpo == (FILE *) 0){
							return(1);	// if no output file, then start next vector
						} else
							po->compare = COMPARE_OFF; // stop comparing after a difference is found
					}
				}

				// write channel output to file if only the encoder is running
				if (po->fpo != (FILE *)0) {
					put_bits(chan_buf, echanbits, po->fpo);
				}
			}
		}

		if (po->run_decoder) {
			// If we only ran decoder then compare output to known decoded bits
			// for the test vectors and write output to file

			TICK();
			ret = ARMAMBE_ReceiveBuffer((char *)&speech_buf, po->dec_ws*sizeof(short));
			TOCK("Elapsed Decode");
			if (ret != po->dec_ws*sizeof(short))
				break;

			rcv_bytes += ret;  

			// Compare to known decoded bits
			if (po->compare > 0) {
				compare_fail = 0;
				ws = (short) po->dec_ws;
				if (get_speech(cmp_speech_buf, ws, po->fpc) <= 0) {
					INFO("EOF on %s\n", po->fpc_name );
					return 1;
				}
				for (i = 0; i < ws; i++) {
					if (cmp_speech_buf[i] != speech_buf[i]) {
						DBG("FAILED WHILE COMPARING 0x%x to 0x%x\n",cmp_speech_buf[i],speech_buf[i]);
						compare_fail = 1;
						break;
					}
				}
				compare_count += i;
				if (compare_fail) {
					ERROR("output differs from %s at sample %ld\n", po->fpc_name, compare_count );
					if (po->compare == COMPARE_FAIL)
						exit(0); // stop processing vectors
					else if (po->fpo == (FILE *) 0) {
						return(1); // if no output file, then start next vector
					} else
						po->compare = COMPARE_OFF; // stop comparing after a difference is found
				}
			}

			// write synthesized speech to output file
			if (po->fpo != (FILE *)0)
				put_speech( speech_buf, (short)po->dec_ws, po->fpo );
		}

		//INFO("*\n");
		INFO("RECEIVED BYTES = %d\n",rcv_bytes);

		//if (exprcv_bytes == rcv_bytes) 
		//	break;
	}

	return (0);
}

/**
 * ----------------------------------------------------------------------------
 * @func	process_test_cryptography
 * @brief  	Simulates encryption/decryption processing by delaying and looping
 * 			back voice data.
 * @details	Tests showed that delays of up to 15 milliseconds could be added
 * 			and the 20 millisecond constraint on P25 transmission would still
 * 			be maintained.
 * @arg		po
 *             Options struct containing parameters on related to encoding/
 *             decoding.
 * @see		process_test_input
 * @author	DVSI
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
int process_test_cryptography(PC_OPTIONS * po) {

	static short chan_buf[CHAN_BUF_SZ];
	static short recv_frame_count = 0;
	static short sent_frame_count = 0;
	short echanbits;
	int mode;
	struct byte_fifo *bf_hold;

	// Compute number of bits to expect depending on encode/decode mode
	echanbits = (po->mode == APCO_MODE) ? APCO_SRCBITS : APCO_HR_SRCBITS_MAX;
	bf_hold = bf_alloc(2*echanbits*sizeof(short));
	
	for (;;) {
		do {
			// Receive data to be encrypted or decrypted and
			mode = ARMAMBE_ReceiveCryptBuffer((char *)&chan_buf, echanbits*sizeof(short));

			// Store received data in a FIFO
			bf_put_n(bf_hold,(char *)&chan_buf,echanbits*sizeof(short));

			recv_frame_count++;
			DBG("************************* recv_frame_count = %d\n\n",recv_frame_count);

			switch (mode) {
			case CRYPT_ENCRYPTION:
				DBG("ENCRYPTING!\n");
				usleep(15000);
				break;
			case CRYPT_DECRYPTION:
				DBG("DECRYPTING!\n");
				usleep(15000);
				break;
			case CRYPT_END:
				INFO("Finished Cryptography\n");
				recv_frame_count = 0;
				sent_frame_count = 0;
				return 1;
			default:
				ERROR("Unexpected Control\n");
			}
		} while (recv_frame_count == 1 || recv_frame_count == 10);

		// Grab voice data from FIFO and send back to DSP
		bf_get_n(bf_hold,(char *)&chan_buf,echanbits*sizeof(short));
		ARMAMBE_SendCryptBuffer((char *)&chan_buf, echanbits*sizeof(short));
		sent_frame_count++;

		// Due to the encryption schedule, our implementation involved
		// sending two AMBE frames to be encrypted at the beginning to allow
		// the encryption algorithm produce its first output vector. One frame
		// after that. This is repeated every 9 frames. This implementation
		// leads to two frames being output every 9 frames. Hence we simulate
		// an extra frame being sent to DSP every 9th and 18th frame.
		if (recv_frame_count == 9 || recv_frame_count == 18) {
			bf_get_n(bf_hold,(char *)&chan_buf,echanbits*sizeof(short));
			ARMAMBE_SendCryptBuffer((char *)&chan_buf, echanbits*sizeof(short));
			sent_frame_count++;
		}
		
		if (recv_frame_count == 18) {
			recv_frame_count = 0;
			sent_frame_count = 0;
		}
	}

	bf_free(bf_hold);	

	return (0);
}
