/**
 * ============================================================================
 * @file    arm_ambe.c
 * @brief   This file contains functions for setting up the ARM processor to
 * 			communicate with the DSP. It also implements routines to send and
 * 			receive data from the DSP.
 * @note	none.
 *
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */

/* ----------------------------- Linux specific ---------------------------- */
#include <string.h>
#include <pthread.h>
#include <semaphore.h>

/* ---------------------------- DSP/BIOS Link ------------------------------ */
#include <dsplink.h>

/* ---------------------------- DSP/BIOS LINK API -------------------------- */
#include <proc.h>
#include <chnl.h>
#include <pool.h>
#include <notify.h>

/* ---------------------------- Application specific ----------------------- */
#include "common.h"
#include "bs_packet.h"
#include "byte_fifo.h"
#include "arm_ambe.h"
#include "dprintf.h"
#include "log_arm.h"

#if defined (__cplusplus)
extern "C" {
#endif /* defined (__cplusplus) */


/**
 * ----------------------------------------------------------------------------
 * @name   	NUMBUFFERPOOLS
 * @brief   Number of buffer pools in this application.
 * ----------------------------------------------------------------------------
 */
#define NUMBUFFERPOOLS	1

/**
 * ----------------------------------------------------------------------------
 * @name	NUMBUFSINPOOL
 * @brief	Number of buffers in pool.
 * ----------------------------------------------------------------------------
 */
#define NUMBUFSINPOOL	8

/**
 * ----------------------------------------------------------------------------
 * @name	POOL_ID
 * @brief	Pool id for this application.
 * ----------------------------------------------------------------------------
 */
#define POOL_ID        0

/**
 * ----------------------------------------------------------------------------
 * @name   	NUMIOREQBUFS
 * @brief   Number of IO Req Buffers.
 * ----------------------------------------------------------------------------
 */
#define NUMIOREQBUFS	4

/*
 * ----------------------------------------------------------------------------
 * @name   	ARMAMBE_BufferSize
 * @brief   Size of buffer to be used for data transfer.
 * ----------------------------------------------------------------------------
 */
STATIC Uint32  ARMAMBE_BufferSize ;

/**
 * ----------------------------------------------------------------------------
 * @name   	ARMAMBE_Buffers
 * @brief   Array of buffers used by the arm_ambe application.
 * ----------------------------------------------------------------------------
 */
STATIC Char8 * ARMAMBE_Buffers [NUMIOREQBUFS] ;

/**
 * ----------------------------------------------------------------------------
 * @name   	ARMAMBE_Send_IOReq
 * @brief   Information for adding or reclaiming a send request to DSP for voice
 * 			data.
 * ----------------------------------------------------------------------------
 */
STATIC ChannelIOInfo ARMAMBE_Send_IOReq;

/**
 * ----------------------------------------------------------------------------
 * @name   	ARMAMBE_Receive_IOReq
 * @brief   Information for adding or reclaiming a receive request from DSP for
 * 			voice data.
 * ----------------------------------------------------------------------------
 */
STATIC ChannelIOInfo ARMAMBE_Receive_IOReq;

/**
 * ----------------------------------------------------------------------------
 * @name   	ARMAMBE_SendC_IOReq
 * @brief   Information for adding or reclaiming a send request to DSP for
 * 			encrypted voice data.
 * ----------------------------------------------------------------------------
 */
STATIC ChannelIOInfo ARMAMBE_SendC_IOReq;

/**
 * ----------------------------------------------------------------------------
 * @name   	ARMAMBE_ReceiveC_IOReq
 * @brief	Information for adding or reclaiming a receive request from DSP for
 * 			encrypted voice data.
 * ----------------------------------------------------------------------------
 */
STATIC ChannelIOInfo ARMAMBE_ReceiveC_IOReq;

/**
 * ----------------------------------------------------------------------------
 * @name   	procId
 * @brief	Holds the processor Id
 * ----------------------------------------------------------------------------
 */
Uint8 procId;

/**
 * ----------------------------------------------------------------------------
 * @name   	thr_host_arm
 * @brief   	Handle for thread that processes data coming from Host to ARM
 * ----------------------------------------------------------------------------
 */
static pthread_t		thr_host_arm;

/**
 * ----------------------------------------------------------------------------
 * @name   	thr_arm_host
 * @brief	Handle for thread that processes data going from ARM to Host
 * ----------------------------------------------------------------------------
 */
static pthread_t		thr_arm_host;

/**
 * ----------------------------------------------------------------------------
 * @name   	thr_dsp_arm
 * @brief	Handle for thread that processes data coming from DSP to ARM
 * ----------------------------------------------------------------------------
 */
static pthread_t		thr_dsp_arm;

/**
 * ----------------------------------------------------------------------------
 * @name   	thr_arm_dsp
 * @brief	Handle for thread that processes data going from ARM to DSP
 * ----------------------------------------------------------------------------
 */
static pthread_t		thr_arm_dsp;

/**
 * ----------------------------------------------------------------------------
 * @name   	mtx_arm_dsp
 * @brief	Mutex for processes with data flowing from ARM to DSP
 * ----------------------------------------------------------------------------
 */
static pthread_mutex_t 	mtx_arm_dsp;

/**
 * ----------------------------------------------------------------------------
 * @name   	mtx_dsp_arm
 * @brief	Mutex for processes with data flowing from DSP to ARM
 * ----------------------------------------------------------------------------
 */
static pthread_mutex_t 	mtx_dsp_arm;


/**
 * ----------------------------------------------------------------------------
 * @name  	bf_arm_dsp
 * @brief	FIFO for data going from ARM to DSP
 * ----------------------------------------------------------------------------
 */
static struct byte_fifo *bf_arm_dsp;

/**
 * ----------------------------------------------------------------------------
 * @name  	bf_dsp_arm
 * @brief	FIFO for data coming from DSP to ARM
 * ----------------------------------------------------------------------------
 */
static struct byte_fifo *bf_dsp_arm;


/**
 * ----------------------------------------------------------------------------
 * @name   	sem_arm_dsp_flush
 * @brief	Semaphore
 * ----------------------------------------------------------------------------
 */
static sem_t			sem_arm_dsp_flush;

/**
 * ----------------------------------------------------------------------------
 * @name   	sem_arm_dsp_empty
 * @brief	Semaphore
 * ----------------------------------------------------------------------------
 */
static sem_t			sem_arm_dsp_empty;

/**
 * ----------------------------------------------------------------------------
 * @name   	sem_arm_dsp_notempty
 * @brief	Semaphore
 * ----------------------------------------------------------------------------
 */
static sem_t			sem_arm_dsp_notempty;

/**
 * ----------------------------------------------------------------------------
 * @name   	sem_dsp_arm_empty
 * @brief	Semaphore
 * ----------------------------------------------------------------------------
 */
static sem_t			sem_dsp_arm_empty;

/**
 * ----------------------------------------------------------------------------
 * @name   	sem_dsp_arm_notempty
 * @brief	Semaphore
 * ----------------------------------------------------------------------------
 */
static sem_t			sem_dsp_arm_notempty;

/**
 * ----------------------------------------------------------------------------
 * @name   	sem_dsp_arm_stateswitch
 * @brief	Semaphore
 * ----------------------------------------------------------------------------
 */
static sem_t			sem_dsp_arm_stateswitch;


/**
 * ----------------------------------------------------------------------------
 * @name   	flush_bf_arm_dsp
 * @brief	Flag to indicate flush request for bf_arm_dsp FIFO
 * ----------------------------------------------------------------------------
 */
static char				flush_bf_arm_dsp;

/**
 * ----------------------------------------------------------------------------
 * @name   	cntrl_bf_arm_dsp
 * @brief	Flag to indicate control packet for bf_arm_dsp FIFO
 * ----------------------------------------------------------------------------
 */
static char				cntrl_bf_arm_dsp;


/**
 * ----------------------------------------------------------------------------
 * @name   	bf_dsp_arm_accept_less
 * @brief	Flag to indicate that processes should accept less data than they
 * 			requested from the bf_arm_dsp FIFO
 * ----------------------------------------------------------------------------
 */
static char 			bf_dsp_arm_accept_less;


/**
 * ----------------------------------------------------------------------------
 * @name   	SendToDspTask_started
 * @brief   Flag to indicate that thread for sending data to DSP has started
 * ----------------------------------------------------------------------------
 */
char 					SendToDspTask_started;

/**
 * ----------------------------------------------------------------------------
 * @name   	receiveFromDspTask_started
 * @brief	Flag to indicate that thread for receiving data from DSP has started
 * ----------------------------------------------------------------------------
 */
char 					receiveFromDspTask_started;

/**
 * ----------------------------------------------------------------------------
 * @see   	arm_running
 * ----------------------------------------------------------------------------
 */
extern char				arm_running; 

/**
 * ----------------------------------------------------------------------------
 * @see   	dsp_ack_stop
 * ----------------------------------------------------------------------------
 */
extern char				dsp_ack_stop; 



/**
 * ----------------------------------------------------------------------------
 * @func	ARMAMBE_Create
 * @brief  	This function allocates and initializes resources used by
 *        	the ARM application.
 * @arg		dspExecutable
 *             DSP executable name.
 * @return	DSP_SOK
 *              Operation Successfully completed.
 *          DSP_EFAIL
 *              Resource allocation failed.
 * @see		ARMAMBE_Delete
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
DSP_STATUS
ARMAMBE_Create (IN Char8 * dspExecutable)
{
	DSP_STATUS status = DSP_SOK;
	Uint32 numBufs[NUMBUFFERPOOLS] = { NUMBUFSINPOOL };

	ChannelAttrs chnlAttrInput;
	ChannelAttrs chnlAttrOutput;
	Uint32 size[NUMBUFFERPOOLS];
#if defined (ZCPY_LINK)
	SMAPOOL_Attrs poolAttrs                ;
#endif /* if defined (ZCPY_LINK) */

	DBG("Entered ARMAMBE_Create\n") ;

	// Create and initialize the proc object.
	status = PROC_setup (NULL) ;

	// Attach to the DSP
	if (DSP_SUCCEEDED (status)) {
		status = PROC_attach (procId, NULL) ;
		if (DSP_FAILED (status)) {
			ERROR("PROC_attach failed. Status = [0x%x]\n", (int)status) ;
		}
	} else {
		ERROR("PROC_setup failed. Status =  [0x%x]\n", (int)status) ;
	}

	// Open the pool of buffers.
	if (DSP_SUCCEEDED (status)) {
		size [0] = ARMAMBE_BufferSize ;
		poolAttrs.bufSizes      = (Uint32 *) &size ;
		poolAttrs.numBuffers    = (Uint32 *) &numBufs ;
		poolAttrs.numBufPools   = NUMBUFFERPOOLS ;
#if defined (ZCPY_LINK)
		poolAttrs.exactMatchReq = TRUE ;
#endif /* if defined (ZCPY_LINK) */
		status = POOL_open (POOL_makePoolId(procId, POOL_ID), &poolAttrs) ;
		if (DSP_FAILED (status)) {
			ERROR("POOL_open () failed. Status = [0x%x]\n", (int)status);
		}
	}

	// Load the executable on the DSP.
	if (DSP_SUCCEEDED (status)) {
		status = PROC_load (procId, dspExecutable, 0, NULL) ;
		if (DSP_FAILED (status)) {
			ERROR("PROC_load failed. Status = [0x%x]\n", (int)status) ;
		}
	}

	// Initialize dprintf routines
	status = dprintf_init_gpp(procId);
	if (DSP_FAILED (status)) {
		ERROR("NOTIFY_register failed. Status = [0x%x]\n", (int)status) ;
	}

	// Create channels for output to DSP
	if (DSP_SUCCEEDED (status)) {
		chnlAttrOutput.mode      = ChannelMode_Output     ;
		chnlAttrOutput.endianism = Endianism_Default      ;
		chnlAttrOutput.size      = ChannelDataSize_16bits ;

		// Channel for outputting error corrected AMBE frames to DSP when in
		// receive mode or PCM data to be encoded when in transmit mode
		status = CHNL_create (procId, CHNL_ID_OUTPUT, &chnlAttrOutput) ;
		if (DSP_FAILED (status)) {
			ERROR("CHNL_create failed (output). Status = [0x%x]\n", (int)status) ;
		}

		// Channel for outputting decrypted AMBE encoded frames to DSP when
		// in receive mode or encrypted AMBE frames encoded when in transmit
		// mode
		status = CHNL_create (procId, CHNL_ID_COUTPUT, &chnlAttrOutput) ;
		if (DSP_FAILED (status)) {
			ERROR("CHNL_create failed (coutput). Status = [0x%x]\n", (int)status) ;
		}
	}

	// Create a channels for input from DSP
	if (DSP_SUCCEEDED (status)) {
		chnlAttrInput.mode      = ChannelMode_Input      ;
		chnlAttrInput.endianism = Endianism_Default      ;
		chnlAttrInput.size      = ChannelDataSize_16bits ;

		// Channel for receiving error corrected AMBE frames from DSP when
		// in transmit mode or decoded PCM data when in receive mode
		status = CHNL_create (procId, CHNL_ID_INPUT, &chnlAttrInput) ;
		if (DSP_FAILED (status)) {
			ERROR("CHNL_create failed (input). Status = [0x%x]\n", (int)status) ;
		}

		// Channel for receiving AMBE encoded frames to be encrypted from DSP
		// when in transmit mode or AMBE frames to be decrypted when in receive
		// mode
		status = CHNL_create (procId, CHNL_ID_CINPUT, &chnlAttrInput) ;
		if (DSP_FAILED (status)) {
			ERROR("CHNL_create failed (cinput). Status = [0x%x]\n", (int)status) ;
		}
	}

	// Allocate all buffers for data transfer to DSP.
	if (DSP_SUCCEEDED (status)) {
		status = CHNL_allocateBuffer (procId, CHNL_ID_OUTPUT, ARMAMBE_Buffers, ARMAMBE_BufferSize , NUMIOREQBUFS) ;
		if (DSP_FAILED (status)) {
			ERROR("CHNL_allocateBuffer failed (output) Status = [0x%x]\n", (int)status) ;
		}
	}

	// Initialize the allocated buffers with zeros.
	int i;
	for (i=0; i < NUMIOREQBUFS; i++) {
		memset(ARMAMBE_Buffers[i],0,ARMAMBE_BufferSize);
	}

	// Allocate byte fifos for buffering data to be transferred usind the
	// DPS Bios Link Channels
	DBG("Allocating buffers for DSP <--> ARM transfers\n");
	bf_arm_dsp = bf_alloc(BS_BUF_LEN8);
	bf_dsp_arm = bf_alloc(BS_BUF_LEN8);

	// Initialize semaphores
	sem_init(&sem_arm_dsp_empty, 0, 0);
	sem_init(&sem_arm_dsp_notempty, 0, 0);
	sem_init(&sem_arm_dsp_flush, 0, 0);
	sem_init(&sem_dsp_arm_empty, 0, 0);
	sem_init(&sem_dsp_arm_notempty, 0, 0);
	sem_init(&sem_dsp_arm_stateswitch, 0, 0);

	// Intialize flags
	flush_bf_arm_dsp = 0;
	cntrl_bf_arm_dsp = 0;
	bf_dsp_arm_accept_less = 0;

	DBG("Leaving ARMAMBE_Create\n") ;

	return status ;
}

/**
 * ----------------------------------------------------------------------------
 * @func  	ARMAMBE_FlushSendBuffer
 * @brief  	Flushes the send buffer to the DSP.
 * @arg		None
 * @note 	Should not be called from ARMAMBE_SendBuffer.
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Void
ARMAMBE_FlushSendBuffer (Void)
{
	Int size;

	DBG("Entered ARMAMBE_FlushSendBuffer\n") ;

	pthread_mutex_lock(&mtx_arm_dsp);
	size = bf_size(bf_arm_dsp);
	pthread_mutex_unlock(&mtx_arm_dsp);

	// Set the flush_bf_arm_dsp flag while the Send FIFO is not empty
	while (size > 0 && arm_running) {
		DBG("Flushing GPP\n");
		if (!flush_bf_arm_dsp)
			flush_bf_arm_dsp = 1;

		// Signal waiting threads that the FIFO is not empty
		sem_post(&sem_arm_dsp_notempty);
		// Wait for signal
		sem_wait(&sem_arm_dsp_flush);
		pthread_mutex_lock(&mtx_arm_dsp);
		size = bf_size(bf_arm_dsp);
		pthread_mutex_unlock(&mtx_arm_dsp);
	}

	DBG("Leaving ARMAMBE_FlushSendBuffer\n") ;
}


/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_SendBuffer
 * @brief  	Puts the specified bytes in the Send FIFO to the DSP.
 * @details
 * @arg  	buffer
 * 				array of chars to be sent
 * 			len
 * 				length of buffer
 * 			control
 * 				indicates if buffer contains control message. 1 for TRUE, 0
 * 			otherwise
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Int
ARMAMBE_SendBuffer (Char8 * buffer, Uint32 len, Char8 control)
{
	Int ret;
	Int space;
	Int size;
	Int pre_size;

	DBG("Entered ARMAMBE_SendBuffer\n") ;
	DBG("Sending %d bytes\n", (int)len);

	if (len > BS_BUF_LEN8) {
		INFO("bs_input len too long %d\n", (int)len);
		len = BS_BUF_LEN8;
	}

	if (control) {
		DBG("Preparing to send Control bytes\n");
		ARMAMBE_FlushSendBuffer();
	}

	pthread_mutex_lock(&mtx_arm_dsp);
	space = bf_space(bf_arm_dsp);
	pthread_mutex_unlock(&mtx_arm_dsp);

	// If space in Send FIFO is less than requested length wait for more
	// space to become available
	while (space < len && arm_running) {
		DBG("Waiting on Semaphore, bf_arm_dsp space = %d bytes\n", space);
		sem_wait(&sem_arm_dsp_empty);
		pthread_mutex_lock(&mtx_arm_dsp);
		space = bf_space(bf_arm_dsp);
		pthread_mutex_unlock(&mtx_arm_dsp);
	}

	pthread_mutex_lock(&mtx_arm_dsp);
	// size before putting data of length len in buffer
	pre_size = bf_size(bf_arm_dsp);
	ret = bf_put_n(bf_arm_dsp, buffer, len);
	pthread_mutex_unlock(&mtx_arm_dsp);
	if (ret != len)
		INFO("bf get failed\n");

	// Set cntrl_bf_arm_dsp only after control buffer has been put in FIFO
	if (control) {
		cntrl_bf_arm_dsp = 1;
		DBG("Forcing Control send\n");
		// Buffer is always flushed before sending a control packet
		ARMAMBE_FlushSendBuffer();
	}

	pthread_mutex_lock(&mtx_arm_dsp);
	size = bf_size(bf_arm_dsp);
	pthread_mutex_unlock(&mtx_arm_dsp);

	// Send "not empty" signal only if the size of the FIFO was previously less
	// than the threshold and is now greater than the threshold.
	if (size >= BLOB_LEN && pre_size < BLOB_LEN)
		sem_post(&sem_arm_dsp_notempty);

	DBG("Sent %d bytes\n", (int)ret);
	DBG("Leaving ARMAMBE_SendBuffer\n") ;

	return ret;
}


/**
 * ----------------------------------------------------------------------------
 * @func	ARMAMBE_ReceiveBuffer
 * @brief   Receive specified length of bytes from DSP
 * @details
 * @arg		buffer
 * 				buffer to store received bytes in
 * 			len
 * 				number of bytes to receive
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Int
ARMAMBE_ReceiveBuffer (Char8 * buffer, Uint32 len)
{
	Int ret;
	Int space;
	Int size;

	DBG("Entered ARMAMBE_ReceiveBuffer\n") ;
	DBG("Requesting %d bytes\n", (int)len);

	if (len > BS_BUF_LEN8) {
		INFO("bs_input len too long %d\n", (int)len);
		len = BS_BUF_LEN8;
	}

	pthread_mutex_lock(&mtx_dsp_arm);
	size = bf_size(bf_dsp_arm);
	pthread_mutex_unlock(&mtx_dsp_arm);

	// If there isn't enough data in the Receive FIFO to fill this request
	// or bf_dsp_arm_accept_less flag is not set then wait for more data to
	// be put into the Receive FIFO.
	while ((size < len) && !bf_dsp_arm_accept_less && arm_running) {
		sem_wait(&sem_dsp_arm_notempty);
		pthread_mutex_lock(&mtx_dsp_arm);
		size = bf_size(bf_dsp_arm);
		pthread_mutex_unlock(&mtx_dsp_arm);
	}

	pthread_mutex_lock(&mtx_dsp_arm);
	ret = bf_get_n(bf_dsp_arm, buffer, len);
	size = bf_size(bf_dsp_arm);
	pthread_mutex_unlock(&mtx_dsp_arm);
	if (ret != len)
		INFO("bf get failed\n");

	// The bf_dsp_arm_accept_less flag is unset iff the flag is currently
	// set, the size of the buffer is currently zero and the number of
	// bytes read in the last bf_get_n call was zero. If we don't check
	// that ret is zero, the calling process may not exit as it depends on
	// zero being returned to stop processing.
	if (bf_dsp_arm_accept_less && size == 0 && ret == 0) {
		bf_dsp_arm_accept_less = 0;
		sem_post(&sem_dsp_arm_stateswitch);
	}

	pthread_mutex_lock(&mtx_dsp_arm);
	space = bf_space(bf_dsp_arm);
	pthread_mutex_unlock(&mtx_dsp_arm);

	// Send 'empty' signal if space in FIFO exceeds or equals threshold
	// @TODO: Might need to condition on pre_space too
	if (space >= BLOB_LEN) 
		sem_post(&sem_dsp_arm_empty);		

	DBG("Got %d bytes\n", (int)ret);
	DBG("Leaving ARMAMBE_ReceiveBuffer\n") ;

	return ret;
}

/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_SendDspTask
 * @brief   Thread (task) responsible for sending bytes from the Send FIFO to
 * 			the DSP using DSP Bios Link channels.
 * @details
 * @arg  	arg
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Void *
ARMAMBE_SendDspTask (Void * arg)
{
	DSP_STATUS status = DSP_SOK;
	struct bs_packet *pkt;
	Int ret;
	Int size;

	DBG("Starting ARMAMBE_SendDspTask\n") ;

	// Signify that this task started successfully
	SendToDspTask_started = 1;

	// Loop forever unless error occurs or ARM shuts down
	while (status == DSP_SOK && arm_running) {
		pkt = (struct bs_packet *) ARMAMBE_Send_IOReq.buffer ;

		pthread_mutex_lock(&mtx_arm_dsp);
		size = bf_size(bf_arm_dsp);
		pthread_mutex_unlock(&mtx_arm_dsp);

		// If there isn't enough data to push into the channel and a flush has
		// not be requested, wait for more data to be put in FIFO
		while (size < BLOB_LEN && !flush_bf_arm_dsp && arm_running) {
			sem_wait(&sem_arm_dsp_notempty);
			pthread_mutex_lock(&mtx_arm_dsp);
			size = bf_size(bf_arm_dsp);
			pthread_mutex_unlock(&mtx_arm_dsp);
		}

		// Check that application didn't issue a shutdown while task was
		// waiting for more data
		if (!arm_running) 
			break;

		pthread_mutex_lock(&mtx_arm_dsp);
		ret = bf_get_n(bf_arm_dsp, pkt->data, BLOB_LEN);
		pthread_mutex_unlock(&mtx_arm_dsp);
		pkt->len = ret;
		pkt->hdr = PKT_HDR_DATA;
		if (cntrl_bf_arm_dsp)
			pkt->hdr = PKT_HDR_CNTRL;

		DBG("Sending to DSP: pkt->len = %d, pkt->hdr = %x, bf_arm_dsp size = %d\n",pkt->len,pkt->hdr,bf_size(bf_arm_dsp));

		// Send data to DSP by issuing 'filled' buffer to the channel.
		DBG("Issuing filled buffer [%p]\n", ARMAMBE_Send_IOReq.buffer);
		status = CHNL_issue (procId, CHNL_ID_OUTPUT, &ARMAMBE_Send_IOReq) ;
		if (DSP_FAILED (status)) {
			ERROR("CHNL_issue failed (output). Status = [0x%x]\n", (int)status) ;
		}

		// Reclaim 'empty' buffer from the channel
		if (DSP_SUCCEEDED (status)) {
			status = CHNL_reclaim (procId, CHNL_ID_OUTPUT, WAIT_FOREVER, &ARMAMBE_Send_IOReq) ;
			DBG("Reclaimed empty buffer [%p]\n", ARMAMBE_Send_IOReq.buffer);
			if (DSP_FAILED (status)) {
				ERROR("CHNL_reclaim failed (output). Status = [0x%x]\n", (int)status) ;
			}
		}

		// Reset control packet flag
		cntrl_bf_arm_dsp = 0;

		pthread_mutex_lock(&mtx_arm_dsp);
		size = bf_size(bf_arm_dsp);
		pthread_mutex_unlock(&mtx_arm_dsp);
		DBG("bf_arm_dsp size = %d\n",size);

		// If flush buffer was requested and the Send FIFO is empty, then unset
		// flush_bf_arm_dsp flag and signal that the Send FIFO has been flushed
		if (flush_bf_arm_dsp && size == 0) {
			flush_bf_arm_dsp = 0;
			sem_post(&sem_arm_dsp_flush);
		}

		// @TODO might have to do an else here so that sem_post(sem_arm_dsp) is
		// only called when non control data is sent.
		sem_post(&sem_arm_dsp_empty);
	}

	DBG("Ending ARMAMBE_SendTask\n") ;

	return (Void *) 0;
}


/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_ReceiveDspTask
 * @brief   Thread (task) responsible for receiving bytes from the DSP using
 * 			the DSP Bios Link channels and putting them in the Receive FIFO
 * @details
 * @arg  	arg
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Void *
ARMAMBE_ReceiveDspTask (Void * arg)
{
	DSP_STATUS status = DSP_SOK;
	AMBE_OPTIONS * ao;
	struct bs_packet *pkt;
	Int ret;
	Int space;
	Int size;

	DBG("Starting ARMAMBE_ReceiveDspTask\n") ;

	// Signify that this task started successfully
	receiveFromDspTask_started = 1;

	// Loop forever unless error occurs or ARM shuts down
	while (status == DSP_SOK && arm_running) {
		pthread_mutex_lock(&mtx_dsp_arm);
		space = bf_space(bf_dsp_arm);
		pthread_mutex_unlock(&mtx_dsp_arm);

		// If there isn't enough space in buffer to put received data, wait for
		// more data to be removed from Receive FIFO
		while (space < BLOB_LEN && arm_running) {
			sem_wait(&sem_dsp_arm_empty);
			pthread_mutex_lock(&mtx_dsp_arm);
			space = bf_space(bf_dsp_arm);
			pthread_mutex_unlock(&mtx_dsp_arm);
		}

		// Check that application didn't issue a shutdown while task was
		// waiting for more data
		if (!arm_running)
			break;

		// Receive data from DSP by issuing 'empty' buffer to the channel.
		if (DSP_SUCCEEDED (status)) {
			DBG("Issuing empty buffer [%p]\n", ARMAMBE_Receive_IOReq.buffer);
			status = CHNL_issue (procId, CHNL_ID_INPUT, &ARMAMBE_Receive_IOReq) ;
			if (DSP_FAILED (status)) {
				ERROR("CHNL_issue failed (input). Status = [0x%x]\n", (int)status) ;
			}
		}

		// Reclaim 'filled' buffer from the channel
		if (DSP_SUCCEEDED (status)) {
			status = CHNL_reclaim (procId, CHNL_ID_INPUT, WAIT_FOREVER, &ARMAMBE_Receive_IOReq) ;
			DBG("Reclaimed filled buffer [%p]\n", ARMAMBE_Receive_IOReq.buffer);
			if (DSP_FAILED (status)) {
				ERROR("CHNL_reclaim failed (input). Status = [0x%x]\n", (int)status) ;
			}
		}


		// Process received buffer
		if (status == DSP_SOK) {
			pkt = (struct bs_packet *) ARMAMBE_Receive_IOReq.buffer;

			DBG("Received from DSP: pkt->len = %d, pkt->hdr = %x\n",pkt->len,pkt->hdr);

			if (pkt->hdr == PKT_HDR_CNTRL) { 
				// The DSP currently only sends a control packet as an ACK to
				// the ARM's request to STOP processing data.
				ao = (AMBE_OPTIONS *) pkt->data;
				if (!ao->run_encoder && !ao->run_decoder)
					dsp_ack_stop = 1;

				DBG("Received Control Packet, ao->run_encoder = %d, ao->run_decoder = %d\n",
						ao->run_encoder, ao->run_decoder);

				// Flush Receive buffer before taking any further actions
				do {
					DBG("Waiting for buffer to be emptied [size = %d]\n",bf_size(bf_dsp_arm));
					// Set bf_dsp_arm_accept_less flag to force
					// ARMAMBE_ReceiveBuffer to return data even if the size is
					// less than what was requested.
					bf_dsp_arm_accept_less = 1;

					// Send signal to notify processes waiting for data in
					// Receive FIFO to proceed
					sem_post(&sem_dsp_arm_notempty);

					// This task will be notified when ARMAMBE_ReceiveBuffer
					// has emptied the Receive FIFO
					sem_wait(&sem_dsp_arm_stateswitch);

					pthread_mutex_lock(&mtx_dsp_arm);
					size = bf_size(bf_dsp_arm);
					pthread_mutex_unlock(&mtx_dsp_arm);
				} while (size > 0);

				// If this is a STOP control packet don't put it in the FIFO
				if (!ao->run_encoder && !ao->run_decoder)
					continue;
			}

			pthread_mutex_lock(&mtx_dsp_arm);
			ret = bf_put_n(bf_dsp_arm, pkt->data, pkt->len);
			pthread_mutex_unlock(&mtx_dsp_arm);
			if (ret != pkt->len)
				INFO("bf import failed\n");

			// Signal waiting processes that the FIFO is not empty
			sem_post(&sem_dsp_arm_notempty);
		}
	}

	DBG("Ending ARMAMBE_ReceiveDspTask\n") ;

	return (Void *) 0;
}


/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_SendCryptBuffer
 * @brief   Send buffer to DSP on its cryptography related channel
 * @details	Data buffer either contains encrypted AMBE voice data (when in TX
 * 			mode) or decrypted AMBE voice data (when in RX mode)
 * @arg  	buffer
 * 				buffer of bytes
 * 			len
 * 				number of bytes to send
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Int
ARMAMBE_SendCryptBuffer (Char8 * buffer, Uint32 len)
{
	DSP_STATUS status = DSP_SOK;
	struct bs_packet *pkt;

	DBG("Starting ARMAMBE_SendCryptBuffer\n") ;

	// Counter for number of sent bytes
	int sent = 0;

	while (status == DSP_SOK && sent < len) {
		pkt = (struct bs_packet *) ARMAMBE_SendC_IOReq.buffer ;

		pkt->hdr = PKT_HDR_DATA;

		// Shorten pkt length to fit in channel buffer
		pkt->len = len-sent;
		if (pkt->len > BLOB_LEN)
			pkt->len = BLOB_LEN;

		memcpy(pkt->data, buffer+sent, pkt->len);
		sent += pkt->len;

		DBG("Sending to DSP: pkt->len = %d, pkt->hdr = %x\n",pkt->len,pkt->hdr);

		// Send data to DSP by issuing 'filled' buffer to the channel.
		DBG("Issuing filled buffer [%p]\n", ARMAMBE_SendC_IOReq.buffer);
		status = CHNL_issue (procId, CHNL_ID_COUTPUT, &ARMAMBE_SendC_IOReq) ;
		if (DSP_FAILED (status)) {
			ERROR("CHNL_issue failed (output). Status = [0x%x]\n", (int)status) ;
		}

		// Reclaim 'empty' buffer from the channel
		if (DSP_SUCCEEDED (status)) {
			status = CHNL_reclaim (procId, CHNL_ID_COUTPUT, WAIT_FOREVER, &ARMAMBE_SendC_IOReq) ;
			DBG("Reclaimed empty buffer [%p]\n", ARMAMBE_SendC_IOReq.buffer);
			if (DSP_FAILED (status)) {
				ERROR("CHNL_reclaim failed (output). Status = [0x%x]\n", (int)status) ;
			}
		}
	}

	DBG("Sent %d bytes\n", sent);
	DBG("Leaving ARMAMBE_SendCryptBuffer\n") ;

	return sent;
}


/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_ReceiveCryptBuffer
 * @brief   Receive buffer from DSP on its cryptography related channel
 * @details	Buffer will contain AMBE voice data to be encrypted (when in TX
 * 			mode) or AMBE voice data to be decrypted (when in RX mode)
 * @arg  	buffer
 * 				buffer of bytes
 * 			len
 * 				number of bytes to receive
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Int
ARMAMBE_ReceiveCryptBuffer (Char8 * buffer, Uint32 len)
{
	DSP_STATUS status = DSP_SOK;
	struct bs_packet *pkt;

	int recv = 0;
	int ret;

	DBG("Starting ARMAMBE_ReceiveCryptBuffer\n") ;
	DBG("Requesting %d bytes\n", (int)len);

	while (status == DSP_SOK && recv < len) {
		// Receive data from DSP by issuing 'empty' buffer to the channel.
		if (DSP_SUCCEEDED (status)) {
			DBG("Issuing empty buffer [%p]\n", ARMAMBE_ReceiveC_IOReq.buffer);
			status = CHNL_issue (procId, CHNL_ID_CINPUT, &ARMAMBE_ReceiveC_IOReq) ;
			if (DSP_FAILED (status)) {
				ERROR("CHNL_issue failed (input). Status = [0x%x]\n", (int)status) ;
			}
		}

		// Reclaim 'filled' buffer from the channel
		if (DSP_SUCCEEDED (status)) {
			status = CHNL_reclaim (procId, CHNL_ID_CINPUT, WAIT_FOREVER, &ARMAMBE_ReceiveC_IOReq) ;
			DBG("Reclaimed filled buffer [%p]\n", ARMAMBE_ReceiveC_IOReq.buffer);
			if (DSP_FAILED (status)) {
				ERROR("CHNL_reclaim failed (input). Status = [0x%x]\n", (int)status) ;
			}
		}


		// Process received buffer
		if (status == DSP_SOK) {
			pkt = (struct bs_packet *) ARMAMBE_ReceiveC_IOReq.buffer;

			DBG("Received from DSP: pkt->len = %d, pkt->hdr = %x\n",pkt->len,pkt->hdr);

			if (pkt->hdr == CRYPT_END) { 
				DBG("Received Control Packet\n");
				return pkt->hdr;
			}

			// Shorten pkt length to enough bytes to meet the requested size
			if (pkt->len > len-recv)
				pkt->len = len-recv;

			memcpy(buffer+recv, pkt->data, pkt->len);
			recv += pkt->len;

			ret = pkt->hdr;
		}
	}

	DBG("Got %d bytes\n", recv);
	DBG("Ending ARMAMBE_ReceiveCryptBuffer\n") ;

	return ret;
}

/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_Execute
 * @brief   Initializes and begins execution of the ARM application
 * @details	Also starts execution of the DSP
 * @arg  	none
 * @ret		DSP_SOK
 *              Operation Successfully completed.
 *          DSP_EFAIL
 *              GppAmbe execution failed.
 * @see		ARMAMBE_Delete , ARMAMBE_Create
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
DSP_STATUS
ARMAMBE_Execute (Void)
{
	DSP_STATUS status = DSP_SOK ;
	int res;
	pthread_attr_t attr;

	DBG("Entered ARMAMBE_Execute\n") ;

	// Initialize flags
	SendToDspTask_started = 0;
	receiveFromDspTask_started = 0;

	// Start execution on DSP.
	status = PROC_start (procId) ;

	// Fill the IO Request structure
	// It gives Information for adding or reclaiming an input request.
	if (DSP_SUCCEEDED (status)) {
		ARMAMBE_Send_IOReq.buffer = ARMAMBE_Buffers [0];
		ARMAMBE_Send_IOReq.size   = ARMAMBE_BufferSize;
		ARMAMBE_Receive_IOReq.buffer = ARMAMBE_Buffers [1];
		ARMAMBE_Receive_IOReq.size   = ARMAMBE_BufferSize;

		ARMAMBE_SendC_IOReq.buffer = ARMAMBE_Buffers [2];
		ARMAMBE_SendC_IOReq.size   = ARMAMBE_BufferSize;
		ARMAMBE_ReceiveC_IOReq.buffer = ARMAMBE_Buffers [3];
		ARMAMBE_ReceiveC_IOReq.size   = ARMAMBE_BufferSize;
	} else {
		ERROR("PROC_start failed. Status = [0x%x]\n", (int)status) ;
	}

	pthread_mutex_init(&mtx_dsp_arm, NULL);
	pthread_mutex_init(&mtx_arm_dsp, NULL);

	// Create threads for Send and Receive Tasks
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	res = pthread_create(&thr_dsp_arm, &attr, ARMAMBE_ReceiveDspTask, (void *)0);

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	res = pthread_create(&thr_arm_dsp, &attr, ARMAMBE_SendDspTask, (void *)0);

	DBG("Waiting for all GPP Tasks to get started\n");
	while (!SendToDspTask_started && !receiveFromDspTask_started)
		sched_yield();

	DBG("Leaving ARMAMBE_Execute ()\n") ;

	return status ;
}

/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_Delete
 * @brief   This function releases resources allocated earlier by call to
 *         	ARMAMBE_Create ().
 * @details	During cleanup, the allocated resources are freed unconditionally.
 * @ret		DSP_SOK
 *              Operation Successfully completed.
 *          DSP_EFAIL
 *              Resource deallocation failed.
 * @see	ARMAMBE_Create
 * @author 	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Void
ARMAMBE_Delete ()
{
	DSP_STATUS status    = DSP_SOK ;
	DSP_STATUS tmpStatus = DSP_SOK ;

	DBG("Entered ARMAMBE_Delete\n") ;

	CHNL_flush (procId, CHNL_ID_INPUT);
	CHNL_flush (procId, CHNL_ID_CINPUT);
	CHNL_flush (procId, CHNL_ID_OUTPUT);
	CHNL_flush (procId, CHNL_ID_COUTPUT);

	sem_post(&sem_arm_dsp_empty);
	sem_post(&sem_arm_dsp_notempty);
	sem_post(&sem_arm_dsp_flush);

	sem_post(&sem_dsp_arm_empty);
	sem_post(&sem_dsp_arm_notempty);
	sem_post(&sem_dsp_arm_stateswitch);

	pthread_join(thr_dsp_arm, NULL);
	pthread_join(thr_arm_dsp, NULL);

	// Free buffers allocated for channel to DSP
	tmpStatus = CHNL_freeBuffer (procId, CHNL_ID_OUTPUT, ARMAMBE_Buffers, 1) ;
	if (DSP_SUCCEEDED (status) && DSP_FAILED (tmpStatus)) {
		ERROR("CHNL_freeBuffer () failed (output). Status = [0x%x]\n", (int)tmpStatus) ;
	}

	//Delete all input and output channels
	tmpStatus = CHNL_delete  (procId, CHNL_ID_INPUT) ;
	if (DSP_SUCCEEDED (status) && DSP_FAILED (tmpStatus)) {
		ERROR("CHNL_delete () failed (input). Status = [0x%x]\n", (int)tmpStatus) ;
	}
	tmpStatus = CHNL_delete  (procId, CHNL_ID_CINPUT) ;
	if (DSP_SUCCEEDED (status) && DSP_FAILED (tmpStatus)) {
		ERROR("CHNL_delete () failed (cinput). Status = [0x%x]\n", (int)tmpStatus) ;
	}
	tmpStatus = CHNL_delete  (procId, CHNL_ID_OUTPUT) ;
	if (DSP_SUCCEEDED (status) && DSP_FAILED (tmpStatus)) {
		ERROR("CHNL_delete () failed (output). Status = [0x%x]\n", (int)tmpStatus) ;
	}
	tmpStatus = CHNL_delete  (procId, CHNL_ID_COUTPUT) ;
	if (DSP_SUCCEEDED (status) && DSP_FAILED (tmpStatus)) {
		ERROR("CHNL_delete () failed (coutput). Status = [0x%x]\n", (int)tmpStatus) ;
	}

	// Stop execution on DSP.
	status = PROC_stop (procId) ;

	// Close the pool
	tmpStatus = POOL_close (POOL_makePoolId(procId, POOL_ID)) ;
	if (DSP_SUCCEEDED (status) && DSP_FAILED (tmpStatus)) {
		ERROR("POOL_close () failed. Status = [0x%x]\n", (int)tmpStatus) ;
	}

	// Detach from the processor
	tmpStatus = PROC_detach  (procId) ;
	if (DSP_SUCCEEDED (status) && DSP_FAILED (tmpStatus)) {
		ERROR("PROC_detach () failed. Status = [0x%x]\n", (int)tmpStatus) ;
	}

	// Destroy the PROC object.
	tmpStatus = PROC_destroy () ;
	if (DSP_SUCCEEDED (status) && DSP_FAILED (tmpStatus)) {
		ERROR("PROC_destroy () failed. Status = [0x%x]\n", (int)tmpStatus) ;
	}

	DBG("Leaving ARMAMBE_Delete\n") ;
}


/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_Main
 * @brief   Entry point for the application
 * @details
 * @arg		processorId
 * 				Id of the DSP to use
 * 			bufferSize
 * 				default buffer size for the buffers in this application
 * 				processorId
 * @ret		DSP_SOK
 * @see		ARMAMBE_Create, ARMAMBE_Execute, ARMAMBE_Delete
 * @author 	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */

DSP_STATUS
ARMAMBE_Main (Uint8 processorId, Uint32 bufferSize) {
	DSP_STATUS status       = DSP_SOK ;

	DBG("Entering ARMAMBE_Main\n") ;

	// Validate the buffer size specified.
	ARMAMBE_BufferSize = DSPLINK_ALIGN (bufferSize, DSPLINK_BUF_ALIGN) ;
	if (ARMAMBE_BufferSize == 0) {
		status = DSP_ESIZE ;
	}

	DBG("GPP Buffer Size = %d\n", (int)ARMAMBE_BufferSize);

	if (processorId >= MAX_DSPS) {
		ERROR("Error: Invalid processor id  specified %d ==\n", processorId) ;
		status = DSP_EFAIL ;

	}

	procId = processorId;

	DBG("Leaving ARMAMBE_Main\n") ;

	return status;
}

#if defined (__cplusplus)
}
#endif /* defined (__cplusplus) */
