/**
 * ============================================================================
 * @file	arm_ambe.h
 * @brief	Defines the configurable parameters for arm_ambe.c
 * @note	none.
 *
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */

#if !defined (ARMAMBE_H)
#define ARMAMBE_H

/* ---------------------------- DSP/BIOS Link ------------------------------ */
#include <dsplink.h>


#if defined (__cplusplus)
extern "C" {
#endif /* defined (__cplusplus) */


/**
 * ----------------------------------------------------------------------------
 * @const	CHNL_ID_OUTPUT
 * @brief	ID of channel used to send PCM data to DSP for encoding.
 * ----------------------------------------------------------------------------
 */
#define CHNL_ID_OUTPUT     0


/**
 * ----------------------------------------------------------------------------
 * @const	CHNL_ID_INPUT
 * @brief	ID of channel used to receive encoded voice data from DSP.
 * ----------------------------------------------------------------------------
 */
#define CHNL_ID_INPUT      1

/**
 * ----------------------------------------------------------------------------
 * @const	CHNL_ID_COUTPUT
 * @brief	ID of channel used to send encrypted voice data to DSP.
 * ----------------------------------------------------------------------------
 */
#define CHNL_ID_COUTPUT    2

/**
 * ----------------------------------------------------------------------------
 * @const	CHNL_ID_CINPUT
 * @brief	ID of channel used to receive encrypted voice data from DSP.
 * ----------------------------------------------------------------------------
 */
#define CHNL_ID_CINPUT     3


/**
 * ----------------------------------------------------------------------------
 * @func	ARMAMBE_Create
 * @brief	This function allocates and initializes resources used by
 *          this application.
 * ----------------------------------------------------------------------------
 */
DSP_STATUS
ARMAMBE_Create (IN Char8 * dspExecutable);


/**
 * ----------------------------------------------------------------------------
 * @func	ARMAMBE_Execute
 * @brief	This function implements the execute phase for this application.
 * ----------------------------------------------------------------------------
 */
DSP_STATUS
ARMAMBE_Execute ();

/**
 * ----------------------------------------------------------------------------
 * @func  	ARMAMBE_FlushSendBuffer
 * @brief  	Flushes the send buffer to the DSP.
 * ----------------------------------------------------------------------------
 */
Void
ARMAMBE_FlushSendBuffer (Void);

/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_SendDspTask
 * @brief   Thread (task) responsible for sending bytes from the Send FIFO to
 * 			the DSP using DSP Bios Link channels.
 * ----------------------------------------------------------------------------
 */
Void *
ARMAMBE_SendDspTask (Void * arg);

/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_ReceiveDspTask
 * @brief   Thread (task) responsible for receiving bytes from the DSP using
 * 			the DSP Bios Link channels and putting them in the Receive FIFO
 * ----------------------------------------------------------------------------
 */
Void *
ARMAMBE_ReceiveDspTask (Void * arg);


/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_SendBuffer
 * @brief  	Puts the specified bytes in the Send FIFO to the DSP.
 * ----------------------------------------------------------------------------
 */
Int
ARMAMBE_SendBuffer (Char8 * buffer, Uint32 len, Char8 control);

/**
 * ----------------------------------------------------------------------------
 * @func	ARMAMBE_ReceiveBuffer
 * @brief   Receive specified length of bytes from DSP
 * ----------------------------------------------------------------------------
 */
Int
ARMAMBE_ReceiveBuffer (Char8 * buffer, Uint32 len);

/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_SendCryptBuffer
 * @brief   Send buffer to DSP on its cryptography related channel
 * ----------------------------------------------------------------------------
 */
Int
ARMAMBE_SendCryptBuffer (Char8 * buffer, Uint32 len);

/**
 * ----------------------------------------------------------------------------
 * @func   	ARMAMBE_ReceiveCryptBuffer
 * @brief   Receive buffer from DSP on its cryptography related channel
 * ----------------------------------------------------------------------------
 */
Int
ARMAMBE_ReceiveCryptBuffer (Char8 * buffer, Uint32 len);

/**
 * ----------------------------------------------------------------------------
 * @func	ARMAMBE_Delete
 * @brief	This function releases resources allocated earlier by call to
 *          ARMAMBE_Create().
 * ----------------------------------------------------------------------------
 */
Void
ARMAMBE_Delete ();


/**
 * ----------------------------------------------------------------------------
 * @func	ARMAMBE_Main
 * @brief	The OS independent driver function for the arm_ambe sample
 * 			application.
 * ----------------------------------------------------------------------------
 */
DSP_STATUS
ARMAMBE_Main (IN Uint8 processorId, IN Uint32  bufferSize);


#if defined (__cplusplus)
}
#endif /* defined (__cplusplus) */


#endif /* !defined (ARMAMBE_H) */
