/**
 * ============================================================================
 * @file 	gpp_host_intf.c
 * @brief	Implementation of ARM to host and host to ARM function for both
 *			vocoding and encryption.
 * @note	none.
 *
 * @author	Cal(it)2
 * @author  Jeffrey Cuenco <jcuenco@ucsd.edu>
 * @author  Justyn Bell <jwbell@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 *
 * @endverbatim
 * ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "tcplib.h"
#include "logger.h"
#include "gpp_host_intf.h"
#include "arm_ambe.h"
#include "ambe_intf.h"
#include "common.h"

// From main.c
extern unsigned int exprcv_bytes;
extern unsigned int rcv_bytes;
extern char arm_running; // 1 or 0. Accessed by all ARMAMBE_* threads
extern char dsp_ack_stop; // 1 or 0. Accessed by ARMAMBE_ReceiveDspTask
extern AMBE_OPTIONS ao;
extern PC_OPTIONS opts;

extern sem_t sem_processDspArm;
extern sem_t sem_processCryptography;

/** @brief State variable for input thread */
int input_finished = 0;
/** @brief State variable for output thread */
int output_finished = 0;

/** @brief Global variables for start state */
int gh_started = 0, gh_crypt_started = 0;

/** @brief Global socket pointers for Encoding/Decoding data flow */
int connfd, crypt_connfd, host_connected = 0, crypt_host_connected = 0;

/**
  * @details Shuts down the socket connection and resets relevant state variables.
  * 
  * @see gh_started
  * @see gh_crypt_started
  */
void gh_intf_shutdown(void)
{
        INFO("Gracefully shutting down...\n");
        gh_started = 0;
	gh_crypt_started = 0;

	// Close connection and TCP server
//	close_tcp_server_connection(connfd);
//	stop_tcp_server(sockfd);
}

/**
  * @details Reads control packet from host from TCP stream.
  *
  * @param connfd  The socket from which data is read.
  * @return Returns the state of the stream.
  * @see I_SOCK_ERROR
  * @see I_SOCK_DISCONNECT
  */
enum imbe_ctrl_t read_ctrl_pkt(int connfd)
{
	int res;
	enum imbe_ctrl_t ctrl;

	res = read_tcp_server_stream(connfd, (char *)&ctrl, sizeof(ctrl));
	// Disconnect from socket if there is an error, and return I_SOCK_ERROR
	if (res < 0) {
		ERROR("ERROR, reading from host ");
		close_tcp_server_connection(connfd);
		return I_SOCK_ERROR;
	}
	// Disconnect from socket if no data is read, and return I_SOCK_DISCONNECT
	else if (res == 0) {
		INFO("Client disconnected.\n");
		close_tcp_server_connection(connfd);
		return I_SOCK_DISCONNECT;
	}
    else {
        if (ctrl == I_SOCK_ERROR) {
            INFO("Got I_SOCK_ERROR that should be IMBE_UNKNOWN\n");
            ctrl = IMBE_UNKNOWN;
        }
    }

	return ctrl;
}

/**
  * @details Reads data payload from host from TCP stream.
  *
  * @param connfd The socket from which data is read.
  * @param buf The buffer in which the data payload is stored.
  * @param len The number of bytes to be read from the socket.
  * @return Returns the actual number of bytes read from the socket if successful, -1 if there is an error on the socket
  *		or 0 if no data is read.
  */
int read_data_payload(int connfd, char *buf, int len)
{
	int res;

	res = read_tcp_server_stream(connfd, buf, len);
	// Disconnect from socket if there is an error
	if (res < 0) {
		ERROR("ERROR, reading from host ");
		close_tcp_server_connection(connfd);
		return -1;
	}
	// Disconnect from socket if was no data read
	else if (res == 0) {
		INFO("Client disconnected.\n");
		close_tcp_server_connection(connfd);
		return -1;
	}

	return res;
}

/**
  * @details Writes data payload to host on TCP stream.
  *
  * @param connfd The socket on which data is written.
  * @param buf The buffer which contains the data to be written to the socket.
  * @param len The number of bytes to be written to the socket.
  * @return Returns the number of bytes actually written to the socket if successful, -1 if there is an error on the socket
  *		or 0 if no data was written.
  */
int write_data_payload(int connfd, char *buf, int len)
{
	if (len == 0) return 0;
        int res;

        res = send_tcp_client_data(connfd, buf, len);
        if (res < 0) {
                ERROR("ERROR, writing to host\n");
		close_tcp_server_connection(connfd);
                return -1;
        }
        else if (res == 0) {
                INFO("Client disconnected.\n");
		close_tcp_server_connection(connfd);
                return -1;
        }

        return res;
}

/** @brief From AMBE test file.
 * convert binary encoder channel data output to desired soft decision size
 */
static void enc_cvt_sdbits( short *cbuf, short len, short sd_bits ) {
	short i;
	for ( i = 0; i < len; i++ )
		cbuf[i] *= ((1 << sd_bits) - 1);
}

/** @brief From AMBE test file.
 * convert decoder channel data input to desired soft decision level
 */
static void dec_cvt_sdbits( short *t, short len, short sd_bits ) {
	short sd_shift;
	short i;

	sd_shift = sd_bits - 1;

	for ( i = 0; i < len; i++ )
		t[i] = t[i] >> sd_shift;
}

/**
  * @details Sets the default options.  Then, based on the control message from the host, sets the vocoder
  *		to either encode, or decode, and sets the state of encryption.
  * 
  * @param o The AMBE vocoder options struct.
  * @param ctrl The control message used to set the AMBE parameters, as well as encryption.
  */
int set_ambe_options(AMBE_OPTIONS *o, enum imbe_ctrl_t ctrl) 
{
	/*
	 * set default options and set encoder/decoder based on the control message from the host
	 */
	o->verbose 		=  0;
	o->mode 		= APCO_MODE;
	o->fec_mode 		= APCO_MODE;			
	o->rep_mode 		= -1;
	o->enc_ws 		= 2*AMBE_WS_NOM;
	o->dec_ws	 	= 2*AMBE_WS_NOM;
	o->vad_level   		= VAD_LEVEL;
	o->nsprsn_enabled   	=  1;
	o->tone_det_enabled 	=  1;
	o->tone_on		=  0;
	o->tone_amp 		= TONE_AMP;
	o->tone_index  		= TONE_INDEX;
	o->sd_bits 		= SD_BITS;
	o->sd_dist 		=  0;
	o->hd_errs 		=  0;
	o->dtx	 		=  0;
	o->bit_steal 		=  0;
	o->cnirate 		=  0;
	o->agc_on		=  0;
	o->use_fec 		=  1;

/*	
	o->compare		= COMPARE_OFF;
	o->fpc_name 		= NULL;
	o->fpc 			= NULL;
	o->fpi_name 		= NULL;
	o->fpi 			= NULL;
	o->fpo_name 		= NULL;
	o->fpo 			= NULL;
*/
	o->arm_mode		= ARM_NORMAL_MODE;

	/* Set encoder and decoder states based on control message received */
	switch (ctrl) {
		case IMBE_ENCODE_START:
			o->run_encoder 		=  1;
			o->run_decoder 		=  0;
			o->encryption		=  0;
			break;
		case IMBE_ENCR_ENCODE_START:
			o->run_encoder 		=  1;
			o->run_decoder 		=  0;
			o->encryption		=  1;
			break;
		case IMBE_DECODE_START:
			o->run_encoder 		=  0;
			o->run_decoder 		=  1;
			o->encryption		=  0;
			break;
		case IMBE_DECR_DECODE_START:
			o->run_encoder 		=  0;
			o->run_decoder 		=  1;
			o->encryption		=  1;
			break;
		case IMBE_ENCODE_STOP:
		case IMBE_DECODE_STOP:
			/* 
			 * Stop control message: 
			 * [o->run_encoder = 0 and o->run_decoder = 0];
			 */
			o->run_encoder 		=  0;
			o->run_decoder 		=  0;
			o->encryption		=  0;
			break;
		default:
			break;
	}

//	INFO("ARM MODE = %d\n", o->arm_mode);
//	INFO("run_encoder = %d, run_decoder = %d, encryption = %d\n", o->run_encoder, o->run_decoder, o->encryption);
}

/**
 *  @brief Sends data to AMBE on DSP for encoding/decoding.
 *
 *  @param ao AMBE options as set by control message sent from host.
 *  @param buf Buffer containing data to be sent to the DSP.
 *  @param len Number of bytes to be sent to the DSP vocoder.
 */
int send_to_ambe(AMBE_OPTIONS *ao, char *buf, int len)
{
	short dttlbits;

	dttlbits = get_ttlbits(ao->rep_mode, ao->mode, ao->use_fec);

	if (ao->run_encoder) {
		// send voice samples to DSP for encoding
//		INFO("ao->run_encoder\n");
		ARMAMBE_SendBuffer(buf, ao->enc_ws*sizeof(short), 0);
		exprcv_bytes += dttlbits*sizeof(short);
		//INFO("EXPECTED RECEIVED BYTES = %d\n",exprcv_bytes);
	} 
	else if (ao->run_decoder) {
		// send encoded bits to DSP for decoding
//		INFO("ao->run_decoder\n");
		ARMAMBE_SendBuffer(buf, dttlbits*sizeof(short), 0);
		exprcv_bytes += ao->enc_ws*sizeof(short);
		//INFO("EXPECTED RECEIVED BYTES = %d\n",exprcv_bytes);
	} else {
		ERROR("Neither ao->run_encoder or ao->run_decoder is set\n");
	}
}

/**
  * @details    While a connection is established to the host side,
  * 			the loop reads first a control from the host. As long as there is no error, it proceeds.
  * 			The loop sets the relevant AMBE and encryption options based on the control message received
  * 			from the host and then starts reading payloads from the host and sending the collected data
  * 			to the actual AMBE vocoder until a STOP message is received.
  * 			Once a stop message is received from the host, the DSP is sent a stop message and the loop
  * 			waits for an acknowledgement from the DSP that all of its cleanup has been completed.
  */
/* Main Host vocoder to ARM Loop */
int gh_intf_loop(int port) 
{
    int res, sockfd;
	enum imbe_ctrl_t ctrl;
	char buf[BUFF_LEN];
	int buflen = 0;
	DSP_STATUS status;

	// State variable that keeps track of the current thread's state.  Goes to 0 when a connection is lost
	gh_started = 1;
	while (gh_started) {
		// Start TCP server (for software vocoder), listen for connections
		sockfd = start_tcp_server(port);
		INFO("Waiting for connection on port %d [sock = %d]\n", port, sockfd);
		do {
			connfd = get_tcp_server_connection(sockfd);
			usleep(100);
		} while (!connfd);
		host_connected = 1;

		// Client connected, start main processing loop
		while (host_connected) {
			// Read in control packet to determine state
			ctrl = read_ctrl_pkt(connfd);
			if (ctrl == I_SOCK_ERROR || ctrl == I_SOCK_DISCONNECT) {
				if (ctrl == I_SOCK_ERROR)
					ERROR("read_ctrl_pkt() failure!\n");
				host_connected = 0;
				break;
			}

			INFO("***** ctrl = %d\n", (int)ctrl);

			// Only valid encoding states are IMBE_ENCODE_START and IMBE_DECODE_START (with or without encryption).
			if ((ctrl == IMBE_ENCODE_START) || (ctrl == IMBE_DECODE_START) ||
			    (ctrl == IMBE_ENCR_ENCODE_START) || (ctrl == IMBE_DECR_DECODE_START)) {

				// Reset byte counters
				exprcv_bytes = 0;
				rcv_bytes = 0;

				// State variable keeps lets rest of code know there is current input processing
				input_finished = 0;		

				// Set encoder/decoder option struct based on control message
				set_ambe_options(&ao, ctrl);

				sem_post(&sem_processDspArm);
				if (ao.encryption)
					sem_post(&sem_processCryptography);

				// Send Start control message to DSP
				status = ARMAMBE_SendBuffer((char *)&ao, sizeof(ao), 1);

				// Sets correct buffer sizes based on state of host/GPP
				switch (ctrl) {
					case IMBE_ENCODE_START:
						INFO("Starting IMBE encoding mode, encryption disabled\n");
						buflen = BUFF_LEN;
						break;
					case IMBE_ENCR_ENCODE_START:
						INFO("Starting IMBE encoding mode, encryption enabled\n");
						buflen = BUFF_LEN;
						break;
					case IMBE_DECODE_START:
						INFO("Starting IMBE decoding mode, decryption disabled\n");
						buflen = E_BUFF_LEN;					
						break;
					case IMBE_DECR_DECODE_START:
						INFO("Starting IMBE decoding mode, decryption enabled\n");
						buflen = E_BUFF_LEN;					
						break;
					default:
						break;
				}

				//usleep(20000);	

				// Start processing data until stop encode/decode control signal received.
				do {
					// Read in control packet to determine state
					ctrl = read_ctrl_pkt(connfd);
					if (ctrl == I_SOCK_ERROR || ctrl == I_SOCK_DISCONNECT) {
						if (ctrl == I_SOCK_ERROR)
							ERROR("read_ctrl_pkt() failure!\n");
						host_connected = 0;
						break;
					}

					// If control state is IMBE_DATA, retrieve data payload
					else if (ctrl == IMBE_DATA) {
						//INFO("IMBE_DATA\n");

						// Read in data payload since were are in Data Encode mode
						res = read_data_payload(connfd, buf, buflen);
//						printf("Read %d bytes\n", res);
						// Exit if error on trying to read data payload
						if (res == -1) {
							ERROR("read_data_payload() failure!\n");
							host_connected = 0;
							break;
						}
						
						if (res != buflen) {
							INFO("res = %d, buflen = %d\n", res, buflen);
							res = buflen;
						}

						// AMBE encode/decode data!
						send_to_ambe(&ao, buf, res);
					}
				// Stop processing input data upon receiving a STOP from host
				} while (!((ctrl == IMBE_ENCODE_STOP) || (ctrl == IMBE_DECODE_STOP)));
				
				if (ctrl == IMBE_ENCODE_STOP) {
					INFO("IMBE_ENCODE_STOP\n");
				}
				else if (ctrl == IMBE_DECODE_STOP) {
					INFO("IMBE_DECODE_STOP\n");
				}

				// Reset state variable to reflect that we are finished processing input data from host
				input_finished = 1;
				
				// Send Stop control message to DSP
				set_ambe_options(&ao, ctrl);
				status = ARMAMBE_SendBuffer((char *)&ao, sizeof(ao), 1);

				// Waits for the DSP to send a control message which serves as an ACK.
				INFO("Waiting for DSP ACK message...\n");
				while (!dsp_ack_stop)
					sched_yield();

				INFO("Received DSP ACK message!\n");
				dsp_ack_stop = 0;

			// In case we get an initial STOP instead of START from host.  Occers when we initialized the ARM
			//		with an initial STOP message.
			} else if (ctrl == IMBE_ENCODE_STOP || ctrl == IMBE_DECODE_STOP ) {
				if (ctrl == IMBE_ENCODE_STOP) {
					INFO("IMBE_ENCODE_STOP when IMBE_ENCODE_START expected\n");
				} else if (ctrl == IMBE_DECODE_STOP) {
					INFO("IMBE_DECODE_STOP when IMBE_DECODE_START expected\n");
				}
			
				// Write control header to Host
				ctrl = HOST_STOP_STREAM;
				if (write_data_payload(connfd, (char *)&ctrl, sizeof(ctrl)) < 0) {
					ERROR("ERROR writing stop stream control packet to host\n");
					host_connected = 0;
				}
				INFO("Sent HOST_STOP_STREAM!\n");
			}
		}		

		// Wait for output thread to complete
		while (!output_finished)
			sched_yield();

		//output_finished = 0;

		INFO("Stopping input TCP server...\n");
		
		// Close connection and TCP server
		close_tcp_server_connection(connfd);
		stop_tcp_server(sockfd);
	}

	return 0;
}

/**
 * @details Once the thread is started, it sends a HOST_START_STREAM to
 * 			the host side to notify that data is incoming.  It then starts receiving encoded/decoded (depending
 * 			on the mode of the vocoder) data from the DSP.  While the host is connected, and there is data available
 * 			to be sent to the host side, the loop will continue writing a DATA control and then a payload to the host.
 * 			Once the input thread has finished, the output loop tests to make sure the same number of data that
 * 			was sent to the DSP in the input thread is received from the DSP in the current output thread and
 * 			sends a STOP control to the host.
 *
 * 	@param ao AMBE control options.
 */
/* Main ARM to host vocoder loop */
int gh_intf_output(AMBE_OPTIONS * ao) 
{
	int i, ret;
	static short speech_buf[SPEECH_BUF_SZ];
	static short chan_buf[CHAN_BUF_SZ];
	char encode, decode;
	short echanbits, dec_ws;
	enum host_ctrl_t ctrl;	
	
	output_finished = 0;

	// Save ao options to local variable because this is reset
	// by gh_intf_loop when it sends the stop control packet
	encode = ao->run_encoder;
	decode = ao->run_decoder;
	dec_ws = ao->dec_ws;

	echanbits = get_ttlbits(ao->rep_mode, ao->mode, ao->use_fec);
	INFO("echanbits = %d\n",echanbits);

	if (host_connected) {
		ctrl = HOST_START_STREAM;
		if (write_data_payload(connfd, (char *)&ctrl, sizeof(ctrl)) < 0) {
			ERROR("ERROR writing control packet to host\n");
			host_connected = 0;
		}
		INFO("Wrote HOST_START_STREAM\n");
	}	

	while (1) {
		if (encode) {
			ret = ARMAMBE_ReceiveBuffer((char *)chan_buf, echanbits*sizeof(short));
			if (ret != echanbits*sizeof(short)) {
				ERROR("ret = %d, echanbits*sizeof(short) = %d\n", ret, echanbits*sizeof(short));
				if (input_finished)
					break;

				continue;
			}

			if (host_connected) {
				// Write control header to Host
				ctrl = HOST_STREAM_DATA;
        			if (write_data_payload(connfd, (char *)&ctrl, sizeof(ctrl)) < 0) {
                			ERROR("ERROR writing control packet to host\n");
					host_connected = 0;
				}

				// Write data payload to Host
				if (write_data_payload(connfd, (char *)chan_buf, ret) < 0) {
                			ERROR("ERROR writing data payload to host\n");
					host_connected = 0;
				}
			}
			else
				INFO("host_connected = 0, it shouldn't be!\n");

			rcv_bytes += ret;  

			//enc_cvt_sdbits(chan_buf, echanbits, ao->sd_bits);

		} else if (decode) {
			ret = ARMAMBE_ReceiveBuffer((char *)speech_buf, dec_ws*sizeof(short));
			if (ret != dec_ws*sizeof(short)) {
				ERROR("ret = %d, dec_ws*sizeof(short) = %d\n", ret, dec_ws*sizeof(short));
				if (input_finished)
					break;
				
				continue;
			}

			if (host_connected) {
				// Write control header to Host
				ctrl = HOST_STREAM_DATA;
        			if (write_data_payload(connfd, (char *)&ctrl, sizeof(ctrl)) < 0) {
                			ERROR("ERROR writing control packet to host\n");
					host_connected = 0;
				}

				// Write data payload to Host
				if (write_data_payload(connfd, (char *)speech_buf, ret) < 0) {
                			ERROR("ERROR writing data payload to host\n");
					host_connected = 0;
				}
			}
			else
				INFO("host_connected = 0, it shouldn't be!\n");

			rcv_bytes += ret;
		}
		else
			break;

		//INFO("RECEIVED BYTES = %d\n",rcv_bytes);
		

		if (input_finished) {
			if (exprcv_bytes == rcv_bytes) {				
				INFO("exprcv_bytes = %d, rcv_bytes = %d\n", exprcv_bytes, rcv_bytes);
				break;
			}
		}
	}

	// Sends the host a HOST_STOP_STREAM message once the input thread is finished, and the output thread
	// 		has received all of the data from the DSP that was sent in the input thread.
	if (host_connected) {
		ctrl = HOST_STOP_STREAM;
		if (write_data_payload(connfd, (char *)&ctrl, sizeof(ctrl)) < 0) {
			printf("ERROR writing control packet to host\n");
			host_connected = 0;
		}
		INFO("HOST_STOP_STREAM\n");
	}

	output_finished = 1;
	return (0);
}

/**
 * @details The thread monitors and controls the encryption
 * 			socket connection between the ARM GPP and the host.  The socket is closed when no more data
 * 			is found on the port.
 */
/* Main host encryption device to ARM loop */
int gh_intf_crypt_loop(int port)
{
	int res, sockfd;
	char buf[200];

	gh_crypt_started = 1;
	while (gh_crypt_started) {
		// Start TCP server (for encoding), listen for connections
		sockfd = start_tcp_server(port);
		INFO("Waiting for connection on port %d [sock = %d]\n", port, sockfd);
		do {
			crypt_connfd = get_tcp_server_connection(sockfd);
			usleep(100);
		} while (!crypt_connfd);
		crypt_host_connected = 1;

		// Client connected, start main processing loop
		while (crypt_host_connected) {
			res = recv(crypt_connfd, buf, 1, MSG_PEEK | MSG_DONTWAIT);
			if (res == 0) {
				ERROR("Client disconnected.\n");
				crypt_host_connected = 0;
				break;
			}
			usleep(100);
		}		

		INFO("Stopping input TCP server...\n");
		
		// Close connection and TCP server
		close_tcp_server_connection(crypt_connfd);
		stop_tcp_server(sockfd);
	}

	return 0;
}

/**
 * @details Thread handles both the reading a writing of data
 * 			to the host for encryption.
 *
 *
 */
int gh_intf_cryptography(PC_OPTIONS * po) 
{
	static short chan_buf[CHAN_BUF_SZ];
	static short recv_frame_count = 0;
	static short sent_frame_count = 0;
	short echanbits;
	int mode, res, payload_size;

	echanbits = (po->mode == APCO_MODE) ? APCO_SRCBITS : APCO_HR_SRCBITS_MAX;
	payload_size = echanbits*sizeof(short);
	INFO("echanbits = %d shorts, payload_size = %d bytes\n", echanbits, payload_size);
	
	while (1) {
		if (crypt_host_connected) {
			do {
				mode = ARMAMBE_ReceiveCryptBuffer((char *)&chan_buf, payload_size);
				//INFO("Mode = %d\n", mode);
				recv_frame_count++;
				//INFO("************************* recv_frame_count = %d\n\n",recv_frame_count);

				switch (mode) {
					case CRYPT_ENCRYPTION:
					case CRYPT_DECRYPTION:
						// Write control header and data payload to Host
						if (write_data_payload(crypt_connfd, (char *)&mode, sizeof(mode)) < 0) {
							ERROR("ERROR writing control packet to host\n");
							crypt_host_connected = 0;
							break;
						}
						//INFO("Mode sent to host = %d\n", mode);
						if (write_data_payload(crypt_connfd, (char *)chan_buf, payload_size) < 0) {
							ERROR("ERROR writing data payload to host\n");
							crypt_host_connected = 0;
							INFO("Wrote payload to host\n");
							break;
						}
						//INFO("Payload sent to host = %d\n", recv_frame_count);

						//usleep(15000);
						break;
					case CRYPT_END:
						INFO("Finished Cryptography\n");
						recv_frame_count = 0;
                        sent_frame_count = 0;

						// Write control header to Host
						if (write_data_payload(crypt_connfd, (char *)&mode, sizeof(mode)) < 0) {
							ERROR("ERROR writing control packet to host\n");
							crypt_host_connected = 0;
						}

						return 1;
					default:
						ERROR("Unexpected Control\n");
				}
			} while (recv_frame_count == 1 || recv_frame_count == 10);

			// Read in encrypted/decrypted data payload.
			res = read_data_payload(crypt_connfd, (char *)&chan_buf, payload_size);
			if (res == -1) {
				ERROR("read_data_payload() failure!\n");
				crypt_host_connected = 0;
			}
			else if (res != echanbits*sizeof(short)) {
				INFO("res = %d, buflen = %d\n", res, payload_size);
				res = echanbits*sizeof(short);
			}

			ARMAMBE_SendCryptBuffer((char *)&chan_buf, payload_size);
			sent_frame_count++;
			//INFO("************************* sent_frame_count = %d\n\n",sent_frame_count);

			if (recv_frame_count == 9 || recv_frame_count == 18) {
				// Read in encrypted/decrypted data payload.
				res = read_data_payload(crypt_connfd, (char *)&chan_buf, payload_size);
				if (res == -1) {
					ERROR("read_data_payload() failure!\n");
					crypt_host_connected = 0;
				}
				else if (res != echanbits*sizeof(short)) {
					INFO("res = %d, buflen = %d\n", res, payload_size);
					res = echanbits*sizeof(short);
				}

				ARMAMBE_SendCryptBuffer((char *)&chan_buf, payload_size);
				sent_frame_count++;
				//INFO("************************* sent_frame_count = %d\n\n",sent_frame_count);
			}

			if (recv_frame_count == 18) {
				recv_frame_count = 0;
				sent_frame_count = 0;
			}
		}
		else
			sched_yield();
	}
	return (0);
}

