/**
 * ============================================================================
 * @file	tskDspAmbeIn.c
 * @brief	TSK based library that uses SIO interface to implement input
 * 			channels from the ARM. Data received over the channels is pushed into
 * 			a FIFO and is accessible to processes that require the data.
 * @note	none.
 *
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */


/*  ----------------------------------- DSP/BIOS Headers            */
#include <std.h>
#include <log.h>
#include <sys.h>
#include <sio.h>
#include <tsk.h>
#include <sem.h>

/*  ----------------------------------- DSP/BIOS LINK Headers       */
#include <failure.h>
#include <dsplink.h>
#include <platform.h>

/*  ----------------------------------- Sample Headers              */
#include "common.h"
#include "dsp_ambe_config.h"
#include "tskDspAmbeIn.h"
#include "byte_fifo.h"
#include "bs_packet.h"
#include "dprintf.h"
#include "log_dsp.h"


#define swap32(x)                           \
     ((((x) & 0x000000ff) << 24) | (((x) & 0x0000ff00) << 8) |   \
     (((x) & 0x00ff0000) >>  8) | (((x) & 0xff000000) >>24))


/** ============================================================================
 *  @const  FILEID
 *  @desc   FILEID is used by SET_FAILURE_REASON macro.
 *  ============================================================================
 */
#define FILEID  FID_APP_C

/** ============================================================================
 *  @name   xferBufSize
 *  @desc   Size of the buffer for TSK based dsp_ambe.
 *  ============================================================================
 */
extern Uint32 xferBufSize ;
extern short inputTskRunning;

static struct byte_fifo *bf_in;

static SEM_Handle sem_bf_in_mtx;
static SEM_Handle sem_bf_in_empty;
static SEM_Handle sem_bf_in_notempty;
static SEM_Handle sem_bf_in_stateswitch;

static char arm_mode;
static char bf_get_accept_less;

static TSKDSPAMBE_IN_TransferInfo *  info = NULL;

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_create
 *  @desc   Create phase function for the TSKDSPAMBE_IN application. Initializes the
 *          TSKDSPAMBE_IN_TransferInfo structure with the information that will be
 *          used by the other phases of the application.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_create (TSKDSPAMBE_IN_TransferInfo ** infoPtr)
{
    Int                     status = SYS_OK ;
    SIO_Attrs               attrs ;
    Uint16                  i ;
    Uint16                  j ;

    /* Allocate TSKDSPAMBE_IN_TransferInfo structure that will be initialized
     * and passed to other phases of the application
     */
    *infoPtr = MEM_calloc (DSPLINK_SEGID, sizeof (TSKDSPAMBE_IN_TransferInfo), DSPLINK_BUF_ALIGN) ;
    if (*infoPtr == NULL) {
        status = SYS_EALLOC ;
    }
    else {
        info = *infoPtr ;
    }

    if (status == SYS_OK) {
        /* Filling up the transfer info structure */
        info->bufferSize   =  xferBufSize ;
        info->numBuffers   =  TSK_NUM_BUFFERS ;
        info->numChnlBuffers   =  1 ;

        /* Attributes for the stream creation */
        attrs = SIO_ATTRS ;
        attrs.nbufs   = info->numChnlBuffers ;
        attrs.segid   = DSPLINK_SEGID ;
        attrs.align   = DSPLINK_BUF_ALIGN ;
        attrs.flush   = TRUE ;
        attrs.model   = SIO_ISSUERECLAIM ;
        attrs.timeout = SYS_FOREVER ;

        /* Creating input stream */
		DBG2("SIO_Create (%s) numBuffers = %d\n",INPUT_CHANNEL, info->numChnlBuffers);
        info->inStream = SIO_create (INPUT_CHANNEL, SIO_INPUT, info->bufferSize, &attrs) ;
		DBG2("SIO_Create (%s) numBuffers = %d\n",CINPUT_CHANNEL, info->numChnlBuffers);
        info->inCryptStream = SIO_create (CINPUT_CHANNEL, SIO_INPUT, info->bufferSize, &attrs) ;

        if (info->inStream  == NULL || info->inCryptStream == NULL) {
            status = SYS_EALLOC ;
			ERROR1("SIO_Create failed [%x]\n",status);
        }
    }

    /* Allocating all the buffers that will be used in the transfer */
    if (status == SYS_OK) {
        for (i = 0 ; (i < info->numBuffers) && (status == SYS_OK) ; i++) {
            status = POOL_alloc (SAMPLE_POOL_ID, (Ptr *) &(info->buffers [i]), info->bufferSize) ;
			DBG2("Allocated pool [%d], ptr = %p\n",i,info->buffers[i]);	
            if (status != SYS_OK) {
                for (j = 0 ; j < i ; j++) {
                    POOL_free (SAMPLE_POOL_ID, info->buffers [j], info->bufferSize) ;
                    info->buffers [j] = NULL ;
                }
                status = SYS_EALLOC ;
				ERROR1("POOL_alloc failed [%x]\n",status);
            }
        }
    }
	
	DBG2("CHNL_SIZE = %d, BS_BUF_LEN8 = %d\n", xferBufSize, BS_BUF_LEN8);

	sem_bf_in_mtx = SEM_create(1, NULL);
	sem_bf_in_empty = SEM_create(0, NULL);
	sem_bf_in_notempty = SEM_create(0, NULL);
	sem_bf_in_stateswitch = SEM_create(0, NULL);

	bf_in = bf_alloc(BS_BUF_LEN8);

	arm_mode = ARM_NORMAL_MODE;
	bf_get_accept_less = 0;

    return status ;
}


/** ============================================================================
 *  @func   TSKDSPAMBE_IN_execute
 *  @desc   Execute phase function for the TSKDSPAMBE_IN application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_execute(TSKDSPAMBE_IN_TransferInfo * info)
{
    Int    status  = SYS_OK ;
    Char * buffer  = info->buffers [0] ;
    Arg    arg     = 0 ;
    Int    nmadus ;

    struct bs_packet *pkt;
	AMBE_OPTIONS * ao;

	int ret;
	int space;
	int size;

	inputTskRunning = 1;

    while (status == SYS_OK) {
		SEM_pend(sem_bf_in_mtx, SYS_FOREVER);
		space = bf_space(bf_in);
		SEM_post(sem_bf_in_mtx);
		while (space < BLOB_LEN) {
			DBG1("Pending on semaphore, bf_in space = %d\n", space);
			SEM_pend(sem_bf_in_empty, SYS_FOREVER);
			SEM_pend(sem_bf_in_mtx, SYS_FOREVER);
			space = bf_space(bf_in);
			SEM_post(sem_bf_in_mtx);
		}

        /* Receive a data buffer from GPP */
		DBG1("Issue empty buffer [%p] to GPP\n", buffer);
        status = SIO_issue(info->inStream, buffer, info->bufferSize, arg) ;
        if (status == SYS_OK) {
            nmadus = SIO_reclaim (info->inStream, (Ptr *) &buffer, &arg) ;
			DBG1("Reclaim filled buffer [%p] from GPP\n", buffer);
 
            if (nmadus < 0) {
                status = -nmadus ;
				ERROR1("Error, status = [0x%x]\n",status);
                SET_FAILURE_REASON (status) ;
            }
            else {
                info->receivedSize = nmadus ;
            }
        }
        else {
			ERROR1("Error, status = [0x%x]\n",status);
            SET_FAILURE_REASON(status);
        }

        /* Do processing on this buffer */
        if (status == SYS_OK) {
			pkt =  (struct bs_packet *) buffer;	

			DBG2("IN: pkt->hdr = %x, pkt->len = %d\n",pkt->hdr, pkt->len);

			if (pkt->hdr == PKT_HDR_CNTRL) { 
				DBG("IN: Control pkt->hdr match\n");

				ao = (AMBE_OPTIONS *) pkt->data;
			
				do {
					DBG1("Waiting for buffer to clear, buf size = %d\n",bf_size(bf_in));
					bf_get_accept_less = 1;
					SEM_post(sem_bf_in_notempty);
					SEM_pend(sem_bf_in_stateswitch, SYS_FOREVER);
					
					SEM_pend(sem_bf_in_mtx, SYS_FOREVER);
					size = bf_size(bf_in);
					SEM_post(sem_bf_in_mtx);
				} while (size > 0);

				// Change ARM MODE
				TSKDSPAMBE_IN_setArmMode(ao->arm_mode);

				// If this is a STOP control packet don't put it in the FIFO
				if (!ao->run_encoder && !ao->run_decoder)
					continue;
        	}

			SEM_pend(sem_bf_in_mtx, SYS_FOREVER);
			ret = bf_put_n(bf_in, pkt->data, pkt->len);
			SEM_post(sem_bf_in_mtx);
        	if (ret != pkt->len)
            	INFO("bf import failed\n");

			//INFO1("bf_in size = %d\n",bf_size(bf_in));
 
			if (SEM_count(sem_bf_in_notempty) < 2)
         		SEM_post(sem_bf_in_notempty);
		}
    }
    return status ;
}

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_getBuffer
 *  @desc   Execute phase function for the TSKDSPAMBE_IN application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_getBuffer(Char * buf, Int len)
{
	Int ret;
	int pre_space;
	int space;
	int size;

	DBG1("Requesting %d bytes\n",len);

	if (len > BS_BUF_LEN8) {
		INFO1("bs_input len too long %d\n", len);
		len = BS_BUF_LEN8;
	}

	SEM_pend(sem_bf_in_mtx, SYS_FOREVER);
	size = bf_size(bf_in);
	//INFO1("bf_in size = %d\n",size);
	SEM_post(sem_bf_in_mtx);
	while ((size < len) && !bf_get_accept_less) {
		SEM_pend(sem_bf_in_notempty, SYS_FOREVER);
		SEM_pend(sem_bf_in_mtx, SYS_FOREVER);
		size = bf_size(bf_in);
		SEM_post(sem_bf_in_mtx);
	}


	SEM_pend(sem_bf_in_mtx, SYS_FOREVER);
	pre_space = bf_space(bf_in);
	ret = bf_get_n(bf_in, buf, len);
	//space = bf_space(bf_in);
	SEM_post(sem_bf_in_mtx);
	if (ret != len)
		INFO("bf get failed\n");

	SEM_pend(sem_bf_in_mtx, SYS_FOREVER);
	size = bf_size(bf_in);
	SEM_post(sem_bf_in_mtx);
	if (bf_get_accept_less && size == 0) {
		bf_get_accept_less = 0;
		SEM_post(sem_bf_in_stateswitch);
	}

	SEM_pend(sem_bf_in_mtx, SYS_FOREVER);
	space = bf_space(bf_in);
	SEM_post(sem_bf_in_mtx);
	if (space >= BLOB_LEN && pre_space < BLOB_LEN)
		SEM_post(sem_bf_in_empty);

	DBG1("Got %d bytes\n",ret);

	return ret;
}

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_getCryptBuffer
 *  @desc   Execute phase function for the TSKDSPAMBE_IN application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_getCryptBuffer(Char * buf, Int len)
{
    Int    status  = SYS_OK ;
    //Char * buffer  = info->buffers [1] ;
    Arg    arg     = 0 ;
    Int    nmadus ;

    struct bs_packet *pkt;

	int recv = 0;

    while (status == SYS_OK && recv < len) {
        /* Receive a data buffer from GPP */
		DBG1("Issue empty buffer [%p] to GPP\n", info->buffers[1]);
        status = SIO_issue(info->inCryptStream, info->buffers[1], info->bufferSize, arg) ;
        if (status == SYS_OK) {
            nmadus = SIO_reclaim (info->inCryptStream, (Ptr *) &(info->buffers[1]), &arg) ;
			DBG1("Reclaim filled buffer [%p] from GPP\n", info->buffers[1]);
 
            if (nmadus < 0) {
                status = -nmadus ;
				ERROR1("Error, status = [0x%x]\n",status);
                SET_FAILURE_REASON (status) ;
            }
            else {
                info->receivedSize = nmadus ;
            }
        }
        else {
			ERROR1("Error, status = [0x%x]\n",status);
            SET_FAILURE_REASON(status);
        }

        /* Do processing on this buffer */
        if (status == SYS_OK) {
			pkt =  (struct bs_packet *) info->buffers[1];	

			DBG2("IN: pkt->hdr = %x, pkt->len = %d\n",pkt->hdr, pkt->len);
			if (pkt->hdr == PKT_HDR_CNTRL) { 
				DBG("IN: Control pkt->hdr match\n");
					continue;
        	}
			
			// Shorten pkt length to enough bytes to meet the requested size
			if (pkt->len > len-recv)
				pkt->len = len-recv;
	
			memcpy(buf+recv, pkt->data, pkt->len);
 			recv += pkt->len;
		}
    }

	DBG1("Got %d bytes\n",recv);
    return recv;
}

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_setArmMode
 *  @desc   Execute phase function for the TSKDSPAMBE_IN application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Void TSKDSPAMBE_IN_setArmMode(Int mode)
{
	arm_mode = mode;
	INFO1("Set ARM MODE to %d\n",(short) arm_mode);
}

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_getArmMode
 *  @desc   Execute phase function for the TSKDSPAMBE_IN application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_getArmMode()
{
	return arm_mode;
}

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_delete
 *  @desc   Delete phase function for the TSKDSPAMBE_IN application. It deallocates
 *          all the resources of allocated during create phase of the
 *          application.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_delete (TSKDSPAMBE_IN_TransferInfo * info)
{
    Int     status     = SYS_OK ;
    Bool    freeStatus = FALSE ;
    Uint16  j ;

    /* Delete input stream */
    if (info->inStream != NULL) {
        status = SIO_delete (info->inStream) ;
        if (status != SYS_OK) {
            SET_FAILURE_REASON (status) ;
			ERROR1("Error: status = [0x%x]\n",status);
        }
    }

    /* Delete the buffers */
    if (info->numBuffers > 0) {
        for (j = 0 ;
             (j < info->numBuffers) && (info->buffers [j] != NULL) ;
             j++) {
            POOL_free (SAMPLE_POOL_ID, info->buffers [j], info->bufferSize) ;
        }
    }

    /* Free the info structure */
    freeStatus = MEM_free (DSPLINK_SEGID, info, sizeof (TSKDSPAMBE_IN_TransferInfo)) ;
    if ((status == SYS_OK) && (freeStatus != TRUE)) {
        status = SYS_EFREE ;
        SET_FAILURE_REASON (status) ;
		ERROR1("Error: status = [0x%x]\n",status);
    }

    return status ;
}
