/**
 * ============================================================================
 * @file	tskDspAmbeOut.c
 * @brief	TSK based library that uses SIO interface to implement output
 * 			channels to the ARM. Data to be sent over the channel is first
 * 			store in a FIFO and then processed by a dedicated task.
 * @note	none.
 *
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */


/*  ----------------------------------- DSP/BIOS Headers            */
#include <std.h>
#include <log.h>
#include <swi.h>
#include <sys.h>
#include <sio.h>
#include <tsk.h>

/*  ----------------------------------- DSP/BIOS LINK Headers       */
#include <failure.h>
#include <dsplink.h>
#include <platform.h>

/*  ----------------------------------- Sample Headers              */
#include "common.h"
#include "dsp_ambe_config.h"
#include "tskDspAmbeOut.h"
#include "byte_fifo.h"
#include "bs_packet.h"
#include "dprintf.h"
#include "log_dsp.h"

/** ============================================================================
 *  @const  FILEID
 *  @desc   FILEID is used by SET_FAILURE_REASON macro.
 *  ============================================================================
 */
#define FILEID  FID_APP_C

/** ============================================================================
 *  @name   xferBufSize
 *  @desc   Size of the buffer size for TSK based dsp_ambeback.
 *  ============================================================================
 */
extern Uint32 xferBufSize ;
extern short outputTskRunning;

static struct byte_fifo *bf_out;

static SEM_Handle sem_bf_out_mtx;
static SEM_Handle sem_bf_out_flush;
static SEM_Handle sem_bf_out_empty;
static SEM_Handle sem_bf_out_notempty;

static char flush_bf_out;
static char cntrl_bf_out;

static TSKDSPAMBE_OUT_TransferInfo *  info = NULL;

/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_create
 *  @desc   Create phase function for the TSKDSPAMBE_OUT application. Initializes the
 *          TSKDSPAMBE_OUT_TransferInfo structure with the information that will be
 *          used by the other phases of the application.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_create (TSKDSPAMBE_OUT_TransferInfo ** infoPtr)
{
	Int                     status = SYS_OK ;
	SIO_Attrs               attrs ;
	Uint16                  i ;
	Uint16                  j ;

	/* Allocate TSKDSPAMBE_OUT_TransferInfo structure that will be initialized
	 * and passed to other phases of the application
	 */
	*infoPtr = MEM_calloc (DSPLINK_SEGID, sizeof (TSKDSPAMBE_OUT_TransferInfo), DSPLINK_BUF_ALIGN) ;
	if (*infoPtr == NULL) {
		status = SYS_EALLOC ;
	}
	else {
		info = *infoPtr ;
	}

	if (status == SYS_OK) {
		/* Filling up the transfer info structure */
		info->bufferSize   =  xferBufSize ;
		info->numBuffers   =  TSK_NUM_BUFFERS ;
		info->numChnlBuffers   =  1 ;

		/* Attributes for the stream creation */
		attrs = SIO_ATTRS ;
		attrs.nbufs   = info->numChnlBuffers ;
		attrs.segid   = DSPLINK_SEGID ;
		attrs.align   = DSPLINK_BUF_ALIGN ;
		attrs.flush   = TRUE ;
		attrs.model   = SIO_ISSUERECLAIM ;
		attrs.timeout = SYS_FOREVER ;

		/* Creating input and output streams */
		DBG2("SIO_Create (%s) numBuffers = %d\n",OUTPUT_CHANNEL, info->numChnlBuffers);
		info->outStream = SIO_create(OUTPUT_CHANNEL, SIO_OUTPUT, info->bufferSize, &attrs);
		DBG2("SIO_Create (%s) numBuffers = %d\n",COUTPUT_CHANNEL, info->numChnlBuffers);
		info->outCryptStream = SIO_create(COUTPUT_CHANNEL, SIO_OUTPUT, info->bufferSize, &attrs);

		if (info->outStream  == NULL && info->outCryptStream == NULL) {
			status = SYS_EALLOC ;
			ERROR1("SIO_Create failed [%x]\n",status);
		}
	}

	/* Allocating all the buffers that will be used in the transfer */
	if (status == SYS_OK) {
		for (i = 0 ; (i < info->numBuffers) && (status == SYS_OK) ; i++) {
			status = POOL_alloc (SAMPLE_POOL_ID, (Ptr *) &(info->buffers [i]), info->bufferSize) ;
			DBG2("Allocated pool [%d], ptr = %p\n",i,info->buffers[i]);	
			if (status != SYS_OK) {
				for (j = 0 ; j < i ; j++) {
					POOL_free (SAMPLE_POOL_ID, info->buffers [j], info->bufferSize) ;
					info->buffers [j] = NULL ;
				}
				status = SYS_EALLOC ;
				ERROR1("POOL_alloc failed [%x]\n",status);
			}
		}
	}

	bf_out = bf_alloc(BS_BUF_LEN8);

	sem_bf_out_mtx = SEM_create(1, NULL);
	sem_bf_out_flush = SEM_create(0, NULL);
	sem_bf_out_empty = SEM_create(0, NULL);
	sem_bf_out_notempty = SEM_create(0, NULL);

	flush_bf_out = 0;
	cntrl_bf_out = 0;

	return status ;
}


/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_execute
 *  @desc   Execute phase function for the TSKDSPAMBE_OUT application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_execute(TSKDSPAMBE_OUT_TransferInfo * info)
{
	Int         status  = SYS_OK ;
	Char *      buffer  = info->buffers [0] ;
	Arg         arg     = 0 ;
	Int         nmadus ;

	struct bs_packet * pkt;

	int ret;
	int size;

	outputTskRunning = 1;

	/* Execute the dsp_ambe for configured number of transfers
	 * A value of 0 in numTransfers implies infinite iterations
	 */
	while (status == SYS_OK) {
		pkt =  (struct bs_packet *) buffer;	

		DBG("Is bf_out empty?\n");
		SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
		size = bf_size(bf_out);
		SEM_post(sem_bf_out_mtx);
		while (size < BLOB_LEN && !flush_bf_out) {
			SEM_pend(sem_bf_out_notempty, SYS_FOREVER);
			SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
			size = bf_size(bf_out);
			SEM_post(sem_bf_out_mtx);
		}

		DBG("no!\n");
		SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
		ret = bf_get_n(bf_out, pkt->data, BLOB_LEN);
		SEM_post(sem_bf_out_mtx);

		pkt->len = ret;
		pkt->hdr = PKT_HDR_DATA;
		if (cntrl_bf_out)
			pkt->hdr = PKT_HDR_CNTRL;

		DBG2("OUT: pkt->hdr = %x, pkt->len = %d\n",pkt->hdr, pkt->len);
		DBG1("OUT: bf_out size = %d\n",bf_size(bf_out));

		/* Send the processed buffer back to GPP */
		if (status == SYS_OK) {
			DBG1("Issue filled buffer [%p] to GPP\n", buffer);
			status = SIO_issue(info->outStream, buffer, info->bufferSize, arg);

			if (status == SYS_OK) {
				nmadus = SIO_reclaim (info->outStream, (Ptr *) &(buffer), &arg) ;
				DBG1("Reclaim empty buffer [%p] to GPP\n", buffer);
				if (nmadus < 0) {
					status = -nmadus ;
					SET_FAILURE_REASON (status) ;
					ERROR1("Error, status = [0x%x]\n",status);
				}
			} else {
				SET_FAILURE_REASON (status) ;
				ERROR1("Error, status = [0x%x]\n",status);
			}
		}

		cntrl_bf_out = 0;

		SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
		size = bf_size(bf_out);
		SEM_post(sem_bf_out_mtx);
		if (flush_bf_out && size == 0) {
			flush_bf_out = 0;
			SEM_post(sem_bf_out_flush);
		}

		SEM_post(sem_bf_out_empty);
	}
	return status ;
}


/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_sendBuffer
 *  @desc   Execute phase function for the TSKDSPAMBE_OUT application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_sendBuffer(Char * buf, Int len, Char control)
{
	Int ret;
	int pre_size;
	int space;
	int size;

	DBG1("Sending %d bytes\n",len);

	if (len > BS_BUF_LEN8) {
		INFO1("bs_output len too long %d\n", len);
		len = BS_BUF_LEN8;
	}

	if (control) {
		DBG("Preparing to send Control bytes\n");
		TSKDSPAMBE_OUT_flushSendBuffer();
	}

	SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
	space = bf_space(bf_out);
	SEM_post(sem_bf_out_mtx);
	while (space < len) {
		SEM_pend(sem_bf_out_empty, SYS_FOREVER);
		SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
		space = bf_space(bf_out);
		SEM_post(sem_bf_out_mtx);
	}

	SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
	pre_size = bf_size(bf_out);
	ret = bf_put_n(bf_out, buf, len);
	SEM_post(sem_bf_out_mtx);
	if (ret != len)
		INFO("bf put failed\n");

	// Set cntrl_bf_out only after control buffer has been put in fifo
	if (control) {
		cntrl_bf_out = 1;
		DBG("Forcing Control send\n");
		TSKDSPAMBE_OUT_flushSendBuffer();
	}

	SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
	size = bf_size(bf_out);
	SEM_post(sem_bf_out_mtx);
	if (size >= BLOB_LEN && pre_size < BLOB_LEN)
		SEM_post(sem_bf_out_notempty);

	DBG1("Sent %d bytes\n",ret);

	return ret;
}

/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_sendCryptBuffer
 *  @desc   Execute phase function for the TSKDSPAMBE_OUT application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_sendCryptBuffer(Char * buf, Int len, Short mode)
{
    Int    status  = SYS_OK ;
    //Char * buffer  = info->buffers[1] ;
    Arg    arg     = 0 ;
    Int    nmadus ;

    struct bs_packet *pkt;

	int sent = 0;

    while (status == SYS_OK && sent < len) {
		pkt =  (struct bs_packet *) info->buffers[1];	

		pkt->hdr = mode;

		// Shorten pkt length to fit in channel buffer
		pkt->len = len-sent;
		if (pkt->len > BLOB_LEN)
			pkt->len = BLOB_LEN;
	
		memcpy(pkt->data, buf+sent, pkt->len);
 		sent += pkt->len;

		DBG2("OUT: pkt->hdr = %x, pkt->len = %d\n",pkt->hdr, pkt->len);


		/* Send the processed buffer back to GPP */
		if (status == SYS_OK) {
			DBG1("Issue filled buffer [%p] to GPP\n", info->buffers[1]);
			status = SIO_issue(info->outCryptStream, info->buffers[1], info->bufferSize, arg);

			if (status == SYS_OK) {
				nmadus = SIO_reclaim (info->outCryptStream, (Ptr *) &(info->buffers[1]), &arg) ;
				DBG1("Reclaim empty buffer [%p] to GPP\n", info->buffers[1]);
				if (nmadus < 0) {
					status = -nmadus ;
					SET_FAILURE_REASON (status) ;
					ERROR1("Error, status = [0x%x]\n",status);
				}
			} else {
				SET_FAILURE_REASON (status) ;
				ERROR1("Error, status = [0x%x]\n",status);
			}
		}
    }

	DBG1("Sent %d bytes\n",sent);
    return sent;
}

/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_flushSendBuffer
 *  @desc   Execute phase function for the TSKDSPAMBE_OUT application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Void TSKDSPAMBE_OUT_flushSendBuffer(Void)
{
	int size;

	DBG("flushing output buffer\n");

	SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
	size = bf_size(bf_out);
	SEM_post(sem_bf_out_mtx);
	while (size > 0) {
		DBG("Flushing DSP\n");
		flush_bf_out = 1;
		SEM_post(sem_bf_out_notempty);
		SEM_pend(sem_bf_out_flush, SYS_FOREVER);

		SEM_pend(sem_bf_out_mtx, SYS_FOREVER);
		size = bf_size(bf_out);
		SEM_post(sem_bf_out_mtx);
	}
}

/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_delete
 *  @desc   Delete phase function for the TSKDSPAMBE_OUT application. It deallocates
 *          all the resources of allocated during create phase of the
 *          application.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_delete (TSKDSPAMBE_OUT_TransferInfo * info)
{
	Int     status     = SYS_OK ;
	Uint16  tmpStatus = SYS_OK ;
	Bool    freeStatus = FALSE ;
	Uint16  j ;

	/* Delete output stream */
	if (info->outStream != NULL) {
		tmpStatus = SIO_delete(info->outStream);
		if ((status == SYS_OK) && (tmpStatus != SYS_OK)) {
			status = tmpStatus ;
			SET_FAILURE_REASON (status) ;
			ERROR1("Error: status = [0x%x]\n",status);
		}
	}

	/* Delete the buffers */
	if (info->numBuffers > 0) {
		for (j = 0 ;
				(j < info->numBuffers) && (info->buffers [j] != NULL) ;
				j++) {
			POOL_free (SAMPLE_POOL_ID, info->buffers [j], info->bufferSize) ;
		}
	}

	/* Free the info structure */
	freeStatus = MEM_free (DSPLINK_SEGID, info, sizeof (TSKDSPAMBE_OUT_TransferInfo)) ;
	if ((status == SYS_OK) && (freeStatus != TRUE)) {
		status = SYS_EFREE ;
		SET_FAILURE_REASON (status) ;
		ERROR1("Error: status = [0x%x]\n",status);
	}

	return status ;
}
