/**
 * ============================================================================
 * @file	tskDspAmbeOut.h
 * @brief	This file contains function definitions and parameters for the
 * 			tskDspAmbeOut.c
 * @note	none.
 *
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */

#if !defined (TSKDSPAMBE_OUT_)
#define TSKDSPAMBE_OUT_

#include <sio.h>

#if defined (__cplusplus)
extern "C" {
#endif /* defined (__cplusplus) */


/** ============================================================================
 *  @const  MAX_BUFFERS
 *  @desc   Size of buffers array.
 *  ============================================================================
 */
#define MAX_BUFFERS            100

/** ============================================================================
 *  @const  TSK_NUM_BUFFERS
 *  @desc   Number of buffers to allocate and use for transfer.
 *  ============================================================================
 */
#define TSK_NUM_BUFFERS          2


/** ============================================================================
 *  @name   TSKDSPAMBE_OUT_TransferInfo
 *  @desc   Structure used to keep various information needed by various phases
 *          of the application.
 *  @field  inputStream
 *              Handle to input stream.
 *  @field  outputStream
 *              Handle to output stream.
 *  @field  numTransfers
 *              Number of transfers to be done.
 *  @field  bufferSize
 *              Buffer size information.
 *  @field  receivedSize
 *              Actual length of the buffer that was received.
 *  @field  numBuffers
 *              Number of valid buffers in "buffers" array.
 *  @field  buffers
 *              Array of buffer pointers.
 *  ============================================================================
 */
typedef struct TSKDSPAMBE_OUT_TransferInfo_tag {
    SIO_Handle outStream;
    SIO_Handle outCryptStream;
    Uint32     bufferSize;
    Uint16     sendSize;
    Uint16     numBuffers;
    Uint16     numChnlBuffers;
    Char *     buffers [MAX_BUFFERS];
} TSKDSPAMBE_OUT_TransferInfo;


/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_create
 *  @desc   Create phase function of TSKDSPAMBE_OUT application.
 *  @arg    transferInfo
 *              Information for transfer.
 *  @ret    SYS_OK
 *              Successful operation.
 *          SYS_EBADIO
 *              Failure occured while doing IO.
 *  @enter  None
 *  @leave  None
 *  @see    None
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_create (TSKDSPAMBE_OUT_TransferInfo ** transferInfo);

/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_execute
 *  @desc   Excecute phase function of TSKDSPAMBE_OUT application.
 *  @arg    transferInfo
 *              Information for transfer.
 *  @ret    SYS_OK
 *              Successful operation.
 *          SYS_EBADIO
 *              Failure occured while doing IO.
 *  @enter  None
 *  @leave  None
 *  @see    None
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_execute (TSKDSPAMBE_OUT_TransferInfo * transferInfo);

/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_sendBuffer
 *  @desc   Excecute phase function of TSKDSPAMBE_OUT application.
 *  @arg    transferInfo
 *              Information for transfer.
 *  @ret    SYS_OK
 *              Successful operation.
 *          SYS_EBADIO
 *              Failure occured while doing IO.
 *  @enter  None
 *  @leave  None
 *  @see    None
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_sendBuffer (Char * buf, Int len, Char control);

/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_sendCryptBuffer
 *  @desc   Excecute phase function of TSKDSPAMBE_OUT application.
 *  @arg    transferInfo
 *              Information for transfer.
 *  @ret    SYS_OK
 *              Successful operation.
 *          SYS_EBADIO
 *              Failure occured while doing IO.
 *  @enter  None
 *  @leave  None
 *  @see    None
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_sendCryptBuffer (Char * buf, Int len, Short mode);

/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_flushSendBuffer
 *  @desc   Excecute phase function of TSKDSPAMBE_OUT application.
 *  @arg    transferInfo
 *              Information for transfer.
 *  @ret    SYS_OK
 *              Successful operation.
 *          SYS_EBADIO
 *              Failure occured while doing IO.
 *  @enter  None
 *  @leave  None
 *  @see    None
 *  ============================================================================
 */
Void TSKDSPAMBE_OUT_flushSendBuffer (Void);

/** ============================================================================
 *  @func   TSKDSPAMBE_OUT_delete
 *  @desc   Delete phase function of TSKDSPAMBE_OUT application.
 *  @arg    transferInfo
 *              Information for transfer.
 *  @ret    SYS_OK
 *              Successful operation.
 *          SYS_EBADIO
 *              Failure occured while doing IO.
 *  @enter  None
 *  @leave  None
 *  @see    None
 *  ============================================================================
 */
Int TSKDSPAMBE_OUT_delete (TSKDSPAMBE_OUT_TransferInfo * transferInfo);


#if defined (__cplusplus)
}
#endif /* defined (__cplusplus) */


#endif /* !defined (TSKDSPAMBE_OUT_) */
