/* A simple loopback program for the host side of a server-client 
	tcp socket setup

	Main program for decoding:
		reads from a file
		sends to tcp socket
	   In another thread
		reads from tcp socket
		plays audio via alsa

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <alsa/asoundlib.h>
#include "alsa_play.h"
#include "tcplib.h"
#include "logger.h"
#include "ticktock.h"

#define DBG_DECODE_TESTFILE
//#define DBG_CLOCK_SENDER
#define DBG_CLOCK_LISTENER

// Length of IMBE-encoded buffer in bytes
#define E_BUFF_LEN	288

// Length of PCM buffer in bytes
#define BUFF_LEN  	320

// Length used for PCM sampling/playback - each sample is 16 bits long
#define S_BUFF_LEN	BUFF_LEN / 2

// PCM audio sampling rate
#define SAMPRATE	8000

enum imbe_ctrl_t {
        I_SOCK_ERROR,
        IMBE_DATA,
        IMBE_ENCODE_START,
        IMBE_ENCODE_STOP,
        IMBE_DECODE_START,
	IMBE_DECODE_STOP,
	IMBE_UNKNOWN
};

enum host_ctrl_t {
        H_SOCK_ERROR,
        HOST_STREAM_DATA,
        HOST_START_STREAM,
	HOST_STOP_STREAM,
	HOST_UNKNOWN
};

// Host->GPP (forward path) interface packet format for Decode mode (E_BUFF_LEN)
struct host_gpp_pkt {
        enum imbe_ctrl_t 	ctrl;
        char 			payload[E_BUFF_LEN];
};

// GPP->Host (return path) interface packet format for Decode mode (BUFF_LEN)
struct gpp_host_pkt {
        enum host_ctrl_t 	ctrl;
	char			payload[BUFF_LEN];	
};

struct tcp_server_t {
	char ip[16];
	int fport;
	int rport;
};

int started = 0;
int end_stream = 0;
void loop_shutdown(int signum)
{
	INFO("Gracefully shutting down...\n");
	started = 0;

	sleep(3);
	exit(signum);
}


void print_usage(char *execname)
{
        printf("usage:  %s <ip> <fport> <rport>\n", execname);
        printf("        <ip>       GPP server IP address\n");
        printf("        <port>     GPP server port [1025-65534]\n\n");
}

/* Reads control packet from GPP */
enum host_ctrl_t read_ctrl_pkt(int connfd)
{
        int res;
        enum host_ctrl_t ctrl;

        res = read_tcp_server_stream(connfd, (char *)&ctrl, sizeof(ctrl));
        if (res < 0) {
                ERROR("ERROR, reading from host ");
                return H_SOCK_ERROR;
        }
        else if (res == 0) {
                INFO("GPP disconnected.\n");
                return H_SOCK_ERROR;
        }
        else {
                if (ctrl == H_SOCK_ERROR) {
                        INFO("Got H_SOCK_ERROR that should be HOST_UNKNOWN\n");
                        ctrl = HOST_UNKNOWN;
                }
        }

        return ctrl;
}

/* Reads data payload from GPP */
int read_data_payload(int connfd, char *buf, int len)
{
        int res;

        res = read_tcp_server_stream(connfd, buf, len);
        if (res < 0) {
                ERROR("ERROR, reading from host ");
                return -1;
        }
        else if (res == 0) {
                INFO("GPP disconnected.\n");
                return -1;
        }

        return res;
}

/* Writes data payload to GPP */
int write_data_payload(int connfd, char *buf, int len)
{
        if (len == 0) return 0;
        int res;

        res = send_tcp_client_data(connfd, buf, len);
        if (res < 0) {
                ERROR("ERROR, writing to host ");
                return -1;
        }
        else if (res == 0) {
                INFO("GPP disconnected.\n");
                return -1;
        }

        return res;
}

/* Main listener thread for decoded PCM data stream sent from GPP */
void *listener_thread(void *arg)
{
	int res;
	int connfd = *((int *)arg);
	struct gpp_host_pkt pkt;

	char tmpstr[40];
	unsigned long count;

#ifdef DBG_DECODE_TESTFILE
	FILE *testfile;
	if ((testfile = fopen("audio_dec_pcm.dat","w")) == NULL)
		INFO("Can't open file to read decoded IMBE data\n");
#endif

	while (started)
	{
                // Read in control packet to determine state
                pkt.ctrl = read_ctrl_pkt(connfd);
                if (pkt.ctrl == H_SOCK_ERROR) {
                        ERROR("read_ctrl_pkt() failure!\n");
                        break;
                }

		// If the control packet received is not HOST_START_STREAM, go to the next
		// while loop iteration (back to reading next control packet above)
		if (pkt.ctrl != HOST_START_STREAM)
			continue;

		count = 0;
		end_stream = 0;
		INFO("HOST_START_STREAM\n");

                // Start processing data until stop encode/decode control signal received.
                do {
		#ifdef DBG_CLOCK_LISTENER
			TICK();
		#endif

                        // Read in control packet to determine state
                        pkt.ctrl = read_ctrl_pkt(connfd);
                        if (pkt.ctrl == H_SOCK_ERROR) {
                                ERROR("read_ctrl_pkt() failure!\n");
				break;
                        }

                        // If control state is HOST_STREAM_DATA, retrieve data payload
                        else if (pkt.ctrl == HOST_STREAM_DATA) {
                                //INFO("HOST_STREAM_DATA\n");

                                // Read in decoded PCM data
                                res = read_data_payload(connfd, pkt.payload, BUFF_LEN);
                                if (res == -1) {
                                        ERROR("read_data_payload() failure!\n");
                                        break;
                                }
				else if (res != BUFF_LEN) {
					ERROR("res = %d, BUFF_LEN = %d\n", res, BUFF_LEN);
					continue;
				}

				//////  Play to audio hardware  //////////////
				while (pcm_play(pkt.payload, S_BUFF_LEN) < 0) {
					pcm_play_prepare();
				}

			#ifdef DBG_DECODE_TESTFILE
				if ((res = fwrite(pkt.payload, sizeof(char), BUFF_LEN, testfile)) < 0) {
					ERROR("Error writing PCM data to file\n");
				}
			#endif

                        } 			

		#ifdef DBG_CLOCK_LISTENER
			sprintf(tmpstr, "count = %8lu:  playback loop ", count++);
			_TOCK(tmpstr, 20);
		#endif
                } while (pkt.ctrl != HOST_STOP_STREAM);

		INFO("HOST_STOP_STREAM\n");
		end_stream = 1;
	}

	//fclose(testfile);

	return NULL;
}


int main(int argc, char **argv)
{
	int res, connfd, count, errorstate;
	struct host_gpp_pkt pkt;
	struct tcp_server_t ts;
    	pthread_attr_t attr;
	pthread_t lThread;
//	int sockfdDec, connfdDec;

	char *snd_play_device = "plughw:0,0";
	
	FILE *testfile;
	if ((testfile = fopen("audio_enc.dat","r")) == NULL)
		printf("Can't open file to read encoded IMBE data\n");

	// Print out usage message if command-line argument length not sufficient.
        if (argc < 3) {
                print_usage(argv[0]);
                exit(0);
        }

	// Command-line processing error state variable
	errorstate = 0;

	// Obtain IP address from command-line parameter
	if (strlen(argv[1]) <= 15) {
		strcpy(ts.ip, argv[1]);
	}
	else {
		ERROR("Invalid IP address string!\n");
		errorstate++;
	}

	// Obtain server port from command-line parameter
	res = sscanf(argv[2], "%d", &ts.fport);
	if (ts.fport < 1025 || ts.fport > 65534) {
		ERROR("Not a valid port:  %s\n", argv[3]);
		errorstate++;
	}

	// Exit program if we encountered errors in the command-line processing
	if (errorstate > 0)
		exit(0);
	
	// Setup termination signal interrupt handler
        signal((int) SIGINT, loop_shutdown);
	
	// Open up PCM playback channel on sound device
	printf("Init PCM playback on sound device:  %s...", snd_play_device);
	pcm_play_init(snd_play_device, SAMPRATE);
	printf("done.\n");

	// Start TCP connection to GPP server
	printf("Init TCP connection to %s:%d...", ts.ip, ts.fport);	
	connfd = start_tcp_client(ts.ip, ts.fport);
	printf("done.\n");

/*	// Start TCP connnection to encode client
	sockfdDec = start_tcp_server(51819);
	INFO("Waiting for connection from encode client\n");
		do {
			connfdDec = get_tcp_server_connection(sockfdDec);
			usleep(100);
		} while (!connfdDec);
*/
	// Start TCP listener thread
	started = 1;
	end_stream = 0;
    	pthread_attr_init(&attr);
        pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
        pthread_attr_setstacksize(&attr, 131072);
    	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    	res = pthread_create(&lThread, &attr, listener_thread, (void *)&connfd);
    	pthread_attr_destroy(&attr);

    	if (res) {
		ERROR("Failed to create TCP listener thread");
		started = 0;
		exit(0);
	}

	pkt.ctrl = IMBE_DECODE_START;
	if (send_tcp_client_data(connfd, (char *)&pkt.ctrl, sizeof(pkt.ctrl)) < 0)
		ERROR("ERROR writing control start to gpp\n");
	
	count = 0;
	do {
	#ifdef DBG_CLOCK_SENDER
		TICK();		
	#endif

		////// Read from a file for testing /////////////
		if ((res = fread(pkt.payload, sizeof(char), E_BUFF_LEN, testfile)) < 0) {
			ERROR("Error reading PCM data from file\n");
			started = 0;
			break;
		}
		else if (res == 0) {
			INFO("EOF reached, file read complete.\n");
			started = 0;
			break;
		}
		usleep(16000);

/*		// Read from TCP client
		res = read_tcp_server_stream(connfdDec, pkt.payload, E_BUFF_LEN);
		
		if (res < 0) {
			ERROR("ERROR, reading from encode side ");
                	close_tcp_server_connection(connfdDec);
                	started = 0;
                	return I_SOCK_ERROR;
        	}
        
		else if (res == 0) {
                	INFO("Client disconnected.\n");
                	started = 0;
                	return I_SOCK_ERROR;
       		}
*/
				
		/////  Send to tcp socket ///////////
		pkt.ctrl = IMBE_DATA;
		if (send_tcp_client_data(connfd, (char *)&pkt, sizeof(struct host_gpp_pkt)) < 0) {    
			ERROR("ERROR writing payload to gpp\n");
			started = 0;
			break;
		}

	#ifdef DBG_CLOCK_SENDER
		TOCK("GPP sender loop");
	#endif
		//INFO("count = %d\n", count++);
	} while (started);
	
	pkt.ctrl = IMBE_DECODE_STOP;
	if (send_tcp_client_data(connfd, (char *)&pkt.ctrl, sizeof(pkt.ctrl)) < 0)
		ERROR("ERROR writing control stop to gpp\n");
	
	if ((fclose(testfile)) < 0)
		ERROR("Error closing PCM file\n");
//	stop_tcp_server(connfdDec);
//	close_tcp_server_connection(sockfdDec);

        INFO("Waiting to receive decoded samples from GPP...\n");
	while (!end_stream)
		usleep(100);

	// Close sound card handles
	pcm_play_close();

	// Close TCP client connection to GPP
	stop_tcp_client(connfd);

        // Wait for TCP listen thread to finish before exiting
        //pthread_join(lThread, NULL);

	return 0;
}





