/* A simple loopback program for the host side of a server-client 
	tcp socket setup

	Main program for encoding:
		captures audio from alsa
		sends to tcp socket
	   In another thread, 
		reads from tcp socket
		writes to encdump.dat file

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <alsa/asoundlib.h>
#include "alsa_capture.h"
#include "alsa_play.h"
#include "tcplib.h"
#include "logger.h"
#include "ticktock.h"

// Length of AMBE-encoded buffer in bytes
#define E_BUFF_LEN      288

// Length of buffer in bytes
#define BUFF_LEN  	320

// Length used for PCM sampling/playback - each sample is 16 bits long
#define S_BUFF_LEN	BUFF_LEN / 2

// PCM audio sampling rate
#define SAMPRATE	8000

enum imbe_ctrl_t {
        SOCK_ERROR,
        IMBE_DATA,
        IMBE_ENCODE_START,
        IMBE_ENCODE_STOP,
        IMBE_DECODE_START,
	IMBE_DECODE_STOP,
	IMBE_UNKNOWN
};

enum host_ctrl_t {
	H_SOCK_ERROR,
	HOST_STREAM_DATA,
	HOST_START_STREAM,
	HOST_STOP_STREAM,
	HOST_UNKNOWN
};

// Host -> GPP (forward path) interface packet format for Encode mode (BUFF_LEN)
struct host_gpp_pkt {
        enum imbe_ctrl_t 	ctrl;
        char 			payload[BUFF_LEN];
};

// GPP -> Host (return path) interface packet format for Encode mode (E_BUFF_LEN)
struct gpp_host_pkt {
	enum host_ctrl_t	ctrl;
	char			payload[E_BUFF_LEN];	
};

struct tcp_server_t {
	char ip[16];
	int fport;
	int rport;

};

int started = 0;
int end_stream = 0;
void loop_shutdown(int signum)
{
	INFO("Gracefully shutting down...\n");
	started = 0;

	// Close TCP client connection and shutdown PCM capture and playback
//	stop_tcp_client(connfd);
	started = 0;

	sleep(10);
	exit(signum);
}

void print_usage(char *execname)
{
        printf("usage:  %s <ip> <port>\n", execname);
        printf("        <ip>       TCP server IP address\n");
        printf("        <port>     TCP server port [1025-65534]\n\n");
}

/* Reads control packet from GPP */
enum host_ctrl_t read_ctrl_pkt(int connfd)
{
	int res;
	enum host_ctrl_t ctrl;

	res = read_tcp_server_stream(connfd, (char *)&ctrl, sizeof(ctrl));
	if (res < 0) {
		ERROR("ERROR, reading from host ");
		return H_SOCK_ERROR;
	}
	else if (res == 0) {
		INFO("GPP disconnected.\n");
		return H_SOCK_ERROR;
	}	
	else {
		if (ctrl == H_SOCK_ERROR) {
			INFO("Got H_SOCK_ERROR that should be HOST_UNKNOWN\n");
			ctrl = HOST_UNKNOWN;
		}
	}

	return ctrl;
}

/* Reads data payload from GPP */
int read_data_payload(int connfd, char *buf, int len)
{
	int res;

	res = read_tcp_server_stream(connfd, buf, len);
	if (res < 0) {
		ERROR("ERROR, reading from GPP ");
		return -1;
	}
	else if (res == 0) {
		INFO("GPP disconnected.\n");
		return -1;
	}

	return res;
}

/* Writes data payload to GPP */
int write_data_payload(int connfd, char *buf, int len)
{
	if (len == 0) return 0;
	int res;

	res = send_tcp_client_data(connfd, buf, len);
	if (res < 0) {
		ERROR("ERROR, writing to GPP ");
		return -1;
	}
	else if (res == 0) {
		INFO("GPP disconnected.\n");
		return -1;
	}

	return res;
}

/* Main listener thread for encoded PCM data stream sent from GPP */
void *listener_thread(void *arg)
{
	int res;
	int connfd = *((int *)arg);
	struct gpp_host_pkt pkt;
//	int connfdEnc;
		
	FILE *testfile_enc = fopen("audio_enc.dat","w");
	if (testfile_enc == NULL)
		printf("Can't open file to output IMBE-encoded PCM data\n"); 


/*	// Start TCP connection to server on decode side
	printf("Init TCP connection to decode side...");
	connfdEnc = start_tcp_client("137.110.115.164", 51819 );
	printf("done.\n");
*/
	
	while (started)
	{
		// Read in control packet to determine state
		pkt.ctrl = read_ctrl_pkt(connfd);
		if (pkt.ctrl == H_SOCK_ERROR) {
			ERROR("read_ctrl_pkt() failure!\n");
			break;
		}

		// If the control packet received is not HOST_START_STREAM, go to the next
		// while loop iteration (back to reading next control packet above)
		if (pkt.ctrl != HOST_START_STREAM)
			continue;

		INFO("HOST_START_STREAM\n");
		end_stream = 0;

		// Start processing data until stop encode/decode control signal received
		do {
			TICK();

			// Read in control packet to determine state
			pkt.ctrl = read_ctrl_pkt(connfd);
			if (pkt.ctrl == H_SOCK_ERROR) {
				ERROR("read_ctrl_pkt() failure!\n");
				break;
			}

			// If control state is HOST_STREAM_DATA, retrieve data payload
			else if (pkt.ctrl == HOST_STREAM_DATA) {
				INFO("HOST_STREAM_DATA\n");
				
				// Read in encoded PCM data
				res = read_data_payload(connfd, pkt.payload, E_BUFF_LEN);
				if (res == -1) {
					ERROR("read_data_payload() failure!\n");
					break;
				}
				else if (res != E_BUFF_LEN) {
					ERROR("res = %d, E_BUFF_LEN = %d\n", res, E_BUFF_LEN);
					continue;
				}
				
				// Write to file
				if (fwrite(pkt.payload, sizeof(char), res, testfile_enc) < 0)
					printf("ERROR writing encoded data to file \n");
				
				fflush(testfile_enc);

			
/*				// Write to server on decode side
				if (send_tcp_client_data(connfdEnc, pkt.payload, res) < 0)
					ERROR("ERROR writing payload to gpp\n");
*/
			}

			TOCK("encode loop ");
		} while (pkt.ctrl != HOST_STOP_STREAM);
		
		INFO("HOST_STOP_STREAM\n");
		end_stream = 1;
	}
	
	fclose(testfile_enc);
//	stop_tcp_client(connfdEnc);
	return NULL;
}

int main(int argc, char **argv)
{
	int res, connfd, count, errorstate;	
	struct host_gpp_pkt pkt;
	struct tcp_server_t ts;
    	pthread_attr_t attr;
	pthread_t lThread;	

	char *snd_cap_device = "plughw:0,0";
/*	
	FILE *testfile;
	if ((testfile = fopen("audio.dat","w")) == NULL)
		printf("Can't open file to write PCM data\n");
*/
	// Print out usage message if command-line argument length not sufficient.
        if (argc < 3) {
                print_usage(argv[0]);
                exit(0);
        }

	// Command-line processing error state variable
	errorstate = 0;

	// Obtain IP address and port from command-line parameters
	if (strlen(argv[1]) <= 15) {
		strcpy(ts.ip, argv[1]);
	}
	else {
		ERROR("Invalid IP address string!\n");
		errorstate++;
	}

	// Obtain forward TCP stream server port from command-line
	res = sscanf(argv[2], "%d", &ts.fport);
	if (ts.fport < 1025 || ts.fport > 65534) {
		ERROR("Not a valid port: %s\n", argv[2]);
		errorstate++;
	}

	//Exit program if we encountered error in the command-line processing
	if (errorstate > 0)
		exit(0);

	// Setup termination signal interrupt handler
        signal((int) SIGINT, loop_shutdown);
	
	// Start TCP connection to GPP server
	printf("Init TCP connection to %s:%d...", ts.ip, ts.fport);
	connfd = start_tcp_client(ts.ip, ts.fport);
	printf("done.\n");

	// Open up PCM capture channel on sound device
	printf("Init PCM capture on sound device:  %s...", snd_cap_device);
	pcm_cap_init(snd_cap_device, SAMPRATE);
	printf("done.\n");

	// Start TCP listener thread
	started = 1;
	end_stream = 0;
    	pthread_attr_init(&attr);
        pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
        pthread_attr_setstacksize(&attr, 131072);
    	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    	res = pthread_create(&lThread, &attr, listener_thread, (void *)&connfd);
     	pthread_attr_destroy(&attr);   	
	
	if (res) {
		ERROR("Failed to create TCP listener thread");
		started = 0;
		exit(0); 
	}

	pkt.ctrl = IMBE_ENCODE_START;
	if (send_tcp_client_data(connfd, (char *)&pkt.ctrl, sizeof(pkt.ctrl)) < 0)
		ERROR("ERROR writing control start to gpp\n");
	
	sleep(1);
	count = 0;
	do {
//		TICK();		

		////////  Capture from audio hardware //////////
		while (pcm_capture(pkt.payload, S_BUFF_LEN) < 0) {
			pcm_capture_prepare();
			ERROR("Buffer Overrun!!!\n");
		}

/*		// Write to file
		if (fwrite(pkt.payload, sizeof(char), BUFF_LEN, testfile) < 0)
			printf("ERROR writing PCM data to file \n");
*/		
		/////  Send to tcp socket ///////////
		pkt.ctrl = IMBE_DATA;
		if (send_tcp_client_data(connfd, (char *)&pkt, sizeof(struct host_gpp_pkt)) < 0)    
			ERROR("ERROR writing payload to gpp\n");

//		TOCK("full loop");
//		INFO("count = %d\n", count++);
//		if (count > 400)
//			started = 0;
	} while (started);
	
	pkt.ctrl = IMBE_ENCODE_STOP;
	if (send_tcp_client_data(connfd, (char *)&pkt.ctrl, sizeof(pkt.ctrl)) < 0)
		ERROR("ERROR writing control stop to gpp\n");

//	if ((fclose(testfile)) < 0)
//		ERROR("Error closing PCM file\n");

	INFO("Waiting to receive encoded samples from GPP...\n");
	while (!end_stream)
		usleep(100);
	
	sleep(5);

	// Close sound card handles
	pcm_cap_close();

	// Close TCP client connection
	stop_tcp_client(connfd);

	// Wait for TCP listen thread to finish before exiting
	//pthread_join(lThread, NULL);

	return 0;
}





