#!/bin/bash
# p25_run.sh:  USRP P25 Demo startup script
#              Authors:  Jeffrey Cuenco   (jcuenco@ucsd.edu)
#                        Anthony Nwokafor (aanwokafor@ucsd.edu)
#
#              Script last modified 04/22/2011
#

# CONSTANTS AND DEFAULTS
RUN_DIR=run

JAVA_GUI=../java_audio/p25_client/p25_client.jar
RADIO_GPP=../gpp/bin/packetizer
RADIO_DSP=../gpp/bin/baseband
RADIO_FPGA=../gpp/bin/modem

# TX output gain
UHD_TX_GAIN=25

# RX input gain
UHD_RX_GAIN=0

# Filenames of Unix Domain Sockets / FIFOs
##UHD_DSP_FIFO=UHD_DSP_FIFO
##DSP_UHD_FIFO=DSP_UHD_FIFO
UHD_DSP_SOCK=uhd_dsp_sock
DSP_UHD_SOCK=dsp_uhd_sock

GPP_DSP_SOCK=gpp_dsp_sock
DSP_GPP_SOCK=dsp_gpp_sock

DSP_CTRL_CHAN=dsp_ctrl_sock
UHD_CTRL_CHAN=uhd_ctrl_sock
MODEM_BASEBAND_CTRL_CHAN=modem_baseband_ctrl_sock

RX_FILTER_COEF=rxfir_coef.bin
TX_FILTER_COEF=txfir_coef.bin

# Run Java GUI by default
RUN_JAVA_GUI=1

# VC55 Serial Port / File paths for TX/RX Filetest
VC55_SERIAL_PORT=/dev/ttyS0
FILETEST_TXFILE=filetest_tx.dat
FILETEST_RXFILE=filetest_rx.dat

# Filetest mode = 1, Live mode (VC55) = 2
GPP_MODE=2

# GPP UDP port
GPP_UDPPORT=8002

# Debug options
GPP_DEBUG_MODE=0
DSP_DEBUG_MODE=0
FPGA_DEBUG_MODE=0
SHOW_CMD_LINE=0

NO_ARGS=0
E_OPTERROR=85

if [ $# -eq "$NO_ARGS" ]    # Script invoked with no command-line args?
then
  echo    "Usage: `basename $0` options (-m:s:t:r:u)"
  echo -e "Runs USRP P25 waveform demo.\n"
  echo -e "Possible arguments:"
  echo -e "  -m  [ LIVE | FILE ]      sets packetizer mode.  LIVE mode       "
  echo -e "                           reads from and writes to the serial    "
  echo -e "                           port specified by -s. FILE mode        "
  echo -e "                           instead reads from a file specified    "
  echo -e "                           by -t when in TX mode and writes to    "
  echo -e "                           a file specified by -r when in RX mode."
  echo -e "                           (default:  LIVE)                     \n"
  echo -e "  -s [ serial_port ]       VC55 serial port, e.g. /dev/ttyS0      "
  echo -e "                           (default:  /dev/ttyUSB0)             \n"
  echo -e "  -t [ tx_file ]           path to source file for TX test mode   "
  echo -e "                           (default:  filetest_tx.dat)          \n"
  echo -e "  -r [ rx_file ]           path to sink file for RX test mode     "
  echo -e "                           (default:  filetest_rx.dat)          \n"
  echo -e "  -u                       use defaults                         \n"

  exit $E_OPTERROR          # Exit and explain usage.
                            # Usage: scriptname -options
                            # Note: dash (-) necessary
fi  

while getopts "m:s:t:r:ud:nc" Option
do
  case $Option in
    m     ) if [ $OPTARG == "LIVE" ]
            then
                echo "LIVE (VC55) mode selected."
                GPP_MODE=2
            else if [ $OPTARG == "FILE" ]
            then
                echo "FILE (Test) mode selected."
                GPP_MODE=1
            fi
            fi
            ;;
    s     ) VC55_SERIAL_PORT=$OPTARG
            echo "Setting VC55 serial port path to:  $VC55_SERIAL_PORT"
            ;;
    t     ) FILETEST_TXFILE=$OPTARG
            echo "Setting Filetest TX filepath to:  $FILETEST_TXFILE"
            ;;
    r     ) FILETEST_RXFILE=$OPTARG
            echo "Setting Filetest RX filepath to:  $FILETEST_RXFILE"
            ;;
    u     ) echo "Running demo using defaults.";;
    d     ) if [ $OPTARG == "gpp" ]
            then
                echo "Packetizer Debug mode enabled."
                GPP_DEBUG_MODE=1
            else if [ $OPTARG == "dsp" ]
            then
                echo "Baseband Debug mode enabled."
                DSP_DEBUG_MODE=1
            else if [ $OPTARG == "fpga" ]
            then
                echo "Radio Debug mode enabled."
                FPGA_DEBUG_MODE=1
            fi
            fi
            fi
            ;;
    n     ) RUN_JAVA_GUI=0
            echo "No GUI mode enabled."
            ;;    
    c     ) SHOW_CMD_LINE=1
            echo "Showing command-lines to be executed only (no run)."
            ;;
    *     ) echo "Unimplemented option chosen.";;   # Default.
  esac
done

shift $(($OPTIND - 1))
#  Decrements the argument pointer so it points to next argument.
#  $1 now references the first non-option item supplied on the command-line
#+ if one exists.

#exit $?

# Increase the buffer for the UHD
sudo sysctl -w net.core.rmem_max=50000000
sudo sysctl -w net.core.wmem_max=50000000

# Assign GPP source and sink to test filepaths
GPP_SOURCE=$VC55_SERIAL_PORT
GPP_SINK=$FILETEST_RXFILE
if [ $GPP_MODE == "1" ]
then
    GPP_SOURCE=$FILETEST_TXFILE
    GPP_SINK=$FILETEST_RXFILE
fi

# For debugger purposes
if [ $GPP_DEBUG_MODE == "1" ]
then
    RADIO_GPP="gdb --args ../gpp/bin/packetizer"
fi

if [ $DSP_DEBUG_MODE == "1" ]
then
    RADIO_DSP="gdb --args ../gpp/bin/baseband"
fi

if [ $FPGA_DEBUG_MODE == "1" ]
then
    RADIO_FPGA="gdb --args ../gpp/bin/modem"
fi

# Kill old instances of terminals and radio components
killall gdb `basename $RADIO_GPP` `basename $RADIO_DSP` `basename $RADIO_FPGA`
killall xterm 

# Create run directory if not created yet
if [ ! -d $RUN_DIR ]
then
	mkdir -p $RUN_DIR
fi

cd $RUN_DIR

#if [ ! -p $UHD_DSP_FIFO ] 
#then
#	mkfifo $UHD_DSP_FIFO
#fi

#if [ ! -p $DSP_UHD_FIFO ] 
#then
#	mkfifo $DSP_UHD_FIFO
#fi

if [ ! -e $TX_FILTER_COEF ]
then
	cp ../$TX_FILTER_COEF .
fi

if [ ! -e $RX_FILTER_COEF ]
then
	cp ../$RX_FILTER_COEF .
fi

if [ $SHOW_CMD_LINE == "0" ]
then
	echo -n "Starting radio1 ...\n"
	#xterm -hold -e $RADIO_GPP $GPP_SOURCE $GPP_SINK $GPP_UDPPORT $DSP_GPP_SOCK $GPP_DSP_SOCK $UHD_CTRL_CHAN $GPP_MODE &
	xterm -hold -e $RADIO_DSP $GPP_DSP_SOCK $DSP_GPP_SOCK $DSP_UHD_SOCK $UHD_DSP_SOCK $DSP_CTRL_CHAN $MODEM_BASEBAND_CTRL_CHAN &
	xterm -hold -e $RADIO_FPGA --rx_osock=$UHD_DSP_SOCK --tx_isock=$DSP_UHD_SOCK --rc_sock=$DSP_CTRL_CHAN --fc_sock=$UHD_CTRL_CHAN --tx_gain $UHD_TX_GAIN --rx_gain $UHD_RX_GAIN --bbc_sock $MODEM_BASEBAND_CTRL_CHAN &
	#xterm -hold -e "$RADIO_FPGA --rx_osock=$UHD_DSP_SOCK --tx_isock=$DSP_UHD_SOCK --rc_sock=$DSP_CTRL_CHAN --fc_sock=$UHD_CTRL_CHAN > modem_output.txt" &
else
	echo "xterm -hold -e $RADIO_GPP $GPP_SOURCE $GPP_SINK $GPP_UDPPORT $DSP_GPP_SOCK $GPP_DSP_SOCK $UHD_CTRL_CHAN $GPP_MODE &"
	echo "xterm -hold -e $RADIO_DSP $GPP_DSP_SOCK $DSP_GPP_SOCK $DSP_UHD_SOCK $UHD_DSP_SOCK $DSP_CTRL_CHAN &"
	echo "xterm -hold -e $RADIO_FPGA --rx_osock=$UHD_DSP_SOCK --tx_isock=$DSP_UHD_SOCK --rc_sock=$DSP_CTRL_CHAN --fc_sock=$UHD_CTRL_CHAN --tx_gain $UHD_TX_GAIN --rx_gain $UHD_RX_GAIN &"
fi

#xterm -e $RADIO_DSP gpp_dsp_sock dsp_gpp_sock dsp_uhd_sock uhd_dsp_sock dsp_ctrl_sock &
#xterm -e $RADIO_FPGA uhd_dsp_sock dsp_uhd_sock dsp_ctrl_sock > fpga_log &

if [ $RUN_JAVA_GUI == "1" ]
then
	java -jar $JAVA_GUI &
fi

